
/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/*
 * callgraph.c - implementation of 
 * 	functions associated with building and maintaining a callgraph 
 * 	for a set of functions
 *
 * author: Jason D. Hiser
 * Creation: 8-1-2000
 * 
 */

#ifdef CALLGRAPH
#ifdef STACHE

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "vpo.h"
#include "assign.h"
#include "callgraph.h"
#include "lig.h"
#include "locals.h"
#include "misc.h"
#include "stache.h"
#include "vectors.h"

#include "placement.h"   /* Machine dependent. */
#include "rtl.h"         /* Machine dependent. */

int   maxGlobalNum = 0;
static struct indirect_functions_t *indirect_functions = 0;

void  dump_call_graph();
int   CG_search(struct call_graph_node *findme,
                struct call_graph_node *current_node);
static void add_call_by_ptr_arcs();

void  fix_local_rep_count();
void  clear_seen();
void  guestimate_stache_usage();
int   make_guestimate(struct call_graph_node *cgn);
void  number_nodes(struct call_graph_node *cgn, int *num);
void  mark_recursive_edges(struct call_graph_node *cgn);
static void read_library_file();
struct call_graph_node *add_to_cg();
void  mini_optimize();
static void build_call_graph_from_static_input();
static void buil_call_graph_from_optimized_funcs();
void  build_call_graph_from_optimized_funcs();
void  calculate_node_estimates(struct call_graph_node *cgn);

/*
 * This is the top node in the call graph.  It may not be a source or sink
 * but it gives us a starting point to search for such things
 */
struct call_graph_node *the_call_graph = 0;

void  annotate_call_graph_with_runtime_info();

/* functions to help create a call graph */
struct call_graph_node *add_func_to_call_graph(char *id)
{
   struct call_graph_node *newNode = find_call_graph_node(id);

   /* don't add it to the graph if it already exists */
   if (newNode != 0)
      return newNode;

   /* initializes to zero */
   newNode = (struct call_graph_node *) CALLOC_PERM(
                                                sizeof(struct call_graph_node));
   bzero(newNode, sizeof(struct call_graph_node));


   newNode->name = (char *) strdup(globalname(id));
   /* then copy the ID into the structure */
   copyloc(newNode->func_id, id);

   /* We don't really need to null-terminate this because *
    * calloc left it initialized to zero.  We do this again *
    * however to avoid confusion.
    */
   *skiploc(newNode->func_id) = 0;

   newNode->next_func = the_call_graph;
   the_call_graph = newNode;
   return newNode;
}


void add_call_to_call_graph_node(struct call_graph_node *current_func,
                                 struct call_graph_node *new_callee, int freq)
{
   struct call_graph_list_node *newListNode;

   if (!isGlobal(current_func->func_id) || !isGlobal(new_callee->func_id)) {
      assert(0);
   }
   /* if it's directly called, we're already a direct child */
   /* no need to do any work */
   if (newListNode = is_directly_called(current_func, new_callee)) {
      newListNode->static_count += freq;
      /* we update the dynamic count here so that
       * we can use the best guess of a dynamic count
       * if we don't really have dynamic data */
      newListNode->dynamic_count += freq;
      if (newListNode = is_direct_parent(new_callee, current_func)) {
         newListNode->static_count += freq;
         newListNode->dynamic_count += freq;
      }
      else {
         fprintf(stderr,
                 "whoa nelly.  our child doesn't know"
                 " that we're the parent\n");
         exit(1);
      }
      return;
   }


   /* otherwise, allocate a new list node */
   newListNode = (struct call_graph_list_node *)
    CALLOC_PERM(1 * sizeof(struct call_graph_list_node));
   bzero(newListNode, sizeof(struct call_graph_list_node));

   /* set the function pointer to point to the new callee */
   newListNode->the_func = new_callee;
   newListNode->static_count = freq;
   newListNode->dynamic_count = freq;

   /* and add it to the beginning of the list */
   newListNode->next_list_node = current_func->next_child;
   current_func->next_child = newListNode;

   /* we also want to repeat the process for the */
   newListNode = (struct call_graph_list_node *)
    CALLOC_PERM(1 * sizeof(struct call_graph_list_node));
   bzero(newListNode, sizeof(struct call_graph_list_node));

   /* set the function pointer to point to the new callee */
   newListNode->the_func = current_func;
   newListNode->static_count = freq;
   newListNode->dynamic_count = freq;

   /* and add it to the beginning of the list */
   newListNode->next_list_node = new_callee->next_parent;
   new_callee->next_parent = newListNode;

   return;
}

/* search for a node by ID */
struct call_graph_node *find_call_graph_node(char *id)
{
   struct call_graph_node *temp;

   for (temp = the_call_graph; temp; temp = temp->next_func) {
      if (cmploc(id, temp->func_id) == 0)
         return temp;
   }
   return 0;
}


/* determine if the call graph has a loop that includes this node. */
int is_indirectly_recursive(struct call_graph_node *node)
{
   struct call_graph_node *cgn;
   struct call_graph_list_node *cgln;

   for (cgn = the_call_graph; cgn; cgn = cgn->next_func)
      cgn->seen = 0;

   for (cgln = node->next_child; cgln; cgln = cgln->next_list_node) {
      if (CG_search(node, cgln->the_func))
         return TRUE;
   }
   return FALSE;
}

int CG_search(struct call_graph_node *findme,
              struct call_graph_node *current_node)
{
   struct call_graph_list_node *cgln;

   if (current_node == findme)
      return TRUE;
   if (current_node->seen) {
      return FALSE;
   }
   else {
      current_node->seen = TRUE;
      for (cgln = current_node->next_child; cgln; cgln = cgln->next_list_node) {
         if (CG_search(findme, cgln->the_func))
            return TRUE;
      }
      return FALSE;
   }
}



/* is the smallest loop of size 1 */
extern int is_directly_recursive(struct call_graph_node *node)
{
   struct call_graph_list_node *temp;

   for (temp = node->next_child; temp; temp = temp->next_list_node) {
      if (node == temp->the_func)
         return TRUE;
   }
   return FALSE;
}


/* determine if a child is a direct, or indirect predecessory of a node */
struct call_graph_list_node *is_directly_called(struct call_graph_node
                                                *caller,
                                                struct call_graph_node
                                                *callee)
{
   struct call_graph_list_node *temp;

   for (temp = caller->next_child; temp; temp = temp->next_list_node) {
      if (callee == temp->the_func)
         return temp;
   }
   return FALSE;
}

/* Determine if a caller is a parent of callee
 * by looking through callee's parent list.
 */
struct call_graph_list_node *is_direct_parent(struct call_graph_node *callee,
                                              struct call_graph_node *caller)
{
   struct call_graph_list_node *temp;

   for (temp = callee->next_parent; temp; temp = temp->next_list_node) {
      if (caller == temp->the_func)
         return temp;
   }
   return FALSE;
}

int is_indirectly_called(struct call_graph_node *caller,
                         struct call_graph_node *callee)
{
   assert(("Unimplemented function, is_indirectly_called", 0));
}

static void release_rtls(struct list *first_rtl)
{
   struct list *rtlptr, *drtl;

   /*
    * Release all of the RTLs in the basic block
    */
   rtlptr = first_rtl;
   while (rtlptr) {
      drtl = rtlptr;
      rtlptr = rtlptr->next;
      FREE(drtl->val);
      FREE(drtl->se);
      FREE(drtl->InitialDeads);
      rfree(drtl->deads);
      /* VET relies on receiving a unique ptr id for each RTL during
       * the compilation of a function. */
      FREETYPE(drtl, LISTTYPE);
   }

}


char *function_name_to_id(char *func_name)
{
   int   i = 0;
   char *temp;

   while (i <= maxGlobalNum) {
      temp = getGlobal(i);
      if (isDefinedGlobal(temp) && strcmp(globalname(temp), func_name) == 0) {
         return temp;
      }
      i++;
   }
   return 0;
}

/*
 * we shouldn't need very many rtls to find out the name of the function called
 * consequently, we only save a few.  we can save more later if necessary
 */
#define MAXRTLSINCALLGRAPHBUILD 10

/* a top-level function that makes a pass over the input and builds
 * the entire callgraph.  Then it rewinds the input for the optimizer
 * to make a second pass over it, doing full optimizations 
 */
void build_call_graph()
{
  build_call_graph_from_static_input();
  
#if 0      
   if (stache_policy == STUPID_STACHE_POLICY)
      build_call_graph_from_static_input();
   else if (stache_policy == COLORING_STACHE_POLICY
            || stache_policy == BOTH_STACHE_POLICY) {
      build_call_graph_from_optimized_funcs();
      build_LIG();
      color_LIG();
   }
   else {          /*if(swpu) */

      /* using the locality profile needs this type of callgraph */
      build_call_graph_from_optimized_funcs();
   }
#endif   
}

void build_call_graph_from_optimized_funcs()
{
   int   i;
   struct bblock *cblk;
   struct list *ptr, *rtl;
   int type;
   struct bblock *exists;
   struct list *prev;
   int   noflow, direct, rt, index, rcont;
   unsigned int ui;
   int   LineNumber = 0;
   char *a, *label, *line, *setptr, setline[MAXSTRING];
   struct call_graph_node *CurrNode = NULL;

#ifndef WIDES
   int   d_f_s_count, b_d_t_count;
#endif

   varargs = FALSE;

   /*
    * Initialize the local items that control the read phase
    */
   noflow = TRUE;
   cblk = top;
   prev = (struct list *) NULL;
   if (swz) {
      assert(FALSE == CollectingGlobals);
      CollectingGlobals = TRUE;
   }

   /*
    * This loop reads in the input file in RTL file format.  The first
    * character of each line indicates what kind of information is present on
    * the rest of the line.  The main switch statement uses the first character
    * to transfer control to the appropriate case which handles the remaining
    * information (if any) on that line.  There is no error recovery here.  The
    * format is either correct or an error message is printed and the process
    * terminates with a non-zero exit condition.
    */
   for (;;) {
      ++LineNumber;
      switch (type = getchar()) {

          /*
           * End of the input file
           */
       case EOF:

          /*
           * Optimize the last procedure, dump it to the output file and
           * exit this loop.
           */
          mini_optimize();
          CurrNode = add_to_cg();

          /* If we are doing data placement, we need to go back over all
           * of the ASMLINEs and gather the ones that pertain to globals
           * into the corresponding global symbol list entries.
           */
          if (swz && swpw) {
             if (NULL != CurrNode)
                RecordCurrentCallGraphNode(CurrNode);
             for (cblk = top; cblk; cblk = cblk->next) {
                for (rtl = cblk->rtls; rtl; rtl = rtl->next) {
                   if (ASMLINE & rtl->type)
                      (void) NotProcessingGlobal(rtl->val);
                }
             }
             /* Make NotProcessingGlobal() wrap up whatever it was doing. */
             (void) NotProcessingGlobal("\0");
          }

          goto annotate_callgraph;

#ifdef VET

          /*
           * Read in the basename of the source file being compiled.
           */

       case (int) 'n':
          strcpy(basename, gtline());
          break;
#endif

          /*
           * End of function
           */
       case (int) '*':

          /*
           * Optimize this procedure and dump it to the output file
           */
          mini_optimize();
          CurrNode = add_to_cg();

          /* If we are doing data placement, we need to go back over all
           * of the ASMLINEs and gather the ones that pertain to globals
           * into the corresponding global symbol list entries.
           */
          if (swz && swpw) {
             if (NULL != CurrNode)
                RecordCurrentCallGraphNode(CurrNode);
             for (cblk = top; cblk; cblk = cblk->next) {
                for (rtl = cblk->rtls; rtl; rtl = rtl->next) {
                   if (ASMLINE & rtl->type)
                      (void) NotProcessingGlobal(rtl->val);
                }
             }
          }

          /*
           * Clear out all the garbage left behind by the optimization
           * process.
           */
          for (type = hregnum; type < regallocptr; type++)
             vecmap[regname(type)] = -1;
          regallocptr = hregnum;
          clrsets();
          clear_locals();
          clear_loops();
          clear_function();
          clear_call_graph();

#ifdef DISPOSE

          /*
           * Reclaim all of the storage used by the function that we have
           * just optimized in one move.  Re-initialize the free list
           * pointers.
           */
          dispose_func();
          for (type = 0; type < ENDFREETYPES; type++)
             freeptrs[type] = (char **) NULL;
#endif

          /*
           * Re-initialize variables to get ready for the next procedure in
           * the input file.
           */
          LeafFunction = TRUE;
          initmap();
          top = new_basic_block((char *) NULL);
          ret_block = NULL;

          /*
           * Re-initialize the local items that control the read phase
           */
          noflow = TRUE;
          cblk = top;
          prev = (struct list *) NULL;

          fn = 0;

          /*
           * Eat the end-of-line marker
           */
          (void) getchar();
          break;

          /* 
           * Metadata line
           */
       case (int) 'm':
          /*
           * We just read it up and pass it along to the output
           */
          ptr = new_rtl(cblk, METADATALINE, prev);
          ptr->val = gtline();
          process_metaline(ptr->val);
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Argument build area size line
           */
       case (int) 'A':
          ptr = new_rtl(cblk, ARGBLINE, prev);
          ptr->val = gtline();
          prev = ptr;
          noflow = FALSE;
          LeafFunction = FALSE;
          break;

          /*
           * Assembly line
           */
       case (int) '-':

          /*
           * All we have to do is read and retain the line for now
           */
          line = gtline();
          ptr = new_rtl(cblk, ASMLINE, prev);
          ptr->val = line;
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Delayed assembly line
           */
       case (int) '_':

          /*
           * All we have to do is read and retain the line for now
           */
          line = gtline();
          ptr = new_rtl(cblk, ASMLINE | DELAYLINE, prev);
          ptr->val = line;
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Trash line for the common subexpression eliminator
           */
       case (int) 't':

          /*
           * All we have to do is read and retain the line for now
           */
          ptr = new_rtl(cblk, TRASHLINE, prev);
          ptr->val = gtline();
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Reserve line for the local register allocator
           */
       case (int) 'r':

          /*
           * All we have to do is read and retain the line for now
           */
          ptr = new_rtl(cblk, RESLINE, prev);
          ptr->val = gtline();
          prev = ptr;
          noflow = FALSE;
          break;

          /* New style reserve line. We need to mark the previous
           * line.  We should check that the register being specified
           * is used in the previous line. Add this later
           */

       case (int) 'R':
          prev->type = RESLINE;
          line = gtline();
          FREE(line);
          break;

          /*
           * Parameter register line. This line contains the registers
           * that have outgoing parameters.  We must build a side
           * effect line and put store it in the side effect field of
           * the previous instruction, which should be a call.  We need
           * to check.
           */
       case (int) 'P':
          /*
           * Read in the line and keep it for now
           */
          ptr = new_rtl(cblk, PARMLINE, prev);
          ptr->val = gtline();
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Set line
           */
       case (int) 'S':
          {
             /*
              * Locate all of the aliases of this item and note each of them as
              * a set.
              */
             char *aline;
             setptr = setline;
             aline = line = gtline();
             for (; *line; line = skiploc(line)) {
                for (label = aliases(line); *label; label++) {
                   copyloc(setptr, label);
                   setptr = skiploc(setptr);
                   label = skiploc(label);
                }
             }
             *setptr = '\0';
             FREE(aline);
             ptr = new_rtl(cblk, SETLINE, prev);
             ptr->val = strcpy(CALLOC(strlen(setline) + 1), setline);
             prev = ptr;
             noflow = FALSE;
             break;
          }

          /*
           * Caller-save line
           */
       case (int) 'C':

          /*
           * Read in the line and keep it for now
           */
          ptr = new_rtl(cblk, CSLINE, prev);
          ptr->val = gtline();
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Comment line
           */
       case (int) '#':

          /*
           * We just read it up and pass it along to the output
           */
          ptr = new_rtl(cblk, COMMENT, prev);
          ptr->val = gtline();
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Register type map
           */
       case (int) 'M':

          /*
           * This line contains a map indicating which letter should be used
           * to denote each register type when printing RTLs in decoded
           * format.
           */
          line = gtline();
          if (strlen(line) < MAXSTRING)
             (void) strcpy(regprt, line);
          FREE(line);
          break;

          /*
           * Function name line
           */
       case (int) 'f':

          /*
           * This line contains the name of the current function, which is
           * read so that it can be referred to using the "fn" variable.
           */
          if (fn) {
             (void) fprintf(stderr, "vpo (main): attempt to rename %s()\n",
                            fn);
             line = gtline();
             FREE(line);
          }
          else {
             ptr = new_rtl(cblk, ASMLINE | FUNCNAME, prev);
             fn = ptr->val = gtline();
             prev = ptr;
             noflow = FALSE;

          }
          break;

          /*
           * Global reference line
           */
       case (int) 'g':

          /*
           * This line contains the name of a function that is referenced
           * somewhere in this function.  This information is used when
           * building a call graph for EASE.
           */
          add_func(line = gtline());
          FREE(line);
          break;

          /*
           * Side effect line
           */
       case (int) 's':

          /*
           * This line contains the side effect of returning
           */
          if (sret) {
             (void) fprintf(stderr, "vpo (main): redefined return effect\n");
             line = gtline();
             FREE(line);
          }
          else
             sret = gtline();
          break;

          /*
           * Varargs line
           */
       case (int) 'V':

          /*
           * This line indicates that varargs is used in this function
           */
          varargs = TRUE;
          (void) getchar();
          break;

          /*
           * Use line
           */
       case (int) 'u':

          /*
           * In order to speed up register allocation in the future, we need
           * to obtain the internal index number of the register being used.
           * In the case that the register used actually takes up more than
           * one index position, new use lines are internally created, each
           * holding one of the register index positions affected by the use
           * of the register.  Note that the cost field of the list structure
           * holds the register index number.
           */
          ptr = new_rtl(cblk, USELINE, prev);
          ptr->val = gtline();
          rt = regtype(ptr->val);
          ptr->cost = index = regtoindexf(rt, regnum(ptr->val));
          if ((rcont = contregf(rt)) > 1) {
             while (--rcont) {
                prev = ptr;
                ptr = new_rtl(cblk, USELINE, prev);
                ptr->cost = ++index;
             }
          }
          prev = ptr;
          noflow = FALSE;
          break;

          /*
           * Declaration line
           */
       case (int) 'd':
             /*
              * If declaration() returns a zero, then the line contained a
              * global identifier.  If it returns a one, the line contained an
              * argument and if it returns a value greater than one, the line
              * was a local.
              */
             declaration(line = gtline());
             break;

          /*
           * Case line
           */
       case (int) 'c':

          /*
           * This line is part of a jump address lookup table indicating
           * where a switch statement may potentially transfer control to.
           * What is important here is that the rest of the line, although
           * treated as an assembly line, contains a label reference that has
           * to be integrated to assemble an accurate representation of the
           * control flow graph for this function.
           */
          ptr = new_rtl(cblk, ASMLINE | CASELINE, prev);
          ptr->val = gtline();
          prev = ptr;
          label = findlabel(ptr->val, &direct);
          if (!(cblk->right = getlabel(label)))
             cblk->right = new_basic_block(label);
          add_predecessor(cblk, cblk->right);
          cblk->down = cblk->left = new_basic_block((char *) NULL);
          cblk->down->up = cblk;
          add_predecessor(cblk, cblk->left);
          cblk = cblk->left;
          cblk->status |= FALLTHRU;
          prev = (struct list *) NULL;
          noflow = FALSE;
          break;

       case (int) 'l':
          line = gtline();
          ptr->type |= CASELINE;
          label = findlabel(line, &direct);
          if (!(cblk->right = getlabel(label)))
             cblk->right = new_basic_block(label);
          add_predecessor(cblk, cblk->right);
          cblk->down = cblk->left = new_basic_block((char *) NULL);
          cblk->down->up = cblk;
          add_predecessor(cblk, cblk->left);
          cblk = cblk->left;
          cblk->status |= FALLTHRU;
          prev = (struct list *) NULL;
          noflow = FALSE;
          FREE(line);
          break;

          /*
           * End-flow label
           */
       case (int) 'E':

          /*
           * Note that this line is a label, and it is treated just as if it
           * began with an 'L', the only difference is that this label cuts
           * the normal flow, so that the code immediately above the label is
           * not assumed to fall into the code below the label.  This type of
           * label can be used to properly represent the control flow created
           * by lookup table sequences on certain target machines.
           */
          noflow = TRUE;

          /*
           * Label line
           */
       case (int) 'L':

          /*
           * We must start a new basic block.  First, see if we have already
           * seen this label referenced and already allocated a block for it.
           */
          if (exists = getlabel(label = gtlab())) {

             /*
              * If we have already allocated a basic block for this label,
              * then there are a couple of things we have to do depending on
              * the state of the current basic block.  First, we check to see
              * if the current basic block has anything in it at all (RTLs or
              * a label).  If it does, then we just indicate that it falls
              * into this new block.  If not, then there is no need to have
              * an empty block just lying around, so we put it aside and
              * connect the new block to the empty blocks predecessor.
              */
             if (cblk->label || cblk->rtls) {
                cblk->down = exists;
                cblk->down->up = cblk;
                if (!noflow) {
                   cblk->left = exists;
                   add_predecessor(cblk, cblk->left);
                }
             }
             else {
                if (exists->up = cblk->up) {
                   exists->up->down = exists;
                   if (exists->up->left == cblk) {
                      exists->up->left = exists;
                      if (cblk->preds)
                         cblk->preds->next = exists->preds;
                      exists->preds = cblk->preds;
                      cblk->preds = (struct blist *) NULL;
                   }
                }
                exists->status = cblk->status;
                remv_basic_block(cblk);
             }
             cblk = exists;
             prev = (struct list *) NULL;
          }
          else if (cblk->label || prev) {
             cblk->down = cblk->left = new_basic_block(label);
             cblk->down->up = cblk;
             if (!noflow) {
                cblk->left = cblk->down;
                add_predecessor(cblk, cblk->left);
             }
             cblk = cblk->down;
             prev = (struct list *) NULL;
          }
          else
             cblk->label = label;
          if (!noflow)
             cblk->status |= FALLTHRU;
          noflow = FALSE;
          break;

          /*
           * RTL line
           */
       case (int) '+':
          /*
           * First we read the line in and store it away, then we determine
           * if it affects the control flow of the function.  If it does, we
           * need to reflect the fact in the control flow data structure.
           */
          noflow = FALSE;
          ptr = new_rtl(cblk, RTLINE, prev);
          gtrtl(&ptr->val, &ptr->InitialDeads);
          prev = ptr;
          if (ptr->val) {
             if ((label = is_pc(ptr->val))
                 && (label = findlabel(label + 1, &direct))) {
                if (direct) {
                   cblk->status |= JUMP;
                   if (!(cblk->left = getlabel(label))) {
                      cblk->left = new_basic_block(label);
                      /* If this is a return instruction, its target is the
                       * return block. Remember the return block if we haven't
                       * already done so.
                       */
                      if (ret_block == NULL && fmatch(label, ret))
                         ret_block = cblk->left;
                   }
                   add_predecessor(cblk, cblk->left);
                   cblk->down = new_basic_block((char *) NULL);
                   cblk->down->up = cblk;
                   cblk = cblk->down;
                   noflow = TRUE;
                }
                else {
                   if (!(cblk->right = getlabel(label)))
                      cblk->right = new_basic_block(label);
                   add_predecessor(cblk, cblk->right);
                   cblk->down = cblk->left = new_basic_block((char *) NULL);
                   cblk->down->up = cblk;
                   add_predecessor(cblk, cblk->left);
                   cblk = cblk->down;
                   cblk->status |= FALLTHRU;
                }
                prev = (struct list *) NULL;
             }
          }
          break;

          /*
           * If we get here, then the input file is not valid RTL format
           */
       default:
          (void) fprintf(stderr, "vpo (main): bad input '%c' near line %d\n",
                         (char) type, LineNumber);
          exit(EXIT_FAILURE);
      }
   }

 annotate_callgraph:

   if (swpr)
      annotate_call_graph_with_runtime_info();

   /* multiply the rep count on the local by the number of
    * times the function is called using best-guess
    * static or dynamic info if available */
   calculate_node_estimates(the_call_graph);
   fix_local_rep_count();

   /* 
    * add the arcs that *might* be there if a function
    * makes a call by pointer.
    */
   add_call_by_ptr_arcs();

   /* figure out which nodes might be recursive.  These can't have
    * any stache usage or the recursion could cause a stache overflow
    */
   determine_callgraph_recursiveness();

   if (swpl)
      read_library_file();

   guestimate_stache_usage();

   /*
    * If we are doing data placement, do it now before the callgraph storage
    * gets released.
    */
   if (swz && swpw)
      PlaceGlobalData(the_call_graph);

   /* If the user requests we dump the call graph, we do so before
    *  we zap the global storage.
    */
   if (debug_opt("dump_call_graph"))
      dump_call_graph();

#ifdef PARTITIONING
   /* Once the call graph is built (and info about globals, locals, etc., is
    * done), we can make partitioning decisions.
    */
   if (swpu && swpw) {
      extern char partition_description_filename[MAXSTRING];
      extern void drive_partitioning(char *);

      drive_partitioning(partition_description_filename);
   }
#endif

   /* here we rewind stdin, clear out all of the global/local
    * storage and re-initialize VPO's data structures
    * so that VPO doesn't accidentally redeclare them 
    */

   rewind(stdin);
   for (type = hregnum; type < regallocptr; type++)
      vecmap[regname(type)] = -1;
   clrsets();
   clear_locals();

   /* get rid of the globals so that the next pass can re-allocate them
    * without an error */
   for (i = 0; i < MAPREGS; i++)
      vecmap[i] = -1;
   gallocptr = 0;
   CollectingGlobals = FALSE; /* Data placement only collects once. */

   /* re-init the hard-regs, etc */
   hregnum = 0;
   regallocptr = 0;
   inithardregs();
   hregnum = regallocptr;

   fn = sret = (char *) NULL;
   last = bottom = (struct bblock *) NULL;
   top = new_basic_block((char *) NULL);
   ret_block = NULL;
   varargs = FALSE;

   /*
    * Initialize the local items that control the read phase
    */
   noflow = TRUE;
   cblk = top;
   prev = (struct list *) NULL;

#ifdef DISPOSE
   if (swz)
      dispose_func();
#endif
}

void build_call_graph_from_static_input()
{
   char  funcName[MAXSTRING];
   char  basename[MAXSTRING];
   int   type;
   char *line;
   register struct list *ptr, *prev, *first_rtl = 0;
   int   line_num = 0;
   int   i;
   int   done = FALSE;
   char  curr_id[MAXSTRING] = { 0, };
   struct call_graph_node *curr_node = 0;

   while (!done) {
      line_num++;
      type = (int) getchar();
      switch (type) {
          /*
           * Read in the basename of the source file being compiled.
           */
       case (int) 'n':
          {
             strcpy(basename, line = gtline());
             FREE(line);
          }
          break;

       case EOF:
          {
             done = TRUE;
             break;
          }
       case (int) '*':
          /*
           * end of function line
           */
          {
             struct locuse *locals = locs;
             int   local_size = 0;
             int   num_locals = 0;
             int   param_size = 0;
             int   num_params = 0;
             for (locals = locs; locals; locals = locals->next) {
                if (!locals->arg) {
                   local_size += locals->size;
                   num_locals++;
                }
                else {
                   param_size += locals->size;
                   num_params++;
                }
             }
             if (curr_node) {
                curr_node->size_of_locals = local_size;
                curr_node->num_of_locals = num_locals;
                curr_node->size_of_params = param_size;
                curr_node->num_of_params = num_params;
             }
             for (type = hregnum; type < regallocptr; type++)
                vecmap[regname(type)] = -1;
             regallocptr = hregnum;
             clrsets();
             clear_locals();
          }
          /*
           * Argument build area size line and other don't care cases.
           */
       case (int) 'A':
       case (int) '-':
       case (int) '_':
       case (int) 'm':
       case (int) 't':
       case (int) 'r':
       case (int) 'R':
       case (int) 'P':
       case (int) 'S':
       case (int) 'C':
       case (int) '#':
       case (int) 'M':
       case (int) 'g':
       case (int) 's':
       case (int) 'V':
       case (int) 'u':
          line = gtline();
          FREE(line);
          break;

       case (int) 'c':
       case (int) 'l':
       case (int) 'E':
       case (int) 'L':
             line = gtline();
             FREE(line);
             release_rtls(first_rtl);
             first_rtl = 0;
             prev = 0;
             break;

       case (int) 'f':
             strcpy(funcName, line = gtline());
             FREE(line);
             strcpy(curr_id, function_name_to_id(funcName));
             curr_node = add_func_to_call_graph(curr_id);
             curr_node->has_data = 1;
             break;

       case (int) '+':
             ptr = new_rtl(NULL, RTLINE, prev);
             gtrtl(&ptr->val, &ptr->InitialDeads);
             prev = ptr;
             if (first_rtl == 0)
                first_rtl = ptr;
             if (ptr->val == 0)
                break;

             if (is_call(ptr->val)) {
                char *new_id = identify_function_name(ptr);
                struct call_graph_node *new_node = 0;

                if (new_id == 0) {
                   curr_node->makes_call_by_pointer = TRUE;
                   break;
                }

                new_node = add_func_to_call_graph(new_id);

                add_call_to_call_graph_node(curr_node, new_node, 1);
                release_rtls(first_rtl);
                first_rtl = 0;
                prev = 0;
             }
             else if (is_pc(ptr->val)) {
                release_rtls(first_rtl);
                first_rtl = 0;
                prev = 0;
             }
             break;

       case (int) 'd':
          {
             int   retVal, num;
             char *a;
             retVal = declaration(line = gtline());
             a = line;
             while (*a++ != 0) ;
             if (isGlobal(a)) {
                num = getGlobalNum(a);
                if (num > maxGlobalNum)
                   maxGlobalNum = num + 1;
             }

             if (!retVal)
                FREE(line);

             break;
          }
       default:
          (void) fprintf(stderr,
                         "vpo (build_call_graph):"
                         " bad input '%c' near line %d\n", (char) type,
                         line_num);
      }
   }


   if (swpr)
      annotate_call_graph_with_runtime_info();

   /* multiply the rep count on the local by the number of
    * times the function is called using best-guess
    * static or dynamic info if available */
   calculate_node_estimates(the_call_graph);
   fix_local_rep_count();

   /* 
    * add the arcs that *might* be there if a function
    * makes a call by pointer.
    */
   add_call_by_ptr_arcs();

   /* figure out which nodes might be recursive.  These can't have
    * any stache usage or the recursion could cause a stache overflow
    */
   determine_callgraph_recursiveness();

   if (swpl)
      read_library_file();

   guestimate_stache_usage();

   /* if the user requests we dump the call graph, we do so before */
   /* we zap the global storage                                    */
   if (1 || debug_opt("dump_call_graph"))
      dump_call_graph();

   /* Here we rewind stdin, clear out all of the global/local
    * storage and re-initialize VPO's data structures
    * so that VPO doesn't accidentally redeclare them.
    */

   rewind(stdin);
   for (type = hregnum; type < regallocptr; type++)
      vecmap[regname(type)] = -1;
   clrsets();
   clear_locals();

   /* get rid of the globals so that the next pass can re-allocate them
    * without an error */
   for (i = 0; i < MAPREGS; i++)
      vecmap[i] = -1;
   gallocptr = 0;

   /* re-init the hard-regs, etc */
   hregnum = 0;
   regallocptr = 0;
   inithardregs();
   hregnum = regallocptr;

   fn = sret = (char *) NULL;
   last = bottom = (struct bblock *) NULL;
   top = new_basic_block((char *) NULL);
   ret_block = NULL;
   varargs = FALSE;
}

void dump_call_graph()
{
   const int cg_linelen = 6;
   struct call_graph_node *cgn = the_call_graph;
   int   count = 0;

   while (cgn) {
      struct call_graph_list_node *cgln;
      fprintf(stderr, "*******************************************\n");
      fprintf(stderr, "'%s' ", globalname(cgn->func_id));
      if (cgn->has_data)
         fprintf(stderr, " [local{%d:%d} params{%d:%d}] ", cgn->num_of_locals,
                 cgn->size_of_locals, cgn->num_of_params,
                 cgn->size_of_params);
      if (cgn->is_indirectly_recursive)
         fprintf(stderr, "[recursive] ");
      if (cgn->has_data)
         fprintf(stderr, "[compiling] ");
      if (cgn->library_func)
         fprintf(stderr, "[library_call] ");
      fprintf(stderr, "[stache: %d] ", cgn->total_stache_usage);
      fprintf(stderr, "[scalls: %d, dcalls:%d] ", cgn->static_calls,
              cgn->dynamic_calls);
      if (cgn->first_local) {
         struct call_graph_local_info *cg_local = cgn->first_local;
         count = 0;
         fprintf(stderr, "\n");
         while (cg_local) {
            fprintf(stderr,
                    "   local: %s\t size: %4d  " "sreps: %4d "
                    "dreps: %4d  reg:%1d arg:%1d vol:%1d\n", cg_local->item,
                    cg_local->size, cg_local->static_reps,
                    cg_local->dynamic_reps, cg_local->reg, cg_local->arg,
                    cg_local->vol);

            cg_local = cg_local->next;
         }

      }
      fprintf(stderr, "calls:\n");
      cgln = cgn->next_child;
      count = 0;
      while (cgln) {
         if (count == 0) {
            fprintf(stderr, "\t");
         }
         count++;
         fprintf(stderr, "%s,%d,%d ", globalname(cgln->the_func->func_id),
                 cgln->static_count, cgln->dynamic_count);
         if (count == cg_linelen) {
            count = 0;
            fprintf(stderr, "\n");
         }
         cgln = cgln->next_list_node;
      }
      if (count != cg_linelen)
         fprintf(stderr, "\n");

      fprintf(stderr, "and function '%s' is called by:",
              globalname(cgn->func_id));

      cgln = cgn->next_parent;
      count = 0;
      fprintf(stderr, "\n");
      while (cgln) {
         if (count == 0) {
            fprintf(stderr, "\t");
         }
         count++;
         fprintf(stderr, "%s,%d,%d ", globalname(cgln->the_func->func_id),
                 cgln->static_count, cgln->dynamic_count);
         if (count == cg_linelen) {
            count = 0;
            fprintf(stderr, "\n");
         }
         cgln = cgln->next_list_node;
      }
      if (count != cg_linelen)
         fprintf(stderr, "\n");
      fprintf(stderr, "\n");

      cgn = cgn->next_func;
   }
   fprintf(stderr, "*******************************************\n");
}


void annotate_call_graph_with_runtime_info()
{
   char  str1[100], str2[100];
   FILE *fin;
   char *func_id;
   char *func_id2;
   struct call_graph_node *func_node, *func_node2, *cgn;
   struct call_graph_list_node *list_node;

   /* clear the static guess on the dynamic count */
   for (cgn = the_call_graph; cgn; cgn = cgn->next_func)
      cgn->dynamic_calls = 0;

   fin = fopen(profileName, "r");

   if (!fin) {
      fprintf(stderr, "Cannot open profile file %s, not profiling\n",
              profileName);
      return;
   }

   while (1) {
      fscanf(fin, "%s%s", str1, str2);
      if (strcmp(str1, "End") == 0 && strcmp(str2, "File") == 0) {
         return;
      }
      else if (strcmp(str1, "function") == 0) {
         func_id = function_name_to_id(str2);
         if (func_id)
            func_node = find_call_graph_node(func_id);
         else
            func_node = 0;
      }
      else if (strcmp(str1, "end") == 0) {
         func_id = 0;
      }
      else if (func_node != 0) {
         func_id2 = function_name_to_id(str1);
         if (func_id2 == 0)
            continue;
         func_node2 = find_call_graph_node(func_id2);
         if (func_node2 == 0)
            continue;
         add_call_to_call_graph_node(func_node, func_node2, 1);
         if (list_node = is_directly_called(func_node, func_node2))
            list_node->dynamic_count += atoi(str2);
         if (list_node = is_direct_parent(func_node2, func_node))
            list_node->dynamic_count += atoi(str2);
      }
   }
   fclose(fin);

}

void calculate_node_estimates(struct call_graph_node *cgn)
{
   struct call_graph_list_node *cgln;

   if (!cgn || cgn->seen)
      return;
   cgn->seen = 1;
   cgn->static_calls = 1;
   cgn->dynamic_calls = 1;

   for (cgln = cgn->next_child; cgln; cgln = cgln->next_list_node) {
      if (!cgln->the_func->seen)
         calculate_node_estimates(cgln->the_func);

      /* if we have profile info, just use the count given */
      /* otherwise calcualte it the same way to we calculate static */
      /* that way we have our best-guess in the dynamic_count */
      if (swpr)
         cgn->dynamic_calls += cgln->dynamic_count;
      else
         cgn->dynamic_calls +=
          (cgln->the_func->static_calls * cgln->static_count);

      cgn->static_calls +=
       (cgln->the_func->static_calls * cgln->static_count);
   }
   return;
}

void fix_local_rep_count()
{
   struct call_graph_node *cgn;

   for (cgn = the_call_graph; cgn; cgn = cgn->next_func) {
      struct call_graph_local_info *locals;
      for (locals = cgn->first_local; locals; locals = locals->next) {
         if (locals->static_reps == 0)
            locals->static_reps = 1;
         if (locals->dynamic_reps == 0)
            locals->dynamic_reps = 1;
      }
      for (locals = cgn->first_local; locals; locals = locals->next) {
         locals->static_reps *= cgn->static_calls;
         locals->dynamic_reps *= cgn->dynamic_calls;
      }
      for (locals = cgn->first_local; locals; locals = locals->next) {
         if (locals->static_reps == 0)
            locals->static_reps = 1;
         if (locals->dynamic_reps == 0)
            locals->dynamic_reps = 1;
      }
   }
}

void determine_callgraph_recursiveness()
{
   struct call_graph_node *cgn = 0;

   for (cgn = the_call_graph; cgn; cgn = cgn->next_func) {
      cgn->is_indirectly_recursive = is_indirectly_recursive(cgn);
      cgn->is_directly_recursive = is_directly_recursive(cgn);
   }

}

static void add_call_by_ptr_arcs()
{
   struct call_graph_node *cgn = 0;
   struct call_graph_node *cgn2 = 0;

   for (cgn = the_call_graph; cgn; cgn = cgn->next_func) {
      if (cgn->makes_call_by_pointer) {
         for (cgn2 = the_call_graph; cgn2; cgn2 = cgn2->next_func) {
            add_call_to_call_graph_node(cgn, cgn2, 1);
         }
      }
   }
}

int make_guestimate(struct call_graph_node *cgn)
{
#if PK  
   struct call_graph_list_node *cgln;

   int   change = FALSE;
   int   max_children = 0;
   int   cur_child = 0;
   int   my_desired = 0;

   /* bail out if we've already seen this */
   if (cgn->seen)
      return 0;

   if (cgn->library_func) {
      return 0;
   }

   if (!cgn->has_data) {
      cgn->total_stache_usage = stache_size;
      return 0;
   }

   /* mark it as seen right away so we don't get into a loop */
   cgn->seen = TRUE;

   for (cgln = cgn->next_child; cgln; cgln = cgln->next_list_node) {
      change |= make_guestimate(cgln->the_func);

      cur_child = cgln->the_func->total_stache_usage;

      if (cur_child > max_children)
         max_children = cur_child;
   }
   if (!cgn->is_indirectly_recursive) {
      my_desired = max_children + cgn->size_of_locals;
      if (my_desired > stache_size)
         my_desired = stache_size;
   }
   else
      my_desired = 0;

   if (my_desired != cgn->total_stache_usage)
      change = TRUE;
   cgn->total_stache_usage = my_desired;

   return change;
#endif   
}

void number_nodes(struct call_graph_node *cgn, int *num)
{
   struct call_graph_list_node *cgln;

   if (cgn->seen)
      return;
   cgn->dfs_num = (*num)++;
   cgn->seen = TRUE;
   for (cgln = cgn->next_child; cgln; cgln = cgln->next_list_node)
      number_nodes(cgln->the_func, num);
   return;
}

#if 0
void mark_recursive_edges(struct call_graph_node *cgn)
{
   struct call_graph_list_node *cgln;

   if (cgn->seen)           
      return;
	cgn->seen = TRUE;

   for (cgln=cgn->next_child; cgln; cgln=cgln->next_list_node) {
		if (cgln->the_func->dfs_num <= cgn->dfs_num)
			cgln->is_recursive_edge=TRUE;
       mark_recursive_edges(cgln->the_func);
	}
   return;
}
#endif

void guestimate_stache_usage()
{
#if PK  
   int   num = 0;
   struct call_graph_node *cgn;
   int   change;

   do {
      change = FALSE;
      clear_seen();
      for (cgn = the_call_graph; cgn; cgn = cgn->next_func)
         change = make_guestimate(cgn);
   } while (change);

   /* Now that we know how much each node's children will take
    * and how much we have for ourselves and our children,
    * calcuate how much we'll have all to ourselves *mmmm*.
    */
   for (cgn = the_call_graph; cgn; cgn = cgn->next_func) {
      cgn->stache_usage =
       cgn->total_stache_usage - cgn->max_child_stache_usage;
   }
#endif   
}

void clear_seen()
{
   struct call_graph_node *cgn;

   for (cgn = the_call_graph; cgn; cgn = cgn->next_func)
      cgn->seen = 0;
}

static void read_library_file()
{
   FILE *fin;
   char  funcName[1000];
   char *id;
   int   stacheSize;

   if (!swpl)
      return;
   fin = fopen(libraryFilename, "r");
   if (!fin) {
      fprintf(stderr, "\n\nCannot open: %s\ndisabling +L ...\n\n\n",
              libraryFilename);
      return;
   }

   while (!feof(fin)) {

      fscanf(fin, "%s%d", funcName, &stacheSize);

      if (debug_opt("dump_library_input")) {
         fprintf(stderr, "Found '%s' with size %d...", funcName, stacheSize);
      }

      id = function_name_to_id(funcName);
      if (id) {
         struct call_graph_node *cgn;
         cgn = add_func_to_call_graph(id);
         cgn->library_func = 1;
         cgn->total_stache_usage = stacheSize;
         if (debug_opt("dump_library_input"))
            fprintf(stderr, "exists in graph");
      }
      if (debug_opt("dump_library_input"))
         fprintf(stderr, " ...Done\n");
   }
   fclose(fin);
}



/*
 * mini_optimize - optimize the current function
 */
void mini_optimize()
{
   register int i;
   struct list *ptr;
   int   change, newallocations, csalloc, r_b_c_count, r_u_j_count,
    r_u_c_count;
   int   r_j_count, m_j_count, i_s_count, l_v_a_count, g_a_count, m_c_s_count;
   int   c_s_e_count, r_d_v_count, c_m_count, s_s_r_count, r_u_b_count;
   int   m_b_b_count, f_d_f_count;
   char *savefn = NULL; /* Used to see if we are still in same function
                         * as a previous iteration, to avoid repeating
                         * global profiling for the same function.
                         */
   struct loopl *loop;
   extern struct loopl *loops;

#ifndef WIDES
   int   d_f_s_count, b_d_t_count;
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_mini_optimize", 1))
      dump_function_i("before mini_optimize()", 1,
                      SHOW_FREQ | SHOW_LABELS | SHOW_STATUS | SHOW_UP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);

   /*
    * Set up invocation counters for some of the more prolific optimization
    * phases.
    */
   r_b_c_count = 1;
   r_u_b_count = 1;
   r_u_j_count = 1;
   r_u_c_count = 1;
   r_j_count = 1;
   m_j_count = 1;
   m_b_b_count = 1;
   i_s_count = 1;
   d_f_s_count = 1;
   f_d_f_count = 1;
   l_v_a_count = 1;
   b_d_t_count = 1;
   g_a_count = 1;
   m_c_s_count = 1;
   c_s_e_count = 1;
   r_d_v_count = 1;
   c_m_count = 1;
   s_s_r_count = 1;

#ifdef VET
   if (fn) {
      moreopts = TRUE;
   }
#endif

   /* Look through the RTL's and mark each callsite with the func called 
    * after we optimize we'll build a call-graph using the frequency estimate 
    * of the basic-block of the call-site.
    */
   mark_callsites();

   /*
    * Perform the following optimization phases only if the user has
    * specified that control flow optimizations should be performed.
    */
   if (swl) {

      /*
       * Locate and remove branch chains.  These are jumps that transfer
       * control directly to another jump.  At this time, we do not care if the
       * phase made any changes to the control flow graph.
       */

      (void) remv_branch_chains(r_b_c_count++);

      /*
       * Remove any useless basic blocks from the function.  These are empty
       * basic blocks that do not provide any useful control flow function.
       * Ignore the change status at this time.
       */

      (void) remv_useless_basic_blocks(r_u_b_count++);

      /*
       * Locate and remove useless jumps.  These are jumps that transfer
       * control to the basic block that follows them.  We do not care if the
       * phase makes any changes to the control flow graph.
       */
      (void) remv_useless_jumps(r_u_j_count++);

      /*
       * Mark basic blocks that are used and dispose of those that are not.
       * Since these basic blocks could never be executed, removing them
       * reduces the size of the function and does not affect the semantics of
       * the code.  At this point, we do not care if the phase makes any
       * changes to the control flow graph.
       */
      (void) remv_unreachable_code(r_u_c_count++);

      /*
       * Eliminate jump sequences that consist of a conditional branch followed
       * by an unconditional jump and then the target of the conditional branch
       * by reversing the condition, changing the target of the conditional
       * branch to match that of the jump and removing the jump.  This sequence
       * has the same semantics, but removes the need for the jump.  Note that
       * we are not concerned with whether or not the control flow of the
       * function is altered by this phase.
       */
      (void) reverse_jumps(r_j_count++);

      /*
       * Perform simple jump minimization by moving the location of some basic
       * blocks. We do not need to know if the phase makes any changes in the
       * function at this point.
       */
      (void) minimize_jumps(m_j_count++);

      /*
       * Merge basic blocks where possible.  This will not only reduce the
       * amount of memory that we will need to optimize this function, it will
       * also make it possible to make some of the combiner links that would
       * have been global local.
       */
      (void) merge_basic_blocks(m_b_b_count++);
   }

   /*
    * Allocate vector positions for pseudo registers, and then go back and fix
    * the dead list on each RTL to use a register vector.  While we are at it,
    * we create a register vector named "pseudos" which contains all pseudo
    * registers (and only pseudo registers).  Note that we do this after having
    * removed unreachable basic blocks (if the user requested control flow
    * optimizations), so that we do not allocate bit positions for pseudo
    * registers that were used only in those blocks.  We must do this before we
    * attempt to perform any data flow analysis or instruction selection
    * phases.
    */
   i = regallocptr;
   allocate_reg_bit_positions();
   rveclen = regallocptr >> intscale;
   if (regallocptr & bitsperint)
      rveclen++;
   for (pseudos = rinit(); i < regallocptr;
        rins((unsigned int) i++, &pseudos)) ;
   vectorize_deads(1);

   /*
    * Perform an initial instruction selection pass through the code to bring
    * it to a more compact, less naive state.  This should, depending on the
    * machine, reduce the number of RTLs we have to work with in half (more on
    * CISC machines).  To do this, we first set up local combiner links between
    * promising RTLs and then invoke the instruction selection phase.  Note
    * that we invoke the instruction selection phase even if the the local
    * combiner links phase reports that no links were introduced.  This is
    * because there are certain special cases where, depending on the target
    * machine, the instruction selection phase will consider some physically
    * adjacent instructions as possible candidates for merging, and we would
    * want to make sure that such cases have been considered at least once.
    */
   (void) set_local_links(1);
#ifdef INTERACT   
   (void) instruction_selection(i_s_count++, FALSE);
#else
   (void) instruction_selection(i_s_count++);
#endif
   
   /*
    * fix-up control-flow/boolean expression from suif
    */
#ifdef FIX
   if (swn)
      fix_control_flow(1);
#endif

   /*
    * If the user specified that evaluation order determination should be
    * performed, then do it at this time (when only local combiner links
    * exist).
    */
   if (swv)
      evaluation_order_determination(1);

   /*
    * Perform a depth-first-search of the control-flow graph.  This
    * information, although not absolutely essential to the proper functioning
    * of the iterative data-flow algorithms, does reduce the number of
    * iterations they have to make significantly.
    */
   depth_first_search(d_f_s_count++);

   /*
    * If the global flag is set, then perform data-flow analysis and set up
    * global combiner links across basic blocks.  Since we currently don't
    * expect full side effect information to be available for each RTL, we
    * abstain from seting global links to hardware registers.
    */
   if (swg) {

      /*
       * Determine the dominance frontier for each basic block in the function
       * so that we can build the static single assignment form.
       */
      build_dominator_tree(b_d_t_count++);
      find_dominance_frontiers(f_d_f_count++);

      /*
       * Perform live variable analysis, wich is also required to properly
       * generate the static single assignment form and to set global combiner
       * links.
       */
      live_variable_analysis(l_v_a_count++, FALSE, FALSE);

      /*
       * Now that we have calculated the dominance frontiers and the live
       * variable information, we can build the static single assignment form
       * of the function, which is required to set combiner links across basic
       * blocks.
       */
      build_minimal_ssa_form(1);

      /*
       * Set up any safe combiner links across basic blocks and, if any new
       * combiner links are created, invoke the instruction selection phase so
       * that it will attempt to combine the instructions containing these new
       * links.
       */

      if (set_global_links(1)) {
#ifdef INTERACT
	if (instruction_selection(i_s_count++, FALSE))
#else
	if (instruction_selection(i_s_count++))
#endif	  
            live_variable_analysis(l_v_a_count++, FALSE, FALSE);
      }
   }

   /*
    * Update the live variable information for the function, since the local
    * register assigner requires this information in order to assign registers
    * correctly.
    */
   else
      live_variable_analysis(l_v_a_count++, FALSE, FALSE);

   /*
    * Perform local register assignment (that is, assign an appropriate
    * hardware register for every pseudo register).  After that is over, Note
    * that the local register allocator will indicate that the function has
    * been changed only if it is forced to add code to the function (to either
    * spill temporaries or fix up assignment conflicts between basic blocks).
    */
   change = local_register_assignment(1);

   /*
    * After local register allocation is performed, no pseudo registers remain
    * in the function, so we no longer have any use for a register vector that
    * contains all of the pseudo registers.
    */
   rfree(pseudos);

   /*
    * Perform another instruction selection pass if any of the previous phases
    * since the last pass made changes to the code.
    */
   if (change)
#ifdef INTERACT
     (void) instruction_selection(i_s_count++, FALSE);
#else
   (void) instruction_selection(i_s_count++);
#endif

   /*
    * Determine if the user has requested some form of global register
    * allocation.
    */
   if (swr || swo) {
      /*
       * Prior to locating the loops in the current function, we must obtain
       * the dominator tree for the function.
       */
      depth_first_search(d_f_s_count++);
      build_dominator_tree(b_d_t_count++);

      /*
       * Perform control flow analysis on the function to locate loops. Also,
       * perform loop-dependent control flow optimizations if the user has
       * requested that control flow optimizations be applied.
       */
      if (find_loops(1)) {
         if (swl) {
            (void) remv_forward_links(1);
            if (minimize_loop_jumps(1)) {
               (void) merge_basic_blocks(m_b_b_count++);
               clear_loops();
               depth_first_search(d_f_s_count++);
               build_dominator_tree(b_d_t_count++);
               (void) find_loops(2);
            }
         }

         /*
          * Order the loop information structures (from outer to inner loops)
          * and determine the exit blocks for each loop if we have to use such
          * information.
          */
         if (swm || (swr && hascs)) {
            order_loops();
            find_loop_exits();
         }
      }

      /*
       * Determine the impact of calls on the use of registers that must be
       * saved and restored around function calls if we are using the simple
       * global register allocator.  Also, we allow for the possibility that
       * unreachable code elimination may have removed callersave points from
       * the function.
       */
      if (swr && hascs)
         hascs = count_caller_save();

      /*
       * Set the "onum" field of each RTL and count them so that the length of
       * RTL vectors, used in global register allocation to keep track of where
       * local references were found, can be determined.
       */
      number_rtls();
   }

   change = TRUE;
   do {
      /*
       * If the user has requested one of the global register allocators, then
       * invoke them as long as changes can be made to the function.
       */
      if (swr || swo) {
         do {

            /*
             * Perform the requested method of global register allocation
             */
            if (swr) {
               csalloc = FALSE;
               (void) number_basic_blocks();
               newallocations =
                simple_global_allocation(g_a_count++, &csalloc);
            }
            else {

               /*
                * Perform live-variable analysis (including local references)
                * on the current function and remove dead variables before
                * performing coloring global register allocation.
                */
               live_variable_analysis(l_v_a_count++, TRUE, FALSE);
               newallocations =
                remv_dead_variables(r_d_v_count++, TRUE, TRUE);
               newallocations |= color_global_allocation(g_a_count++);
            }

            /*
             * If we have caller save registers and we have used the simple
             * global register allocator, then we should attempt to move save
             * and restore items out of loops.
             */
            if (newallocations && swr && csalloc)
               move_caller_save(m_c_s_count++);

            /*
             * If we have allocated any local items to registers, then there is
             * a good chance that we have created opportunities to select
             * better instructions.
             */
            if (newallocations) {
               change = TRUE;
#ifdef INTERACT	       
               newallocations = instruction_selection(i_s_count++, FALSE);
#else
	       newallocations =  instruction_selection(i_s_count++);
#endif
               clrsets();
            }

            /*
             * If phase iteration has been disabled, then exit this loop
             */
            if (swd && debug_opt("no_phase_iteration"))
               break;

            /*
             * If we are using the simple global register allocator, the
             * changes that have been made to the function may have released
             * registers that were previously used, so we determine which
             * registers are still in use before we re-invoke the allocator.
             */
            if (newallocations && swr)
               registers_used();

         } while (newallocations);
      }
      else
         newallocations = FALSE;

      /* If we are doing global data profiling, we want to do it before
       * loop-based code motion and strength reduction, as these
       * will move the names of some globals out of loops, making it
       * appear to global profiling that these globals are used less
       * frequently than they really are.  However, we will need to perform
       * some of the loop and induction variable analyses in order to get
       * an accurate determination of the loop iteration count for each
       * loop, which is useful for static profiling.
       */
      if (swz && swpw && swpg && (NULL != fn) && (fn != savefn)) {
         /*
          * Perform common subexpression elimination and remove dead variables
          * at this time, if the user has requested it.
          */
         if (swc) {
            change |= common_subexpression_elimination(c_s_e_count++);
            live_variable_analysis(l_v_a_count++, swo, FALSE);
            change |= remv_dead_variables(r_d_v_count++, swo, TRUE);
         }
         else 
            live_variable_analysis(l_v_a_count++, swo, FALSE);

         /* Here we perform some loop analyses that are normally performed
          * from within code_motion() and loop_strength_reduction().
          * We don't care if the swm flag is set, because we need these
          * for global profiling regardless of whether the user requested
          * code motion.
          */
	 for(loop=loops ; loop ; loop=loop->next)
	   LoopAnalysesForProfiling(loop, 0, 0, 0, 0);

         savefn = fn;
         BuildBlockGlobalProfile(L2_DATA_CACHE_SIZE);
      } /* end if (swz && swpw && (NULL != fn) && (fn != savefn)) */

      while (change) {
         change = FALSE;

         /*
          * Perform common subexpression elimination and remove dead variables
          * at this time, if the user has requested it.
          */
         if (swc) {
            change |= common_subexpression_elimination(c_s_e_count++);
            live_variable_analysis(l_v_a_count++, swo, FALSE);
            change |= remv_dead_variables(r_d_v_count++, swo, TRUE);
         }

         /*
          * Perform code motion, loop strength reduction, recurrence
          * elimination and induction variable elimination if the user
          * requested them.
          */
         if (swm && (swr || swo)) {
            if (!swc)
               live_variable_analysis(l_v_a_count++, FALSE, FALSE);
            change |= code_motion(c_m_count++);
         }

         if (swl)
            change |= remv_useless_jumps(r_u_j_count++);

         /*
          * Perform simple strength reduction if the user requested it
          */
         if (sws) {
            live_variable_analysis(l_v_a_count++, FALSE, FALSE);
            if (remv_dead_variables(r_d_v_count++, swo, TRUE))
               live_variable_analysis(l_v_a_count++, FALSE, FALSE);
            change |= simple_strength_reduction(s_s_r_count++);
         }

         /*
          * Perform another instruction selection pass in case changes have
          * been made to the function.
          */
#ifdef INTERACT	 
         change |= instruction_selection(i_s_count++, FALSE);
#else
	 change |= instruction_selection(i_s_count++);
#endif

         /*
          * If phase iteration has been disabled, then exit this loop
          */
         if (swd && debug_opt("no_phase_iteration"))
            break;

         newallocations |= change;
      }

      /*
       * If phase iteration has been disabled, then exit this loop
       */
      if (swd && debug_opt("no_phase_iteration"))
         break;

      /*
       * If we have made changes to the function and the simple register
       * allocator is to be used next, then determine which registers are still
       * used in the function.
       */
      if (newallocations && swr)
         registers_used();

   } while (newallocations);


   /* Here is the end of any iterative optimizations.  Things will get applied
    * again, but not iteratively */

   /*
    * Attempt to optimize the control flow of the function a second time. It is
    * worth attempting these optimizations again because some of the
    * optimizations that have been performed on this function may have made
    * changes to the control flow of the function that allows for these
    * optimizations to be performed.
    */
   if (swl) {
      (void) remv_branch_chains(r_b_c_count++);
      (void) remv_unreachable_code(r_u_c_count++);
      (void) remv_useless_basic_blocks(r_u_b_count++);
      (void) reverse_jumps(r_j_count++);
      (void) minimize_jumps(m_j_count++);
   }

   live_variable_analysis(0, 1, 0);

#ifdef PARTITIONING
   /* The partitioning needs to gather info about which global variables are
    * used and also which local variables we are compiling.
    */
   if (swpu)
      memorize_variables();
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_mini_optimize", 1))
      dump_function_i("after mini_optimize()", 1,
                      SHOW_FREQ | SHOW_LABELS | SHOW_STATUS | SHOW_UP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);
}

struct callee {
   char *name;
   struct callee *next;
};

struct callee *callees;
int   calls_indirect;

mark_callsites()
{
   struct bblock *cblk;
   struct list *ptr;

   calls_indirect = FALSE;
   callees = NULL;

   for (cblk = top; cblk; cblk = cblk->next) {
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
         if (ptr->type & RTLINE) {
            if (ptr->val && is_call(ptr->val)) {
               char *temp = identify_function_name(ptr);
               if (temp) {
                  struct callee *temp_callee = (struct callee *)
                   CALLOC_PERM(sizeof(struct callee));
                  temp_callee->name = CALLOC_PERM(10);
                  copyloc(temp_callee->name, temp);
                  *skiploc(temp_callee->name) = 0;
                  temp_callee->next = callees;
                  callees = temp_callee;
               }
               else {
                  calls_indirect = TRUE;
                  fprintf(stderr, "%s makes call by pointer\n", fn);
               }
               ptr->variable = 0;
            }
         }
      }
   }
}

struct call_graph_node *add_to_cg()
{
   struct bblock *cblk;
   struct list *ptr;
   struct call_graph_node *curr_node = NULL;
   char  curr_id[MAXSTRING] = { 0, };

   struct callee *temp_callee;
   struct locuse *locals = locs;
   int   local_size = 0;
   int   num_locals = 0;
   int   param_size = 0;
   int   num_params = 0;

   if (!fn)
      return NULL;

   strcpy(curr_id, function_name_to_id(fn));
   curr_node = add_func_to_call_graph(curr_id);
   curr_node->has_data = 1;
   if (swz) {
      curr_node->topCFG = NULL;
      curr_node->CallSiteCount = 0;
      curr_node->InlinedIntoCallers = FALSE;
      curr_node->IrreducibleCFG = FALSE;
      curr_node->InlineEligible = TRUE;
      curr_node->topEstimate = 1.0;
   }

   for (locals = locs; locals; locals = locals->next) {
      if (!locals->arg && !locals->reg && locals->reps) {
         struct call_graph_local_info *cg_local =
          (struct call_graph_local_info *)
          CALLOC_PERM(sizeof(struct call_graph_local_info));
         bzero(cg_local, sizeof(struct call_graph_local_info));

         cg_local->item = (char *) strdup(locals->item);
         copyloc(cg_local->code, locals->code);
         cg_local->size = locals->size;
         cg_local->static_reps = locals->reps;
         cg_local->dynamic_reps = locals->reps;
         cg_local->reg = locals->reg;
         cg_local->arg = locals->arg;
         cg_local->vol = locals->vol;
         cg_local->function_cgn = curr_node;
         cg_local->next = curr_node->first_local;
         curr_node->first_local = cg_local;
         cg_local = 0;
      }
      /* don't count register vars */
      if (locals->reg)
         continue;
      if (!locals->arg) {
         local_size += locals->size;
         num_locals++;
      }
      else {
         param_size += locals->size;
         num_params++;
      }
   }
   if (curr_node) {
      curr_node->size_of_locals = local_size;
      curr_node->num_of_locals = num_locals;
      curr_node->size_of_params = param_size;
      curr_node->num_of_params = num_params;
   }

   for (temp_callee = callees; temp_callee; temp_callee = temp_callee->next) {
      char *new_id = temp_callee->name;
      struct call_graph_node *new_node = 0;

      /* lookup or add the function */
      new_node = add_func_to_call_graph(new_id);

      /* then add the edge */
      add_call_to_call_graph_node(curr_node, new_node, 10);

   }
   if (calls_indirect)
      curr_node->makes_call_by_pointer = TRUE;

   return curr_node;
}

#endif  /* STACHE */
#endif  /* CALLGRAPH */
