/*
 * callgraph.h - prototypes and declarations for structures and
 *    functions associated with building and maintaining a callgraph 
 *    for a set of functions
 *
 * author: Jason D. Hiser
 * Creation: 8-1-2000
 * 
 */
#ifndef CALLGRAPH_H 
#define CALLGRAPH_H 

#ifdef CALLGRAPH
#ifdef STACHE

#include "vpo.h"

extern int maxGlobalNum;

struct call_graph_list_node;
struct LIG_node;
struct  call_graph_node;

/* Struct to hold global profiling info.  Smaller than struct DataProfile to
 * save space. If an edge weight is negative, that implies the edge is a back
 * edge, which saves space by eliminating back edge flags.
 */

struct GlobalDataProfile {
   int LeftBackEdgeFlag;
   int RightBackEdgeFlag;
   struct varl *FirstGlobal;
   struct varl *LastGlobal;
};

/* Struct to hold the loop list information about loops that contain a
 * CFG node.
 */
struct CFGloopl {
   int loopID;  /* Unique number across all loops in the program. */
   int blockCount; /* Number of basic blocks in loop. */
   int iterations; /* Iteration count of loop. */
   struct CFGloopl *next;
};

/* Struct to hold the control-flow-graph nodes, with global data access info,
 * for each basic block in a function in the call graph.  Used for data
 * placement of globals.
 */
struct GlobalPlacementCFGNode {
   struct GlobalPlacementCFGNode *next;  /* Next in ordered list of nodes. */
   struct GlobalDataProfile *prof;       /* Includes list of globals.      */
   struct GlobalPlacementCFGNode *left;  /* Unconditional successor.       */
   struct GlobalPlacementCFGNode *right; /* Conditional successor.         */
   struct CFGloopl *LoopList;            /* List of containing loops.      */
   unsigned int done:1;                  /* Processed this node? (boolean) */
   unsigned int bypassed:1;              /* Bypassed this node? (boolean)  */
   float estimate;                       /* Execution frequency estimate.  */
   int NodeNum;                          /* index #, with 0 for top, etc.  */
};
 
struct call_graph_local_info
{
   /* the 2 or 4 character code used for this variable */
   char *item; 

   /* the 2 or 4 character code used for this variable */
   char code[5];  
   /* how many stache bytes it would need */
   unsigned int size;

   /* how many times it's referenced in it's function */
   unsigned int static_reps;  /* the static guess */
   unsigned int dynamic_reps; /* a profiled guess */

   /* is it assigned to a register? */
   unsigned int reg:1;

   /* is arg */
   unsigned int arg:1;

   /* is it volatile? */
   unsigned int vol:1;


   /* the LIG node associated with the local */
   struct LIG_node* the_LIG_node;

   struct call_graph_node* function_cgn;

   /* the next local in the list */
   struct call_graph_local_info* next;
};


struct  call_graph_node
{
   /* the 2 or 4 character global identifier of a function */
   char func_id[REGSTRING];

   /* the name of the func, so we can print it even if
      the global info is no longer around */
   char* name;

   /* a list of children of this node */
   struct call_graph_list_node* next_child;

   /* a list of parents of this node */
   struct call_graph_list_node* next_parent;

   /* the list of functions in the graph */
   struct call_graph_node* next_func;

   /* Beginning of control-flow-graph nodes that contain
    * data profiling info about globals for data placement.
    */
   struct GlobalPlacementCFGNode *topCFG;

   /* Estimate of how many times this function is called,
    * starting with 1.0 for main() and statically estimated
    * for other functions based on the estimate fields in
    * the CFG nodes that call them.
    */
   float topEstimate;

   /* Count of call sites to other functions; 0 => is_leaf
    * or InlinedIntoCallers.
    */
   int CallSiteCount;

   /* Count of RTL basic blocks in the call graph node. */
   int blockCount;

   /* Has this call graph node's CFG been inlined into all callers yet? */
   unsigned int InlinedIntoCallers:1;

   /* Is the control flow graph for this function irreducible? */
   unsigned int IrreducibleCFG:1;

   /* Is this function small enough to be eligible for inlining? */
   unsigned int InlineEligible:1;

   /* has this node been seen before when traversing the graph */
   unsigned int seen:1;

   /* this function makes a call by pointer and can potentially
      call any function that is_addressed */
   unsigned int makes_call_by_pointer:1;

   /* has_data gets set when an 'f' line is seen in the RTL file. This means
    * that we'll be compiling this function during this run of VPO.
    */
   unsigned int has_data:1;

   /* is this a node that we have info about as a library function */
   unsigned int library_func:1;

   /* does this function make calls to other functions? */
   unsigned int is_leaf:1;

   /* a pass over the callgraph is made and these are marked.
      this is done after the entire file is parsed. */
   unsigned int is_directly_recursive:1;
   unsigned int is_indirectly_recursive:1;

   /* the depth-first search number for this node */
   int dfs_num;

   /* the total size of all local variables */
   /* parameters are not counted here */
   unsigned short size_of_locals;

   /* the number of local variables */
   /* parameters are not counted here */
   unsigned short num_of_locals;

   /* the total size of all parameters*/
   /* only declared parameters (as opposed to variadic) are counted */
   unsigned short size_of_params;

   /* the number of parameters */
   /* only declared parameters (as opposed to variadic) are counted */
   unsigned short num_of_params;

   /* how much stache space will our children need? */
   /* this includes grand-children, great grandchildren,
      etc */
   unsigned int max_child_stache_usage;

   /* how much will I use + how much my children will use */
   unsigned int total_stache_usage;

   /* how much will I use */
   unsigned int stache_usage;

   /* information about local variables */
   struct call_graph_local_info* first_local;

   /* sum of all incoming edges to this node */
   unsigned int static_calls;

   /* sum of all incoming edges to this node */
   unsigned int dynamic_calls;

};

struct call_graph_list_node
{
   /* the pointer to the function that's being called*/
   struct call_graph_node* the_func;

   /* the next arc in the list */
   struct call_graph_list_node* next_list_node; 

   /* the static guess at how many times this arc is used */
   int static_count;

   /* the dynamic guess at how many times this arc is used */
   int dynamic_count;

   /* is this edge part of a cycle? */
   int is_recursive_edge:1;
};


struct indirect_functions_t
{
   char* func_id;
   struct indirect_functions* next;
};

extern struct call_graph_node *the_call_graph;

/* Functions to help create a call graph */
struct call_graph_node* add_func_to_call_graph(char* id);
void add_call_to_call_graph_node(struct call_graph_node *current_func, 
                              struct call_graph_node * new_callee, int freq);

/* Search for a node by ID */
struct call_graph_node* find_call_graph_node(char* id);


/* Determine if the call graph has a loop that includes this node. */
int is_indirectly_recursive(struct call_graph_node* );
/* Is the smallest loop of size 1 */
int is_directly_recursive(struct call_graph_node* );


/* Determine if a child is a direct, or indirect predecessory of a node */
struct call_graph_list_node* 
is_directly_called(struct call_graph_node* caller, 
                   struct call_graph_node* callee);
struct call_graph_list_node* 
is_direct_parent(struct call_graph_node* caller, 
                 struct call_graph_node* callee);
int 
is_indirectly_called(struct call_graph_node* caller, 
                     struct call_graph_node* callee);

void determine_callgraph_recursiveness();


/* A top-level function that makes a pass over the input and builds
 * the entire callgraph.  Then it rewinds the input for the optimizer
 * to make a second pass over it, doing full optimizations 
 */
void build_call_graph();

char* function_name_to_id(char* func_name);


#endif  /* STACHE */
#endif  /* CALLGRAPH */
#endif  /* CALLGRAPH_H */

