/*
 * codesize.c - File contains optimizations to reduce code size.
 */

#include <stdio.h>
#include <stdlib.h>

#include "vpo.h"
#include "view.h"


#define HTABSIZE 47
#define MAX_RTLS_FUNC 5000

struct Htable{
  int index;          /* count measuring each distinct RTL in func */
  int length;         /* RTL length */
  char *rtlval;
  struct Htable *next;
};

struct RTLrep{
  char *val;
  struct RTLrep *next;
};

int arrangeRTLs();
int RTLHashvalue(char *str);
int isRTLMatch(struct Htable *hent, char *rtl, int length);
void encodeRTLs();
void encodeBlk(struct bblock *blk);



/*variable to hold the RTL hashtable*/
struct Htable *RTLTable[HTABSIZE];

/*global array sequence of RTLs in the function*/
int RTLArray[MAX_RTLS_FUNC];

/*header to the list of RTLs represented in the format we need*/
struct RTLrep *encRTL, *endRTL;


/*
 * procedural_abs - Main outside interface function for procedural
 * abstraction
 */
int procedural_abs()
{
  int i, RTLcount;
  struct bblock *cblk;
  struct list *rtl;
  struct RTLrep *ptr;
  extern struct bblock *top;
  
  /*arrange the RTLs in this function into an array*/
  RTLcount = arrangeRTLs();

  /*print out the RTL array for this function*/
  for(i=0 ; i<RTLcount ; i++)
    fprintf(stderr, "%d ", RTLArray[i]);
  fprintf(stderr, "\n");

  /*free the allocated memory*/
  dispose_pabs();
}


/*
 * arrangeRTLs - Each RTL in the current function is entered into a
 *               hashtable, such that two identical RTLs map to the same hash
 *               bucket and each distinct RTL gets a distinct index. This
 *               index is then used to form a sequence of RTLs in the
 *               function. Returns the number of RTLs in the function
 */
int arrangeRTLs()
{
  int i, count, index, sum;
  int ai, hi;
  char *rtlval;
  struct bblock *cblk;
  struct Htable *new, *hent, *prev;
  struct RTLrep *ptr;
  extern struct bblock *top;

  /*encode the RTLs in this function in the format we need*/
  encRTL = NULL;
  encodeRTLs();

  /*initialize the hashtable*/
  for(i=0 ; i<HTABSIZE ; i++)
    RTLTable[i] = (struct Htable *)NULL;

  /*place all the RTLs in the hashtable*/
  count = 0;
  ai = hi = 0;
  for(ptr=encRTL ; ptr ; ptr=ptr->next){

    rtlval = ptr->val;
    sum = RTLHashvalue(rtlval);

    /*check if an identical copy of this RTL already exists and has been
      encountered before*/
    index = sum % HTABSIZE;
    prev = NULL;
    for(hent=RTLTable[index] ; hent ; hent=hent->next){
      
      /*perform a comparison of the current RTL with the RTL at this
	hashtable location to check for a match*/
      if(isRTLMatch(hent, rtlval, sum)){
	RTLArray[ai++] = hent->index;
	break;
      }
      
      prev = hent;
    }
    
    /*if this RTL does not match any previously encountered*/
    if(!hent){
      new = (struct Htable*)myalloc_pabs(sizeof(struct Htable));
      new->length = sum;
      new->index = hi++;
      new->rtlval = (char *)myalloc_pabs(strlen(rtlval)+1);
      strcpy(new->rtlval, rtlval);
      new->next = NULL;
      
      if(prev)
	prev->next = new;
      else
	RTLTable[index] = new;
      
      RTLArray[ai++] = new->index;
    }
    
    /*increment count of legal RTLs in function*/
    count++;
  }

  return count;
}


/*
 * RTLHashvalue - find the hashvalue of this RTL. Currently this is just a
 *                summation of all chars in the rtl->val
 */
int RTLHashvalue(char *str)
{
  char *c;
  int count;

  count = 0;
  for(c=str ; c && *c ; c++){
    if(*c > 0)
      count += *c;
    else
      count -= *c;
  }

  return count;
}
    

/*
 * isRTLMatch - Check if the current RTL matches the RTL in the structure
 */
int isRTLMatch(struct Htable *hent, char *rtlval, int length)
{
  char *p1, *p2;

  if(hent->length != length)
    return FALSE;

  for(p1=hent->rtlval, p2=rtlval ; p1 && p2 && *p1 && *p2 ; p1++, p2++){
    if(*p1 != *p2)
      return FALSE;
  }

  return TRUE;
}


/*
 * encodeRTLs - RTLs in the function are encoded to make it easier to
 *              identify RTLs that may not be textually identical, but might
 *              be isomorphic within a renaming. All register definitions are
 *              replaced by a '*'. Let the current RTL be I, and let there be
 *              a previous def/use of the same register in Q. Then the
 *              register in I is replaced by a triple <O, T, R>, where O is
 *              the distance between I and Q, T is def/use, and R is the
 *              index of the register within Q. All this is within the same
 *              basic block
 */
void encodeRTLs()
{
  struct bblock *cblk;
  struct list *rtl;
  extern struct bblock *top;
  
  for(cblk=top ; cblk ; cblk=cblk->down)
    encodeBlk(cblk);
}


/*
 * encodeBlk - a recursive function first called from encodeRTLs to encode
 *             the RTLs in a function
 */
void encodeBlk(struct bblock *blk)
{
  int i, ind, rnum, regcnt;
  char *c, *t, str[MAXSTRING], buf[MAXSTRING];
  struct RTLrep *new;
  struct list *rtl;
  struct REGinfo{
    int defuse, offset, index;
  }rinfo[MAXREGS];

  /*initialize the register structure*/
  for(i=0 ; i<MAXREGS ; i++)
    rinfo[i].offset = -1;
  
  /*for each RTL in this block*/
  for(rtl=blk->rtls ; rtl ; rtl=rtl->next){

    if((c=cnts(rtl))){

      ind = 0;
      regcnt = 0;
      while(c && *c){

	/*if the current character in the RTL is a register*/
	if(isRegister(c)){

	  /*if the register is definied then replace it by a '*'*/
	  if((t=skiploc(c)) && *t=='='){

	    str[ind++] = '*';

	    /*update the register record*/
	    rnum = getRegNum(c);
	    rinfo[rnum].offset = 0;
	    rinfo[rnum].defuse = 1;
	    rinfo[rnum].index = 0;
	  }
	  
	  /*if the register is used then look for previous def/use of the
	    register in this block, else replace by '*'*/
	  else{
	    rnum = getRegNum(c);
	    if(rinfo[rnum].offset != -1){

	      sprintf(buf, "<%d,", rinfo[rnum].offset);
	      if(rinfo[rnum].defuse == 1)
		strcat(buf, "d,");
	      else
		strcat(buf, "u,");
	      sprintf(buf, "%s%d>", buf, rinfo[rnum].index);

	      /*copy this to the encoded RTL*/
	      for(t=buf ; *t ; t++)
		str[ind++] = *t;
	    }
	    else{

	      str[ind++] = '*';
	    }

	    /*update the register record for this use*/
	    regcnt++;
	    rinfo[rnum].offset = 0;
	    rinfo[rnum].defuse = 0;
	    rinfo[rnum].index = regcnt;
	  }
	}
	else{

	  /*just add the character as it is to the encoded RTL*/
	  str[ind++] = *c;
	}

	c = skiploc(c);
      }

      /*add this encoded RTL to the list of encoded RTLs*/
      str[ind] = '\0';
      new = (struct RTLrep*)myalloc_pabs(sizeof(struct RTLrep));
      new->val = (char *)myalloc_pabs(strlen(str)+1);
      strcpy(new->val, str);
      new->next = NULL;

      if(encRTL)
	endRTL->next = new;
      else
	encRTL = new;
      endRTL = new;

      /*increment offset counters for all encountered registers*/
      for(i=0 ; i<MAXREGS ; i++){

	if(rinfo[i].offset != -1)
	  rinfo[i].offset++;
      }
    }
  }
}
