/**********************************************************************
Functions to do hand-specified transformations.
hand_trans(msg): do the hand-specified transformation according to the
specified msg.
  **********************************************************************/
#ifdef INTERACT

#include <stdio.h>
#include <stdlib.h>

#include "vpo.h"
#include "hand_trans.h"
#include "undo.h"
#include "misc.h"
#include "vpo2.h"
#include "view.h"
#include "encode_rtl.h"
#include "dataflow.h"

#define ERR_RET(MSG) {send_error(MSG); return;}

extern int hand_measure;

/*
 * get_next_component - get the next component of *src, and copy it to dst
 *                      At the same time, modify *src to point to the
 *                      current position.
 */
int get_next_component(char **src, char *dst)
{
   int i;
   char *srcptr, *dstptr;
   for (i = 0, srcptr = *src, dstptr = dst; *srcptr; i++, srcptr++, dstptr++) {
      if (*srcptr == ' ')
         break;
      *dstptr = *srcptr;
   }
   *(dstptr++) = '\0';

   if (*srcptr) {
      *src = srcptr +1;
      return i;
   }
   else {
      *src = NULL;
      return 0;
   }
}

/*
 * get_real_string - get the next component of *src, and copy it to dst
 *                   At the same time, modify *src to point to the current
 *                   position.  Difference from get_next_componet: stop
 *                   when met with '\0'. */
int get_real_string(char **src, char *dst)
{
   int len;
   strcpy(dst, *src);
   len = strlen(*src) +1;
   *src += len;
   return len;
}

/*
 * get_assem - used for debugging purposes, better use get_real_string
 *             Here the assembly is ended with ;.
 */
int get_assem(char **src, char *dst)
{
   int len;
   int i;
   char *srcptr, *dstptr;
   for (i = 0, srcptr = *src, dstptr = dst; *srcptr; i++, srcptr++, dstptr++) {
      if (*srcptr == END_ASSEM)
         break;
      *dstptr = *srcptr;
   }
   *(dstptr++) = '\0';

   /*skip the ; and a blank space*/
   if (*srcptr) {
      *src = skiploc(srcptr);
      return i;
   }
   else {
      *src = NULL;
      return 0;
   }
}

#define MAX_ITEM_NUM 10
#define MAX_ITEM_LEN 100


/*
 * get_hand_msg_string - get the hand specified message string from
 *                       options.txt.  Stop when met with -1.
 */
void get_hand_msg_string(char **src, char *hand_msg)
{
   int i;
   char tmpstr[20];

   hand_msg[0] = '\0';
   for ( ; ;) {
      get_string(src, tmpstr);
      strcat(hand_msg, tmpstr);
      strcat(hand_msg, " ");
      if (*tmpstr == '-' && *(tmpstr+1) == '1') {
         *src -= strlen(tmpstr);
         break;
      }
   }
}

/*
 * get_hand_msg_file - get the hand specified message string from options.txt
 */
void get_hand_msg_file(FILE *fopt, char *hand_msg)
{
   int i;
   char tmpstr[20];

   for ( ; ;) {
      fscanf(fopt, "%s", tmpstr);
      strcat(hand_msg, tmpstr);
      strcat(hand_msg, " ");
      if (*tmpstr == ENDTRANS)
         break;
   }
}

/*
 * verify_hand_trans - verify hand transform messages
 */
void verify_hand_trans(char *hand_msg)
{
  int i, num_msgs, status;
  char ret_msg[BIGSTRING], s_msg[10];
  struct message messages[MAX_MSG_NUM];

  /*parse the messages */
  num_msgs = parse_messages(hand_msg, messages);

  /*verify that the change associated with this message is valid. Does not
    actually perform the change*/
  status = (verify_hand_msg(&messages[0], ret_msg));
  sprintf(s_msg, "%d", status);
  write_line(s_msg);
  write_line(ret_msg);
}

/*
 * hand_trans - do the hand specified transformation
 */
void hand_trans(char *hand_msg)
{
   int i, num_msgs;
   struct message messages[MAX_MSG_NUM];

   /* parse the messages */
   num_msgs = parse_messages(hand_msg, messages);

   /* execute the messages */
   for (i=0; i<num_msgs; i++)
      exec_msg(&messages[i]);
}

/*
 * is_digits - Is the string a number?
 */
int is_digits(char *str)
{
   char *strptr;
   for (strptr = str; *strptr; strptr++) {
      if (!isdigit(*strptr))
         return 0;
   }
   return 1;
}

/*
 * parse_label_msg - If the message begins with LABELBB
 */
void parse_label_msg(char **msg, struct message *message_ptr)
{
   int n, item_num;
   char items[2][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   get_next_component(&strptr, items[0]);
   get_next_component(&strptr, items[1]);

   if (!is_digits(items[0]))
      ERR_RET("parse_label_msg(): The id must be digits");
   message_ptr->msg_type = LABELBB;
   sscanf(items[0], "%d", &message_ptr->block_id);
   message_ptr->label = copy_string(items[1]);

   *msg =strptr;
}

/*
 * parse_newbb_msg - If the message begins with NEWBB
 */
void parse_newbb_msg(char **msg, struct message *message_ptr)
{
   int n, item_num;
   char items[2][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   get_next_component(&strptr, items[0]);

   if (!is_digits(items[0]))
      ERR_RET("parse_newbb_msg(): The id must be digits");
   message_ptr->msg_type = NEWBB;
   sscanf(items[0], "%d", &message_ptr->block_id);

   *msg = strptr;
}

/*
 * parse_delbb_msg - If the message begins with DELBB
 */
void parse_delbb_msg(char **msg, struct message *message_ptr)
{
   int n, item_num;
   char items[2][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   get_next_component(&strptr, items[0]);

   if (!is_digits(items[0]))
      ERR_RET("parse_delbb_msg(): The id must be digits\n");
   message_ptr->msg_type = DELBB;
   sscanf(items[0], "%d", &message_ptr->block_id);

   *msg = strptr;
}

/*
 * parse_modbb_ptrs - If the message begins with MODBB
 */
void parse_modbb_ptrs(char **msg, struct message *message_ptr, char type)
{
   int n, item_num;
   char items[2][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   get_next_component(&strptr, items[0]);
   get_next_component(&strptr, items[1]);


   if (!is_digits(items[0]) || !is_digits(items[1]))
      ERR_RET("parse_modbb_msg(): The id must be digits");
   message_ptr->msg_type = type;
   sscanf(items[0], "%d", &message_ptr->block_id);
   sscanf(items[1], "%d", &message_ptr->new_ptr_id);

   *msg = strptr;
}

/*
 * parse_delrtl_msg - If the message begins with DELRTL
 */
void parse_delrtl_msg(char **msg, struct message *message_ptr)
{
   int n, item_num;
   char items[2][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   get_next_component(&strptr, items[0]);
   get_next_component(&strptr, items[1]);

   if (!is_digits(items[0]) || !is_digits(items[1]))
      ERR_RET("parse_delrtl_msg(): The id must be digits");

   message_ptr->msg_type = DELRTL;
   sscanf(items[0], "%d", &message_ptr->rtl_id);
   sscanf(items[1], "%d", &message_ptr->block_id);

   *msg = strptr;
}

/*
 * parse_movertl_msg - If the message begins with MOVERTLu
 */
void parse_movertl_msg(char **msg, struct message *message_ptr)
{
   int i, item_num;
   char items[4][MAX_ITEM_LEN], *strptr;
   strptr = *msg;
   
   for (i =0; i<4; i++)
      get_next_component(&strptr, items[i]);

   if (!is_digits(items[0]) || !is_digits(items[1]) ||
       !is_digits(items[2]) || !is_digits(items[3]))
      ERR_RET("parse_movertl_msg(): The id must be digits");

   message_ptr->msg_type = MOVERTL;
   sscanf(items[0], "%d", &message_ptr->rtl_id);
   sscanf(items[1], "%d", &message_ptr->block_id);
   sscanf(items[2], "%d", &message_ptr->before_id);
   sscanf(items[3], "%d", &message_ptr->new_block_id);

   *msg = strptr;
}

/*
 * parse_newrtl_msg - If the message begins with NEWRTL
 */
void parse_newrtl_msg(char **msg, struct message *message_ptr)
{
   int i, item_num;
   char items[5][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   for (i =0; i<5; i++)
      get_next_component(&strptr, items[i]);

   if (!is_digits(items[0]) || !is_digits(items[3]) || !is_digits(items[4]))
      ERR_RET("parse_newrtl_msg(): The id must be digits\n");

   message_ptr->msg_type = NEWRTL;
   sscanf(items[0], "%d", &message_ptr->rtl_id);
   sscanf(items[1], "%d", &message_ptr->rtl_type);
   message_ptr->rtl_val = copy_string((char *)items[2]);
   sscanf(items[3], "%d", &message_ptr->block_id);
   sscanf(items[4], "%d", &message_ptr->before_id);

   *msg = strptr;
}

/*
 * parse_modrtl_msg - If the message begins with MODRTL
 */
void parse_modrtl_msg(char **msg, struct message *message_ptr)
{
   int i, item_num;
   char items[4][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   for (i =0; i<4; i++)
      get_next_component(&strptr, items[i]);

   if (!is_digits(items[0]) || !is_digits(items[3]))
      ERR_RET("parse_modrtl_msg(): The id must be digits");

   message_ptr->msg_type = MOD_RTL;
   sscanf(items[0], "%d", &message_ptr->rtl_id);
   sscanf(items[1], "%d", &message_ptr->rtl_type);
   message_ptr->rtl_val = copy_string(items[2]);
   sscanf(items[3], "%d", &message_ptr->block_id);

   *msg = strptr;
}

/*
 * parse_modassem_msg - If the message begins with MOD_ASSEM
 */
void parse_modassem_msg(char **msg, struct message *message_ptr)
{
   int i, item_num;
   char items[4][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   
   /* The command format is:
         MOD_ASSEM rtl_id rtl_type string_assem\0 blk_id. */
   get_next_component(&strptr, items[0]);
   get_next_component(&strptr, items[1]);

   /* get the assembly value which is ended by ;
      if ; is not ok, we can use \0, and get_assem should be
      replaced by get_real_string */
   get_assem(&strptr, items[2]);

   get_next_component(&strptr, items[3]);   
   if (!is_digits(items[0]) || !is_digits(items[3]))
      ERR_RET("parse_modassem_msg(): The id must be digits");

   message_ptr->msg_type = MOD_ASSEM;
   sscanf(items[0], "%d", &message_ptr->rtl_id);
   sscanf(items[1], "%d", &message_ptr->rtl_type);
   message_ptr->rtl_val = copy_string(items[2]);
   sscanf(items[3], "%d", &message_ptr->block_id);

   *msg = strptr;
}

/*
 * parse_newassem_msg - If the message begins with NEW_ASSEM
 */
void parse_newassem_msg(char **msg, struct message *message_ptr)
{
   int i, item_num;
   char items[5][MAX_ITEM_LEN], *strptr;
   strptr = *msg;

   /* The command format is:
         NEW_ASSEM rtl_id rtl_type assem_val blk_id rtl_id */

   /* get rtl_id */
   get_next_component(&strptr, items[0]);

   /* get rtl_type */
   get_next_component(&strptr, items[1]);

   /* get the assembly value which is ended by ;
      if ; is not ok, we can use \0, and get_assem should be
      replaced by get_real_string */
   get_assem(&strptr, items[2]);

   /* get blk_id */
   get_next_component(&strptr, items[3]);

   /* get rtl_id */
   get_next_component(&strptr, items[4]);   

   if (!is_digits(items[0]) || !is_digits(items[3]) || !is_digits(items[4]))
      ERR_RET("parse_newassem_msg(): The id must be digits\n");

   message_ptr->msg_type = NEW_ASSEM;
   sscanf(items[0], "%d", &message_ptr->rtl_id);
   sscanf(items[1], "%d", &message_ptr->rtl_type);
   message_ptr->rtl_val = copy_string((char *)items[2]);
   sscanf(items[3], "%d", &message_ptr->block_id);
   sscanf(items[4], "%d", &message_ptr->before_id);

   *msg = strptr;
}

/*
 * parse_messages - parse the messages
 */
int parse_messages(char *hand_msg, struct message messages[])
{
   int i,n, num_msgs;
   char *strptr;
   char tmpchar;
   char tmpstr[20];
   
   messages[0].msg_type = hand_msg[0];

   if (hand_msg[0] == BEGINTRANS)
      return 1;
   
   if (hand_msg[0] == ENDTRANS)
      return 1;
   
   strptr = hand_msg;
   
   get_next_component(&strptr, tmpstr);
   if (strlen(tmpstr) != 1)
      ERR_RET("parse_messages(): Parse error: Not begin with a flag\n");
      
   switch (tmpstr[0]) {
      case LABELBB:
         parse_label_msg(&strptr, &messages[0]);
         break;

      case NEWBB:
         parse_newbb_msg(&strptr, &messages[0]);
         break;

      case DELBB:
         parse_delbb_msg(&strptr, &messages[0]);
         break;                  

      case MOD_DOWN:
      case MOD_LEFT:
      case MOD_RIGHT:
         parse_modbb_ptrs(&strptr, &messages[0], tmpstr[0]);
         break;

      case DELRTL:
         parse_delrtl_msg(&strptr, &messages[0]);
         break;

      case MOVERTL:
         parse_movertl_msg(&strptr, &messages[0]);
         break;

      case NEWRTL:
         parse_newrtl_msg(&strptr, &messages[0]);
         break;

      case MOD_RTL:
         parse_modrtl_msg(&strptr, &messages[0]);
         break;

      case MOD_ASSEM:
         parse_modassem_msg(&strptr, &messages[0]);
         break;

      case NEW_ASSEM:
         parse_newassem_msg(&strptr, &messages[0]);
         break;

   #ifdef NOTNOW
      case LIVE_REGS_QUERY:
         parse_liveregsquery_msg(&strptr, &messages[0]);
         break;
   #endif

      default:
         ERR_RET("parse_messages(): No such type or not implemented.\n");

   } /*end switch*/

   return 1;
}

/*
 * exec_msg - Execute the message parsed
 */
void exec_msg(struct message *message_ptr)
{
   int is_rt(char *src);
   int Translate(char *source, char *cexresult);
   
   switch (message_ptr->msg_type) {
      case BEGINTRANS:
         begin_phase(HAND_TRANS_NUMBER);
         begin_trans(VET_OPT_TRANS);
         break;

      case ENDTRANS:
         end_trans();
         end_phase(HAND_TRANS_NUMBER);
	 hand_measure = TRUE;
         break;

      case LABELBB:
         change_bb_label(message_ptr->block_id, message_ptr->label);
         break;

      case NEWBB:
         insert_basic_block(message_ptr->block_id);
         break;

      case DELBB:
         delete_basic_block(message_ptr->block_id);
         break;

      case MOD_DOWN:
      case MOD_LEFT:
      case MOD_RIGHT:      
         change_bb_ptrs(message_ptr->msg_type, message_ptr->block_id,
                               message_ptr->new_ptr_id);
         break;

      case DELRTL:
         {
            struct bblock *cblk;
            struct list *rtlptr;

            /* first, find the current basic block and deleting rtl */
            rtlptr = find_rtl_id(message_ptr->rtl_id,
                                 message_ptr->block_id);
            cblk = find_block_id(message_ptr->block_id);
   
            if (!rtlptr || !cblk)
               ERR_RET("exec_msg(): rtl_id or block_id not found\n");

            /* If a transfer of control being deleted, then update control flow
               to have only a left successor being the next block. */
            update_control_flow_delete(rtlptr, cblk);

            /* end changing the control flow */
            remv_rtl(rtlptr);         
         }
         break;
   
      case MOVERTL:
         {
            /* previous code for move rtl */
            move_rtl_id(message_ptr->rtl_id, message_ptr->block_id,
                        message_ptr->before_id, message_ptr->new_block_id);
         }
         break;

      case NEWRTL:
         {
            register struct list *before, *rtlptr;
            struct bblock *cblk;         
            char code[200];
            CODE_TYPE type_code;
            
            encode_rtl(message_ptr->rtl_val, code);

            /* added to check the syntax of the rtl */
            if (check_syntax(code))
               ERR_RET("exec_msg(): Invalid RTL\n");
   
            /* determine the code type */
            type_code = get_code_type(code);

            /* check the semantics */
            if (check_jump_semantics(code, type_code, message_ptr))
               return;
            
            /* check the control flow before the insert rtl action */
            check_cf(1);
            
            /* insert the new rtl */
            cblk = find_block_id(message_ptr->block_id);
            before = find_rtl_id(message_ptr->before_id, message_ptr->block_id);
            rtlptr = add_rtl_with_type(cblk, before, code, NULL, RTLINE);

            /* Another important thing is that we must update the preds of the
               "exist" bblock */
            /* if it is like PC=L3; */
            if (type_code != NORMAL)
               update_control_flow_insert(cblk,
                                          get_label_down(get_label_str(code)),
                                          type_code);
            
            /* check to see if the control flow is correct */
            check_cf(1);
         }
         break;

      case MOD_RTL:
         {
            char code[200];
            struct list *rtlptr;
            struct bblock *cblk, *label_dst;
            CODE_TYPE type_code;
            
            rtlptr = find_rtl_id(message_ptr->rtl_id,
                                 message_ptr->block_id);
            cblk = find_block_id(message_ptr->block_id);
            if (!rtlptr)
               ERR_RET("exec_msg(): Can't find the rtl_id\n");
            encode_rtl(message_ptr->rtl_val, code);
   
            /* check the syntax of this code */
            if (check_syntax(code))
               ERR_RET("exec_msg(): Invalid RTL\n");
            
            /* determine the code type */
            type_code = get_code_type(code);
   
            /* we do not allow indirect jump */
            if (type_code == INDJMP)
               ERR_RET("exec_msg(): Indirect jump not allowed\n");
             
            /* check the semantics of this code */
            if (type_code != NORMAL) {
               if (rtlptr != find_last_rtl(cblk))
                  ERR_RET("exec_msg(): The transfer of control must be at the end of a block.\n");
            }

            /* check if it has a label, if so, does this label exist? */
            if (type_code == UCJMP || type_code == RJMP) {
               label_dst = get_label_down(get_label_str(code));
               if (!label_dst)
                  ERR_RET("exec_msg(): Branch target does not exist\n");
            }

            /* change the rtl type and value, and send messages to the viewer */
            rtlptr->type = message_ptr->rtl_type;
            rtlptr->val = copy_string(code);
            mod_rtl(rtlptr, code, rtlptr->val);
   
            /* modify the control flow */
            /* assume that this MOD_RTL operation can be divided into delete
               and insert operations */
            
            /* for deleting rtl, the control flow is updated as below: */
            update_control_flow_delete(rtlptr, cblk);

            /* for inserting rtl, the control flow is updated as below: */
            if (type_code != NORMAL)
               update_control_flow_insert(cblk,
                                          get_label_down(get_label_str(code)),
                                          type_code);

            /* check to see if the control flow is correct */
            check_cf(1);
         }
         break;

      case   MOD_ASSEM:
         {
            char code[200];
            struct list *rtlptr;
            struct bblock *cblk, *label_dst;
            CODE_TYPE type_code;
            
            rtlptr = find_rtl_id(message_ptr->rtl_id,
                                 message_ptr->block_id);
            cblk = find_block_id(message_ptr->block_id);
            if (!rtlptr)
               ERR_RET("exec_msg(): Can't find the rtl_id\n");
   
            if (Translate(message_ptr->rtl_val, code))
               ERR_RET("exec_msg(): Translation from assembly to RTL failed\n");
   
   #ifdef DEBUG
            fprintf(stderr, "Encoded RTL:%s\n", code);
   #endif
            
            /* check the syntax of this code */
            if (check_syntax(code))
               ERR_RET("exec_msg(): Invalid RTL\n");
   
            /* determine the code type */
            type_code = get_code_type(code);
   
            /* we do not allow indirect jump */
            if (type_code == INDJMP)
               ERR_RET("exec_msg(): Indirect jump not allowed\n");
             
            /* check the semantics of this code */
            if (type_code != NORMAL) {
               if (rtlptr != find_last_rtl(cblk))
                  ERR_RET("exec_msg(): The transfer of control must be at the end of a block.\n");
            }

            /* check if it has a label, if so, does this label exist? */
            if (type_code == UCJMP || type_code == RJMP) {
               label_dst = get_label_down(get_label_str(code));
               if (!label_dst)
                  ERR_RET("exec_msg(): Branch target does not exist\n");
            }

            /* change the rtl type and value, and send messages to the viewer */
            rtlptr->type = message_ptr->rtl_type;
            rtlptr->val = copy_string(code);
            mod_rtl(rtlptr, code, rtlptr->val);
   
            /* for deleting rtl, the control flow is updated as below: */
            update_control_flow_delete(rtlptr, cblk);

            /* for inserting rtl, the control flow is updated as below: */
            if (type_code != NORMAL)
               update_control_flow_insert(cblk,
                                          get_label_down(get_label_str(code)),
                                          type_code);

            /* check to see if the control flow is correct */
            check_cf(1);
         }
         break;

      case NEW_ASSEM:
         {
            register struct list *before, *rtlptr;
            struct bblock *cblk;         
            char code[200];
            CODE_TYPE type_code;
            
            Translate(message_ptr->rtl_val, code);

            /* added to check the syntax of the rtl */
            if (check_syntax(code))
               ERR_RET("exec_msg(): Invalid RTL\n");
   
            /* determine the code type */
            type_code = get_code_type(code);

            /* check the semantics */
            if (check_jump_semantics(code, type_code, message_ptr))
               return;
            
            /* check the control flow before the insert rtl action */
            check_cf(1);
            
            /* insert the new rtl */
            cblk = find_block_id(message_ptr->block_id);
            before = find_rtl_id(message_ptr->before_id, message_ptr->block_id);
            rtlptr = add_rtl_with_type(cblk, before, code, NULL, RTLINE);

            /* Another important thing is that we must update the preds of the
               "exist" bblock */
            /* if it is like PC=L3; */
            if (type_code != NORMAL)
               update_control_flow_insert(cblk,
                                          get_label_down(get_label_str(code)),
                                          type_code);
            
            /* check to see if the control flow is correct */
            check_cf(1);
         }
         break;
   
   #ifdef NOTNOW
      case LIVE_REGS_QUERY:
         {
            vector regs;
            struct bblock *cblk;
            struct list *rtlptr;
            char regs_string[MAXSTRING];
            extern int reg_assign_done, fill_delay_done;
   
            /* This query should not be made unless register
               assignment has been done. */
            if (!reg_assign_done) {
               send_error("exec_msgs: needs register assignment to be done first.");
               break;
            }
   
            /* perform the live variable analysis analysis first */
            live_variable_analysis(0, FALSE, fill_delay_done);
   
            /* get the live register information */
            if (!(cblk = find_block_id(message_ptr->block_id))) {
               send_error("exec_msgs: block_id not found");
               break;
            }
            if (!(rtlptr = find_rtl_id(message_ptr->rtl_id,
                                       message_ptr->block_id))) {
               send_error("exec_msgs: rtl_id not found");
               break;
            }
            regs = live_regs(cblk, rtlptr);
            vec2string(regs_string, regs);
   
            /* write the live register information to the user interface */
            write_record_id(BEGIN_LIVEREG);   
            write_to_socket(regs_string);
            write_record_id(END_LIVEREG);   
            fprintf(stderr, "the registers are: %s\n", regs_string);
         }
         break;
   #endif
   
      default:
         ERR_RET("exec_msg(): No such type or not implemented.\n");      
   }
      
   return;
}


/*
 * verify_hand_msg - verify that the change specified in the message is legal
 */
int verify_hand_msg(struct message *msg_ptr, char ret_msg[])
{
   int is_rt(char *src);
   int Translate(char *source, char *cexresult);
   struct bblock *cblk;
   struct list *rtlptr;

   switch (msg_ptr->msg_type) {
      case BEGINTRANS:
      case ENDTRANS:
	return TRUE;

      case LABELBB:
	/*this is legal if the basic block exists*/
	if(find_block_id(msg_ptr->block_id) == NULL){
	  strcpy(ret_msg, "Source basic block not found"); 
	  return FALSE;
	}
	strcpy(ret_msg, "Label can be successfully modified");
	return TRUE;

      case NEWBB:
	/*new block can be created if the before_block_id specified in the
	  message is not the first block. before_block_id==0 indicates
	  creation of block at the end of the function, which is legal*/
	cblk = find_block_id(msg_ptr->block_id);
	if((cblk && cblk->up!=NULL) || msg_ptr->block_id==0){
	  strcpy(ret_msg, "New basic block can be created");
	  return TRUE;
	}
	else{
	  strcpy(ret_msg, "New block cannot be created");
	  return FALSE;
	}

      case DELBB:
	/*A block can be deleted if it is present, empty, and is not the
	  entry block to the function*/
	cblk = find_block_id(msg_ptr->block_id);
	if(cblk && cblk->up && !(cblk->rtls)){
	  strcpy(ret_msg, "Block cannot be deleted");
	  return TRUE;
	}
	else{
	  strcpy(ret_msg, "Block cannot be deleted");
	  return FALSE;
	}

      case MOD_DOWN:
      case MOD_LEFT:
      case MOD_RIGHT:
	/*block pointers can be changed if both the source and target blocks
	  are present*/
	if(find_block_id(msg_ptr->block_id)!=NULL &&
	   find_block_id(msg_ptr->new_ptr_id)!=NULL){
	  strcpy(ret_msg, "Block pointer cannot be modified");
	  return TRUE;
	}
	else{
	  strcpy(ret_msg, "BLock pointer cannot be modified");
	  return FALSE;
	}

      case DELRTL:
	/*RTL can be deleted if it is present*/
	if(find_block_id(msg_ptr->block_id)!=NULL &&
	   find_rtl_id(msg_ptr->rtl_id, msg_ptr->block_id)!=NULL){
	  strcpy(ret_msg, "RTL can be deleted");
	  return TRUE;
	}
	else{
	  strcpy(ret_msg, "RTL cannot be deleted");
	  return FALSE;
	}
   
      case MOVERTL:
	/*RTL can be moved if both basic blocks and RTLs exist*/
	if(find_block_id(msg_ptr->block_id) &&
	   find_block_id(msg_ptr->new_block_id) &&
	   find_rtl_id(msg_ptr->block_id, msg_ptr->rtl_id)){

	  /*jumps and returns can only occur at the end of a block*/
	  if(msg_ptr->before_id==0 ||
	     (msg_ptr->before_id!=0 && get_code_type==NORMAL &&
	      find_rtl_id(msg_ptr->new_block_id, msg_ptr->before_id))){
	    strcpy(ret_msg, "RTL can be moved to a new location");
	    return TRUE;
	  }
	}
	else{
	  strcpy(ret_msg, "RTL cannot me moved");
	  return FALSE;
	}

      case NEWRTL:
	{
	  extern int rtl_id;
	  
	  /*check RTL id sent from the viewer, if it is next in the list*/
	  if(msg_ptr->rtl_id != rtl_id+1){
	    strcpy(ret_msg, "Inconsistent RTL id");
	    return FALSE;
	  }
	}
      case MOD_RTL:

	/*check instruction syntax and semantics*/
	if(!check_syn_sem(msg_ptr->rtl_val, msg_ptr->before_id, ret_msg))
	  return FALSE;
	
	/*check if the blocks are present*/
	if(find_block_id(msg_ptr->block_id)==NULL ||
	   (msg_ptr->before_id &&
	    find_rtl_id(msg_ptr->before_id, msg_ptr->block_id)==NULL)){
	  strcpy(ret_msg, "Insert block/rtl not found");
	  return FALSE;
	}
	strcpy(ret_msg, "RTL can be created/modified");
	return TRUE;
	
      case MOD_ASSEM:
      case NEW_ASSEM:
	{
	  char rtl_code[BIGSTRING];
	  
	  /*check if legal translation to RTL is possible*/
	  if (Translate(msg_ptr->rtl_val, rtl_code)){
	    strcpy(ret_msg, "Translation from assembly to RTL failed");
	    return FALSE;
	  }
	  
	  /*check instruction syntax and semantics*/
	  if(!check_syn_sem(rtl_code, msg_ptr->before_id, ret_msg))
	    return FALSE;
	  
	  /*check if the blocks are present*/
	  if(find_block_id(msg_ptr->block_id)==NULL ||
	     (msg_ptr->before_id &&
	      find_rtl_id(msg_ptr->before_id, msg_ptr->block_id)==NULL)){
	    strcpy(ret_msg, "Insert block/rtl not found");
	    return FALSE;
	  }
	  strcpy(ret_msg, "Assembly statement can be added/modified");
	  return TRUE;
	}
	
   #ifdef NOTNOW
      case LIVE_REGS_QUERY:
	return TRUE;
   #endif
   
      default:
         strcpy(ret_msg, "exec_msg(): No such type or not implemented");
	 return FALSE;
   }
      
   return TRUE;
}

/*
 * check_syn_sem - verify RTL syntax and some semantics
 */
int check_syn_sem(char *rtl, int before_id, char msg_ret[])
{
  struct bblock *label_dst;
  int ins_type;
  char en_rtl[BIGSTRING];

  encode_rtl(rtl, en_rtl);
  if(check_syntax(en_rtl)){
    strcpy(msg_ret, "Invalid RTL syntax");
    return FALSE;
  }

  ins_type = get_code_type(en_rtl);
  /* first, check if it is a transfer of control.
     if so, it is must be at the end of the block */
  if (ins_type == INDJMP) {
      strcpy(msg_ret, "Indirect jump not allowed");
      return FALSE;               
   }
      
   if (ins_type != NORMAL) {
     if (before_id) {
         strcpy(msg_ret, "The transfer of control must be at the end of a block");
         return FALSE;
      }
   }

   /* check if it has a label, if so, does this label exist? */
   if (ins_type == UCJMP || ins_type == RJMP) {
      label_dst = get_label_down(get_label_str(en_rtl));
      if (!label_dst) {
         strcpy(msg_ret, "Branch target does not exist");
         return FALSE;
      }
   }

   return TRUE;
}


/*
 * find_block_id - defined functions for executing the changes.
 */
struct bblock *find_block_id(int block_id)
{
   struct bblock *cblk, *nblk;

   if (block_id == 0)
      return (struct bblock *) NULL;
   for(cblk=top, nblk=top ; cblk && nblk ; cblk=cblk->down, nblk=nblk->next){
     if(cblk->block_id == block_id)
       return cblk;
     if(nblk->block_id == block_id)
       return nblk;
   }

   if(cblk){
     for (; cblk ; cblk = cblk->down)
       if (cblk->block_id == block_id)
         return cblk;
   }
   if(nblk){
     for (; nblk ; nblk = nblk->next)
      if (nblk->block_id == block_id)
         return nblk;
   }
   return (struct bblock *) NULL;
}

/*
 * find_rtl_id - Given rtl_id and block_id, find the rtlptr
 */
struct list *find_rtl_id(int rtl_id, int block_id)
{
   struct bblock *cblk;
   struct list *rtlptr;

   /* first, find the basic block that contains the rtl_id */
   if (!rtl_id) return NULL;
   cblk = find_block_id(block_id);
   if (!cblk)
      return NULL;

   /* find the rtl_id in the basic block */
   for (rtlptr = cblk->rtls; rtlptr; rtlptr = rtlptr->next) {
      if (rtlptr->rtl_id == rtl_id)
         return rtlptr;
   }
   return NULL;
}

/*
 * change_bb_label - change the label of the basic block
 */
void change_bb_label(int block_id, char *label)
{
   struct bblock *cblk;

   /* find the basic block through the block_id */
   if (!(cblk = find_block_id(block_id)))
      ERR_RET("change_bb_label(): Can't find this block_id\n");
   
   /* label the basic block with the new label */
   update_label(cblk, copy_string(label));
}

/*
 * insert_basic_block - insert a new basic block
 */
void insert_basic_block(int before_block_id)
{
   struct bblock *new_blk, *before_blk;
   struct bblock *find_bottom();

   /* check to see if the rtl_id is consistent */
   check_cf(1);

   if (before_block_id) {
      before_blk = find_block_id(before_block_id);
      if (!before_blk->up)
         ERR_RET("insert_basic_block(): Can't create a new block at the very beginning\n");

      /* create the new basic block */
      new_blk = new_basic_block(NULL);
      
      /* update the down list */
      before_blk->up->down = new_blk;
      new_blk->up = before_blk->up;
      mod_bb_ptrs(MOD_DOWN, before_blk->up, new_blk);
      
      new_blk->down = before_blk;
      before_blk->up = new_blk;
      mod_bb_ptrs(MOD_DOWN, new_blk, before_blk);      

      /* if the block above has a FALLTHRU to this block */
      if (before_blk->status & FALLTHRU) {
         new_blk->up->left = new_blk;
         new_blk->status |= FALLTHRU;
         mod_bb_ptrs(MOD_LEFT, new_blk->up, new_blk);      
         new_blk->left = before_blk;
         mod_bb_ptrs(MOD_LEFT, new_blk, before_blk);

         /* Since the left pointers has been changed, recalculate all preds */
         clear_all_preds();
         recalculate_all_preds();
      }
   }
   else {
      extern struct bblock *bottom;

      /* create the new basic block */
      bottom = find_bottom();
      new_blk = new_basic_block(NULL);

      bottom->down = new_blk;
      new_blk->up = bottom;
      bottom = new_blk;
      mod_bb_ptrs(MOD_DOWN, new_blk->up, new_blk);      
   }

   check_cf(1);

   /* end update the down list */
}

/*
 * find_bottom - find the bottom of the down list
 */
struct bblock *find_bottom()
{
   struct bblock *cblk;

   for (cblk = top; cblk->down; cblk = cblk->down)
      ;
   return cblk;
      
}
/*
 * delete_basic_block - delete the basic block from the down list.
 *                      first, check to see if the basic block is empty.
 */
void delete_basic_block(int block_id)
{
   struct bblock *cblk, *down;
   register struct blist *bptr;
   extern struct bblock *top, *bottom;
   
   if (!(cblk = find_block_id(block_id)))
      ERR_RET("delete_basic_block(): Can't find this block_id: \n");

   /* make sure that the block is empty */
   if (cblk->rtls)
      ERR_RET("delete_basic_block(): The basic block is not empty\n");
   if (!cblk->up)
      ERR_RET("delete_basic_block(): Cannot delete the entry basic block\n");
      
   /* Make other blocks that jump to deleted block, now jump to the following
      block. */
   for (bptr = cblk->preds; bptr; bptr = bptr->next) {
      struct bblock *blk_ptr = bptr->ptr;

      if (blk_ptr) {
         if (blk_ptr->left == cblk) {
            blk_ptr->left = cblk->down;
            mod_bb_ptrs(MOD_LEFT, blk_ptr, cblk->down);
         }
         else {
            blk_ptr->right = cblk->down;
            mod_bb_ptrs(MOD_RIGHT, blk_ptr, cblk->down);            
         }
      }
   }

   /* modify the down pointer of the above block to the next block */
   cblk->up->down = cblk->down;
   if (cblk->down)
      cblk->down->up = cblk->up;
   else
      bottom = cblk->up; 

   /* send information to the viewer */
   if (cblk->up)
   mod_bb_ptrs(MOD_DOWN, cblk->up, cblk->down);

   /* recalculate all the predecessors */
   clear_all_preds();
   recalculate_all_preds();
   calc_status();
 
   /* this remv_basic_block do not modify the down pointer.
      has to modify the down pointer myself to be continued. */
   remv_basic_block(cblk);
   
}

/*
 * change_bb_ptrs - Change the down, left or right pointer of the basic block
 */
void change_bb_ptrs(char mod_type, int block_id, int new_ptr_id)
{
   struct bblock *cblk, *new_ptr;

   /* find the basic block through the block_id */
   if (!(cblk = find_block_id(block_id)))
      ERR_RET("change_bb_ptrs(): Can't find this block_id\n");
   if (!(new_ptr = find_block_id(new_ptr_id)))
      ERR_RET("change_bb_ptrs(): Can't find this new basic block\n");
   
   /* label the basic block with the new label */
   switch (mod_type) {
      case MOD_DOWN:
         cblk->down = new_ptr;
         if (new_ptr)
            new_ptr->up = cblk;
         break;

      case MOD_LEFT:
         cblk->left = new_ptr;
         break;

      case MOD_RIGHT:
         cblk->right = new_ptr;
         break;         

      default:
         ERR_RET("change_bb_ptrs(): No such type");
   }
   
   /* update the predecessors */
   clear_all_preds();
   recalculate_all_preds();
   calc_status();
   
   /* call mod_bb_ptrs to send message to the viewer, and save the
      changed info */
   mod_bb_ptrs(mod_type, cblk, new_ptr);
}

/*
 * move_rtl_id - move the rtl through the ids
 */
void move_rtl_id(int rtl_id, int block_id, int before_id, int new_block_id)
{
   struct bblock *old_blk, *new_blk, *label_dst;
   struct list *rtlptr, *before;
   CODE_TYPE type_code;

   /* find the bblock ptrs and rtlptrs. */
   old_blk = find_block_id(block_id);
   new_blk = find_block_id(new_block_id);
   rtlptr = find_rtl_id(rtl_id, block_id);
   before = find_rtl_id(before_id, new_block_id);
   
   if (!rtlptr)
      ERR_RET("move_rtl_id(): Can't find the rtl_id\n");
      
   /* determine the code type */
   type_code = get_code_type(rtlptr->val);

   /* check the semantics of this code, make sure that it will be inserted
      at the end of a basic block */
   if (type_code != NORMAL) {
      if (before_id)
         ERR_RET("move_rtl_id(): The transfer of control must be at the end of a block.\n");
   }
   real_move_rtl (rtlptr, old_blk, before, new_blk);

   /* send message to the viewer */
   move_rtl(rtlptr, old_blk, before, new_blk);

   /* for deleting rtl, the control flow is updated as below: */
   update_control_flow_delete(rtlptr, old_blk);

   /* for inserting rtl, the control flow is updated as below: */
   if (type_code != NORMAL)
      update_control_flow_insert(new_blk,
                                 get_label_down(get_label_str(rtlptr->val)),
                                 type_code);

   /* check to see if the control flow is correct */
   check_cf(1);
}


/*
 * insert_new_rtl - real insert a new rtl
 */
struct list *insert_new_rtl(int rtl_id, unsigned int rtl_type, char *rtl_val,
                            int block_id, int before_id)
{
   struct list *before, *rtlptr;
   struct bblock *cblk;

   /* Insert the new structure into the basic block indicated */
   cblk = find_block_id(block_id);
   before = find_rtl_id(before_id, block_id);
   rtlptr = add_rtl_with_type(cblk, before, rtl_val, NULL, RTLINE);
   return rtlptr;
}

/*
 * send_error - just to send an error message to the user interface
 */
void send_error(char *err_msg)
{
   char str[100];
   char c = ERR_HAND;

   sprintf(str, "%c %s", c, err_msg);
   fprintf(stderr, "%s\n", str);
   write_to_socket(str);
}

/*
 * quite - quit function: used so we can set a breakpoint when vpo exits
 */
void quit(int n)
{
   exit(n);
}

/*
 * get_label_str - get the label from src
 *                 if there is no label, then return NULL
                   instructions having label:
                      PC=IC:0,L78; PC=L77;
 */
char *get_label_str(char *src)
{
   char *ptr;
   char dst[10];
   int i;
   
   if (!is_pc_begin(src))
      return NULL;
   for (ptr = src; *ptr; ptr++) {
      if (*ptr == 'L')
         break;
   }
   if (!(*ptr))
      return NULL;
   
   dst[0] = 'L';
   ptr ++;
   for (i=1; isdigit(*ptr); i++, ptr ++)
      dst[i] = *ptr;
   dst[i] = '\0';
   
   return (strcpy(CALLOC(strlen(dst)), dst));
}

/*
 * get_label_down - Get the bblock pointer of the label from the down list
 *                  of the basic blocks.
 */
struct bblock *get_label_down(char *label)
{
    register struct bblock *cblk;

    if (!label)
       return NULL;
    for (cblk = top; cblk; cblk = cblk->down)
       if (cblk->label && !strcmp(cblk->label, label))
          return (cblk);
    return ((struct bblock *) NULL);
}

/*
 * is_pc_begin - see if the string begins with the program counter
 *               almost the same as is_pc, just ignore the blank character.
 */
int is_pc_begin(register char *s)
{
   while (isspace(*s))
      s++;
   if (*s++ == 'P' && *s++ == 'C')
      return (1);
   else
      return (0);
}

/*
 * is_rt - verify RTL is a PC=RT;
 */
int is_rt(char *s)
{
    return (*s == 'P' && *(s + 1) == 'C' && *(s + 3) == 'R' &&
            *(s + 4) == 'T' );   
}

/*
 * get_code_type - get a type of a code (RTL)
 */
CODE_TYPE get_code_type(char *code)
{
   /*determine the code type*/
   if (is_ucjmp(code))
      return UCJMP;
   else if (is_rjmp(code))
      return RJMP;
   else if (is_indjmp(code))
      return INDJMP;
   else if (is_rt(code))
      return RETURN;
   else
      return NORMAL;
}

/*
 * check_jump_semantics - check the jump semantics, if the semantics is wrong,
 *                        return 1; otherwise return 0
 */
int check_jump_semantics(char *code, CODE_TYPE type_code,
                         struct message *message_ptr)
{
   struct bblock *label_dst;
   extern int rtl_id;

   /* check to see if the rtl_id is consistent */
   if (message_ptr->rtl_id != rtl_id+1) {
      send_error("Inconsistent rtl_id\n");
      return 1;
   }
             
   /* first, check if it is a transfer of control.
      if so, it is must be at the end of the block */
   if (type_code == INDJMP) {
      send_error("Indirect jump not allowed. \n");
      return 1;               
   }
      
   if (type_code != NORMAL) {
      if (message_ptr->before_id) {
         send_error("The transfer of control must be at the end of a block.\n");
         return 1;               
      }
   }

   /* check if it has a label, if so, does this label exist? */
   if (type_code == UCJMP || type_code == RJMP) {
      label_dst = get_label_down(get_label_str(code));
      if (!label_dst) {
         send_error("Branch target does not exist\n");
         return 1;
      }
   }
   
   return 0;
}

/*
 * update_control_flow_insert - update the control flow if the type code is
 *                              not normal
 */
int update_control_flow_insert(struct bblock *cblk, struct bblock *label_dst,
                               CODE_TYPE type_code)
{
   switch (type_code) {
      case UCJMP:
         cblk->left = label_dst;
         cblk->right = NULL;
         break;

      case RJMP:
         cblk->left = cblk->down;
         cblk->right = label_dst;
         break;

      case RETURN:
         {
            extern struct bblock *ret_block;
            cblk->left = ret_block;
            cblk->right = NULL;
         }
         break;
      
      default:
         break;
   }
   mod_bb_ptrs(MOD_LEFT, cblk, cblk->left);
   mod_bb_ptrs(MOD_RIGHT, cblk, cblk->right);
   
   /* recalculate the preds to make the control flow consistent */
   if (type_code != NORMAL) {
      clear_all_preds();
      recalculate_all_preds();
      calc_status();
   }
   
   return 1;
}

/*
 * update_control_flow_delete - update the control flow in deleting operation
 */
int update_control_flow_delete(struct list *rtlptr, struct bblock *cblk)
{
   if (is_pc_begin(rtlptr->val)) {

      /* check to see if the control flow is correct */
      cblk->left = cblk->down;
      cblk->right = NULL;

      /* sending messages to the viewer */
      mod_bb_ptrs(MOD_LEFT, cblk, cblk->left);
      mod_bb_ptrs(MOD_RIGHT, cblk, cblk->right);

      /* recalculate the predecessors */
      clear_all_preds();
      recalculate_all_preds();
      calc_status();      
   }
}

/*
 * find_last_rtl - find the last rtl in a basic block
 */
struct list *find_last_rtl(struct bblock *cblk)
{
   struct list *last;

   /* find the last rtl in the block */
   for (last = cblk->rtlend; last; last = last->prev)
      if (cnts(last))
         break;
   return last;
}
#endif
