#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <strings.h>
#include <math.h>


#include "../tools/pdtool/build_data_structure.h"
#include "../tools/pdtool/misc.h"

#include "vpo.h"
#include "callgraph.h"
#include "misc.h"
#include "readlp.h"

#include "rtl.h"  /* Machine dependent. */

extern partition_description_t pd;
int total_number_of_partitions=0;
static int is_heap_variable(locality_profile_variable_t* var);


static locality_profile_variable_t* find_variable(                
		locality_profile_t *p,         /* locality profile */
		char* name, 			/* name of variable */
		int addr, 			/* address of variable */
		int size, 			/* size of variable */
		int index 			/* index number of variable */
	)
{
	
	/* if the profile isn't big enough, extend it */
	if(p->array_size<=index)
	{


		int new_array_size = 2*p->array_size < index ? index+1 : p->array_size*2;

		/* realloc fails if passed a null pointer.  malloc(4) returns a pointer to no-memory 
			so that realloc won't fail */
		if(p->variables==NULL)	p->variables=malloc(4);

	
		p->variables=(locality_profile_variable_t*)realloc(p->variables,(new_array_size)*sizeof(locality_profile_variable_t));
		assert(p->variables);
		bzero(p->variables+p->array_size,(new_array_size-p->array_size)*sizeof(locality_profile_variable_t));
		p->array_size=new_array_size;
	}

	if(p->variables[index].name==NULL)
	{
		assert(p->variables[index].addr==0);
		assert(p->variables[index].size==0);

		p->variables[index].name=strdup(name);
		assert(p->variables[index].name);

		p->variables[index].addr=addr;
		p->variables[index].size=size;
	}
	else
	{
		/* else, check that its the same variable */
		assert(strcmp(p->variables[index].name,name)==0);

		/* we can't assert that the addresses are the same, or even non-zero for stack variables */
		/* (stack variables can have no address if they are not in-scope at the end of the program */
		assert(p->variables[index].addr==addr || strstr(p->variables[index].name,"()->"));
		assert(p->variables[index].size==size);

	}

	return &p->variables[index];
	
}


static void add_range(
		locality_profile_t *p, 	/* locality profile */
		char* name, 			/* name of variable */
		int addr, 			/* address of variable */
		int size, 			/* size of variable */
		int index, 			/* index number of variable */
		long long start_ref, 		/* starting reference for range */
		int num_refs, 			/* number of refs in range */
		long long last_ref		/* last reference in range */
	)
{

	locality_profile_variable_t* var=find_variable(p,name,addr,size,index); 	
	locality_profile_range_t *the_range;


	if(var->ranges==NULL)	var->ranges=malloc(4);
	
	if(var->array_size<=var->num_ranges)
	{

		int new_array_size = var->array_size==0 ? 10 : 2*var->array_size;
		
		var->ranges=(locality_profile_range_t*)realloc(var->ranges,(new_array_size)*sizeof(locality_profile_range_t));
		assert(var->ranges);
		var->array_size=new_array_size;
	}

	the_range=&var->ranges[var->num_ranges];
	var->num_ranges++;	 /* move to next range after getting the addr of the last range */


	the_range->start_ref=start_ref;
	the_range->num_refs=num_refs;
	the_range->last_ref=last_ref;

	return;
	
}

int read_profile(FILE* fp, locality_profile_t **p)
{

	char name[10000];

	assert(p);

	*p=(locality_profile_t*)calloc(1,sizeof(locality_profile_t));

	/* get header string */
	fgets(name,10000,fp);
	while(!feof(fp))
	{
		long long addr, size, index, start_ref, num_refs, last_ref;
		int sf;
		fscanf(fp, "(%s%llx%lld%lld%lld%lld%lld)", name, &addr, &size, &index, &start_ref, &num_refs, &last_ref);
		if(feof(fp)) break;
		add_range(*p, name, addr, size, index, start_ref, num_refs, last_ref);

                fgets(name,10000,fp);
                if(feof(stdin)) break;


	}
	return 0;

}


#define min(a,b) ((a)<(b) ? (a) : (b))
#define max(a,b) ((a)>(b) ? (a) : (b))


long long tcf_counter=0;

#define cost(a) (2*(a))


int calc_tcf(locality_profile_t* p, int var_a_index, int var_b_index)
{

	locality_profile_variable_t* var_a=&p->variables[var_a_index];
	locality_profile_variable_t* var_b=&p->variables[var_b_index];

	int var_a_ranges=var_a->num_ranges;
	int var_b_ranges=var_b->num_ranges;

	int var_a_counter=0, var_b_counter=0;

	int tcf=0;
	long long max_end, max_start;

	while(var_a_counter<var_a_ranges && var_b_index < var_b_ranges)
	{
		locality_profile_range_t *range_a=&var_a->ranges[var_a_counter];
		locality_profile_range_t *range_b=&var_b->ranges[var_b_counter];

		long long min_end=min(range_a->last_ref, range_b->last_ref);
		long long max_start=max(range_a->start_ref, range_b->start_ref);

		long long overlap_cycles=min_end-max_start;

		/* if overlap cycles is positive, these two ranges overlap */
		if(overlap_cycles>0)
		{
			double percent_overlap_a=overlap_cycles/(float)(range_a->last_ref - range_a->start_ref);
			double percent_overlap_b=overlap_cycles/(float)(range_b->last_ref - range_b->start_ref);

			tcf+=cost((percent_overlap_a*range_a->num_refs)+(percent_overlap_b*range_b->num_refs));
		}


		/* update */
		if(range_a->last_ref < range_b->last_ref)
			/* move range A */
			var_a_counter++;
		else if(range_b->last_ref < range_a->last_ref)
			/* move range B */
			var_a_counter++;
		else if (range_b->last_ref==range_a->last_ref)
		{ 
			/* both counters can move if they end on the same instr */ 
			var_a_counter++;
			var_b_counter++;
		}

		tcf_counter++;
	}
		
	return tcf;
}

static int name_compare(const void *p1, const void *p2)
{
        locality_profile_variable_t *i = ((locality_profile_variable_t *)p1);
        locality_profile_variable_t *j = ((locality_profile_variable_t *)p2);
	char *ci, *cj;
	int result;
	int heap1, heap2;

	if(i->name!=NULL &&  j->name==NULL)
		return -1;
	if(i->name==NULL &&  j->name!=NULL)
		return 1;
	if(i->name==NULL &&  j->name==NULL)
		return 0;


	/* we should have 2 names here */
	assert(i->name && j->name);

	result=strcmp(i->name,j->name);

	if(result==0)
		return 0;
	if(result<0)
		return -1;
	if(result>0)
		return 1;

	assert(0);

}


static int heapaddress_compare(const void *p1, const void *p2)
{
        locality_profile_variable_t *i = ((locality_profile_variable_t *)p1);
        locality_profile_variable_t *j = ((locality_profile_variable_t *)p2);
	char *ci, *cj;
	int result;
	int heap1, heap2;

	if(i->name!=NULL &&  j->name==NULL)
		return -1;
	if(i->name==NULL &&  j->name!=NULL)
		return 1;
	if(i->name==NULL &&  j->name==NULL)
		return 0;


	heap1=is_heap_variable(i);
	heap2=is_heap_variable(j);


	/* sort the heap vars first and others later */
	if(heap1 && !heap2)
		return -1;
	if(!heap1 && heap2)
		return 1;
	if(!heap1 && !heap2)
		return 0;


	/* if we get here, they both have a name */
	/* and they are both heap variables */
	ci=i->name; 
	cj=j->name;
	while(ci && cj && *ci && *cj)
	{
		if(fmatch(ci,"->("))
			ci=epos;
		else
		{
			ci=NULL;
		}
		if(fmatch(cj,"->("))
			cj=epos;
		else
		{
			cj=NULL;
			break;
		}

		if(ci && cj)
		{
			int addr1, addr2;
			fmatch(ci,",");
			addr1=atoi(epos);
			fmatch(cj,",");
			addr2=atoi(epos);
	
			if(addr1 > addr2) 
             			return -1;
			if(addr1 < addr2) 
             			return 0;
		}

	}
	if(ci && !cj)
             return -1;
	if(!ci && cj)
             return 1;
	
        if (i->total_refs > j->total_refs)
             return -1;
        if (i->total_refs < j->total_refs)
             return 1;

        return (0);
}


static int ref_compare(const void *p1, const void *p2)
{
        locality_profile_variable_t *i = ((locality_profile_variable_t *)p1);
        locality_profile_variable_t *j = ((locality_profile_variable_t *)p2);

	/* if i is assigned a part, and j isn't, put i first */
	if(i->assigned_part!=NULL &&  j->assigned_part==NULL)
		return -1;

	/* if j is assigned a part, and i isn't, put j first */
	if(i->assigned_part==NULL &&  j->assigned_part!=NULL)
		return 1;
		
	/* otherwise, they're both assigned a part, or both not assigned.. compare refs */

        if (i->total_refs > j->total_refs)
             return -1;
        if (i->total_refs < j->total_refs)
             return 1;

        return (0);
}


static part_list_t* find_default_part(partition_description_t* pd)
{
	part_list_t* parts;
	for(parts=pd->parts;parts; parts=parts->next)
		if(parts->default_item)
			return parts;

	return NULL;
}


static int is_local_variable(locality_profile_variable_t* var)
{
	return (strstr(var->name,"()->")!=NULL);
}

static int is_heap_variable(locality_profile_variable_t* var)
{
	return ((strstr(var->name,"->")!=NULL) && !is_local_variable(var));
}


static int is_global_variable(locality_profile_variable_t* var)
{
	return (strstr(var->name,"->")==NULL);
}

#define assign_heap TRUE
#define assign_locals TRUE
#define assign_globals TRUE

static void do_precolor(locality_profile_t* p, partition_description_t* pd)
{
	int i=NULL;

	locality_profile_variable_t* var=NULL;
	part_list_t* default_part=find_default_part(pd);
	assert(default_part);
	for(i=0;i<p->array_size;i++)
	{
		var=&p->variables[i];	

		/* there are holes in the array, skip them */
		if(var->name==NULL)
			continue;
		if(is_heap_variable(var))
		{
			if(!assign_heap)
				var->assigned_part=default_part;
			else
				var->compiling=TRUE;
		}
		else if(is_local_variable(var))
		{
			if(!assign_locals)
			{
				var->assigned_part=default_part;
			}
			else
			{
				/* check to see if stack vars can be assigned to a different partition */
				char func_name[1000];
				struct call_graph_node* cgn;
		
				strcpy(func_name,var->name);
				fmatch(func_name,"()->");
				*bpos=NULL;
				cgn=find_call_graph_node(function_name_to_id(func_name));
	
				/* can't move vars from recursive functions */
				if(cgn->is_indirectly_recursive)
				{
					var->assigned_part=default_part;
				}
			}
	
		}
		else
		{
			assert(is_global_variable(var));
			if(!assign_globals)
				var->assigned_part=default_part;
		}
		if(!var->compiling)
		{
			/* ick, it's global but we aren't compiling it, must be assigned to default */
			var->assigned_part=default_part;
		}
	}

	return;

}


static void do_setup(locality_profile_t* p, partition_description_t* pd)
{
	part_list_t* parts;
	cache_list_t* caches;
	int changes;
	int count=0;

	for(parts=pd->parts;parts; parts=parts->next)
	{
		parts->user_data=calloc(sizeof(my_part_user_data_t),1);
		((my_part_user_data_t*)parts->user_data)->estimated_average_access_time=parts->latency;
		((my_part_user_data_t*)parts->user_data)->part_num=count;
		count++;
	}
	total_number_of_partitions=count;

	for(caches=pd->caches; caches; caches=caches->next)
		caches->user_data=calloc(sizeof(my_cache_user_data_t),1);
	changes=TRUE;
	while(changes)
	{
		changes=FALSE;
		for(caches=pd->caches; caches; caches=caches->next)
		{
			if(((my_cache_user_data_t*)caches->user_data)->estimated_miss_latency==0)
			{
				if(caches->misses_to_cache && 
					((my_cache_user_data_t*)caches->misses_to_cache->user_data)->estimated_miss_latency)
				{
					((my_cache_user_data_t*)caches->user_data)->estimated_miss_latency=
						caches->misses_to_cache->hit_latency+
						0.1*((my_cache_user_data_t*)caches->misses_to_cache->user_data)->estimated_miss_latency;
					changes=TRUE;
				}
				if(!caches->misses_to_cache)
				{
					int i;
					for(i=0;i<caches->num_parts_cached;i++)
					{
						((my_cache_user_data_t*)caches->user_data)->estimated_miss_latency+=
							caches->the_parts_cached[i]->latency;
					}
					((my_cache_user_data_t*)caches->user_data)->estimated_miss_latency/=caches->num_parts_cached;
					changes=TRUE;
				}
			}
		}
	}


	for(caches=pd->caches; caches; caches=caches->next)
	{
		int i;
		for(i=0;i<caches->num_parts_cached;i++)
		{
			float *latency;
			float estimated_miss_latency;
			parts=caches->the_parts_cached[i];
			latency=& ((my_part_user_data_t*)parts->user_data)->estimated_average_access_time;

			estimated_miss_latency=caches->hit_latency + 
				0.1*((my_cache_user_data_t*)caches->user_data)->estimated_miss_latency;	
			if(estimated_miss_latency<*latency)
				*latency=estimated_miss_latency;
		}
	}
	
	for(parts=pd->parts;parts; parts=parts->next)
	{
		fprintf(stderr,"Estimated access time for part %s is %f\n", parts->name, 
			((my_part_user_data_t*)parts->user_data)->estimated_average_access_time);
	}
}


static int var_fits_in_part(locality_profile_variable_t* var2, int j, locality_profile_t* p, partition_description_t* pd)
{
	int size_so_far=0;
	int i;
	part_list_t* parts;
	for( parts=pd->parts; parts; parts=parts->next)
	{
		if(((my_part_user_data_t*)parts->user_data)->part_num==j)
			break;
	}
	assert(parts);
	if(parts->default_item)
		return TRUE;

	
	/* if this variable is bigger than the partition size, skip out early */
	if(var2->size > parts->size)
		return FALSE;

	for(i=0;&p->variables[i]!=var2; i++)
	{
		if(p->variables[i].assigned_part==parts)
		{
			size_so_far+=p->variables[i].size;
			if(size_so_far+var2->size>=parts->size)
				return FALSE;
		}
	}
	return TRUE;
}

void dump_assignments(locality_profile_t* p)
{
	int i;
	char temp[10000];
	for(i=0;i<p->array_size;i++)
	{
		if(p->variables[i].name==NULL)
			continue;
		sprintf(temp,"Variable %40s (%d):", p->variables[i].name, p->variables[i].total_refs);
		fprintf(stderr,"%50s",temp);
		if(p->variables[i].assigned_part!=NULL)
			fprintf(stderr,"Assigned to partition %s\n",p->variables[i].assigned_part->name);
		else
			fprintf(stderr,"Not assigned a partition\n");
	}
}

char* shorten_name(char* inpname)
{
	static char outname[25];
	strncpy(outname,inpname,21);
	outname[20]=0;
	return outname;
	
}

part_list_t* find_part(char* name, int* part_index, partition_description_t* pd)
{
        part_list_t* parts;
	int i=0;
        for( parts=pd->parts; parts; parts=parts->next)
	{
		if(strcmp(parts->name,name)==0)
		{
			if(part_index) 
				*part_index=i;
			return parts;
		}
		i++;
	}
	return NULL;

}


void read_optimal_assignments(locality_profile_t* p, partition_description_t* pd)
{

	char var_name[1000]="";
	char part_name[1000]="";
	part_list_t* parts=NULL;
	int var_index=-1;
	int part_index=-1;
	int i;

	/* only do this if swpo is turned on*/
	if(!swpo)
		return;

	while(!feof(optimal_assignment_fn))
	{
		fscanf(optimal_assignment_fn,"%s%s\n",var_name, part_name);

		parts=find_part(part_name,&part_index,pd);
		if(!parts)
			fprintf(stderr, "Cannot find part %s\n", part_name), exit(1);
		
		var_index=-1;
		for(i=0;i<p->array_size;i++)
		{
			if(p->variables[i].name==NULL)
                       	 	continue;

			if(strcmp(var_name,p->variables[i].name)==0)
			{
				assert(var_index==-1);
				/* we found the variable */	
				var_index=i;
			}
		}
		if(var_index==-1)
			fprintf(stderr, "Cannot find variable %s in pgm\n", var_name), exit(1);

		if(p->variables[var_index].assigned_part!=NULL && p->variables[var_index].assigned_part!=parts)
			fprintf(stderr, "Cannot reassign variable %s to part %s from part %s\n", var_name, parts->name,
				p->variables[var_index].assigned_part->name), exit(1);

		if(!var_fits_in_part(&p->variables[var_index], part_index,  p,  pd))
			fprintf(stderr, "Partition %s out of room for variable %s\n", parts->name,var_name), exit(1);
		
		p->variables[var_index].assigned_part=parts;
	}

	
	
}


void calc_locality(locality_profile_t* p, partition_description_t* pd)
{
	int i,j,k,tcf;
	locality_profile_variable_t* var=NULL;
	part_list_t* default_part=find_default_part(pd);
	part_list_t* parts;
	float *cost, *part_refs;
	int instructions_part=-1;
	int total_size=0;

	/* setup some constant info about the parts */
	/* sets total_number_of_partitions too */
	do_setup(p,pd);

	/* precolor any variables that can't be assigned to a partition */
	do_precolor(p,pd);

	/* if swpo is turned on, we want to read the optimal assignments from the
	 * user specified file
	 */
	read_optimal_assignments(p,pd);


	/* calc total refs of each one */
	for(i=0;i<p->array_size;i++)
	{
		p->variables[i].total_refs=0;
		for(j=0;j<p->variables[i].num_ranges;j++)
			p->variables[i].total_refs+=p->variables[i].ranges[j].num_refs;
	}
	for(i=0;i<p->array_size;i++)
        { 
		total_size+= p->variables[i].size;
	}
	fprintf(stderr,"Total size of partitionable variables:  %d\n", total_size);


	/* sort highest refs first (assigned parts before unassigned parts) */
     	qsort(p->variables,p->array_size, sizeof(p->variables[0]), ref_compare);

	/* allocate memory for the partition info */
	cost=alloca(sizeof(float)*total_number_of_partitions);
	part_refs=alloca(sizeof(float)*total_number_of_partitions);

	/* find the part number for the instructions part, we need to skip that */
	for(parts=pd->parts;parts;parts=parts->next)
	{
		if(parts->instructions)
			instructions_part=((my_part_user_data_t*)parts->user_data)->part_num;
	}
	assert(instructions_part!=-1);


	/* OK, now we're ready to actually make assignments */
	for(i=0;i<p->array_size;i++)
	{
		float best_cost; 
		int best_cost_index;
		var=&p->variables[i];	
		/* skip the holes in the array */
		if(var->name==NULL)
			continue;
		/* skip any parts that're already assigned */
		if(var->assigned_part!=NULL)
			continue;

		bzero(cost,sizeof(float)*total_number_of_partitions);
		bzero(part_refs,sizeof(float)*total_number_of_partitions);
		for(j=0;j<p->array_size;j++)
		{
			locality_profile_variable_t* var2=NULL;
			var2=&p->variables[j];
			/* skip any holes */
			if(var2->name==NULL)
				continue;
			if(var2->assigned_part==NULL)
				continue;

			tcf=calc_tcf(p,i,j);
			
			cost[((my_part_user_data_t*)var2->assigned_part->user_data)->part_num]+=tcf;
			part_refs[((my_part_user_data_t*)var2->assigned_part->user_data)->part_num]+=var2->total_refs;
		}

		/* find the part with the best cost */
		/* first, multiply by latency of part */
		/* make the cost nonzero, and factor in part latecy */

		for(parts=pd->parts; parts; parts=parts->next)
		{

			float tcf=cost[((my_part_user_data_t*)parts->user_data)->part_num];
			float this_part_refs= part_refs[((my_part_user_data_t*)parts->user_data)->part_num];

			/* no conflicts if no caching */
			if(parts->latency==((my_part_user_data_t*)parts->user_data)->estimated_average_access_time)
				tcf=0;
	

			/* estimate time for this variable */
			cost[((my_part_user_data_t*)parts->user_data)->part_num]=
				pow(((my_part_user_data_t*)parts->user_data)->estimated_average_access_time*var->total_refs,2)+
				pow(tcf,2);
		
		}

		best_cost_index=((my_part_user_data_t*)default_part->user_data)->part_num;
		best_cost=cost[best_cost_index];
		for(j=0;j<total_number_of_partitions;j++)
		{
			if(cost[j]<best_cost && j!=instructions_part &&
				var_fits_in_part(var,j,p,pd))
			{
				best_cost=cost[j];
				best_cost_index=j;
			}
		}

		/* make the assignment */
		for(j=0, parts=pd->parts; j<best_cost_index; j++,parts=parts->next)
			assert(parts);

		/* make sure we actually found the best cost */
		assert(parts);
		assert(((my_part_user_data_t*)parts->user_data)->part_num==best_cost_index);

		var->assigned_part=parts;
		fprintf(stderr, "Variable %20s(%10d) in part %20s\tcosts:", shorten_name(var->name), var->total_refs, parts->name);

		for(j=0;j<total_number_of_partitions;j++)
			fprintf(stderr,"%.3e ",cost[j]);
		fprintf(stderr,"\n");
		
	}

}


#ifdef STANDALONE_PARTITIONING

#define usage() fatal("Usage: readlp parts.pd profile.lp")

main(int argc, char* argv[])
{
	FILE *lp_in, *parts_in;
	extern FILE* pdtoolin;
	locality_profile_t* p;
	extern partition_description_t pd;


	if(argc!=3)
	{
		usage();
	}


	parts_in=fopen(argv[1],"r");
	if(!parts_in) fatal("cannot open partition description %s", argv[1]);

	lp_in=fopen(argv[2],"r");
	if(!lp_in) fatal("cannot open locality profile %s", argv[2]);


	/* parse the partition description, not the locality profile, dummy */
	pdtoolin=parts_in;
	if(pdtoolparse())
		fatal("partition description parse unsuccessful\n");

	if(read_profile(lp_in, &p))
		fatal("locality profile parse unsuccessful\n");

	calc_locality(p,&pd);
	printf("total count %lld", tcf_counter);
}

#else 
/* the locality profile  
 */
static locality_profile_t* p=NULL;

/* a safety check to make sure +g is turned on */
static int globals_gathered=FALSE;


/* emit the partition free list for partition "parts" */
void emit_heap_freelist(part_list_t* parts, int space_used)
{
	printf(".globl %s_freelist_head\n", parts->name );	
	printf(".align 4\n");
	printf("%s_freelist_head:\n", parts->name	);
	printf("\t.word heapstart_%s\n", parts->name);
	printf("\t.word 0\t# freelist head.\n"); 
	printf("\t.word %d\n", parts->size-space_used);
}

void emit_heap_assignments(locality_profile_t* p, partition_description_t *pd)
{
	int vars_assigned_to_non_default_parts=0;
	char* next_set;
	char* temp_ptr;
	char name[10000];
	char temp[10000];
	int i,j;

	/* sort the variables so the heap ones are first, and are in order by stack trace */ 
     	qsort(p->variables,p->array_size, sizeof(p->variables[0]), heapaddress_compare);

	printf(".globl partition_assignments\n");
	printf(".align 4\n");
	printf("partition_assignments:\n");
	for(i=0;i<p->array_size;i++)
	{
		if(p->variables[i].name && !p->variables[i].assigned_part->default_item 
			&& is_heap_variable(&p->variables[i]))
		{
			vars_assigned_to_non_default_parts++;

			strcpy(name,p->variables[i].name);
			next_set=name;


			/* calc ret addr depth */
			epos=next_set;
			for(j=0;j<5;j++)
				if(!fmatch(epos,"->"))
					break;

			/* emit ret_addr depth */
			printf("\t.word %d\n", j>4 ? 4 : j);

			/* emit the ret_addr trace */
			for(j=0;j<4;j++)
			{

				if( next_set && fmatch(next_set, "->("))
				{
					next_set=epos;
					/* get the set into a new string */
					strcpy(temp,next_set);
					fmatch(temp,")");
					*bpos=0;
					fmatch(temp,",");
					temp_ptr=epos;
					fmatch(epos,",");
					/* remove the extra ret_addr */
					sub(temp_ptr,epos,"");


					/* the nums are in hex, add 0x so assembler knows */
					fmatch(temp,"+");	sub(bpos,epos,"+0x");
					fmatch(temp,",");	sub(bpos,epos,",0x");
					fmatch(temp,",");	fmatch(epos,","); 	sub(bpos,epos,",0x");
					/* return addr, frame size, and ret_addr offset */
					printf("\t.word %s\n",temp);
				}
				else
				{
					printf("\t.word 0,0,0\n");
				}
			}
			/* emit the partition to assign it to */
			printf("\t.word %s_freelist_head\n", p->variables[i].assigned_part->name);
		}
	}

	/* emit the size of the partition assignments array */
	printf(".globl partition_assignments_size\n");
	printf(".align 4\n");
	printf("partition_assignments_size:\n");
	printf("\t.word %d\n", vars_assigned_to_non_default_parts);
}

/*
typedef struct free_list free_list_t;
struct free_list
{
        char* next_chunk;
        int chunk_size;
};

 0 means no promotions to any non-default partition 
int partition_assignments_size=0;
struct
{
        int size;
        int depth_of_ras;
        struct ras_entry
        {
                int ret_addr;
                int frame_size;
                int ret_addr_offset;
        } ras[4];      
        free_list_t* free_list_head;
} partition_assignments [0]={0,};
*/

void emit_partition_assignments()
{
	extern partition_description_t pd;
	int i;
	part_list_t* parts;
	int space_used;


	
	/* we sort variables by name so they're assigned deterministically
	 * otherwise, the order is semi-random 
	 */
     	qsort(p->variables,p->array_size, sizeof(p->variables[0]), name_compare);

	printf(".data\n");
	for(parts=pd.parts; parts; parts=parts->next)
	{
		space_used=0;
		if(parts->instructions)
			continue;
		printf(".global %s\n", parts->name);
		printf("%s:\n", parts->name);
		
		for(i=0;i<p->array_size;i++)
		{

			/* move stack/global variables to this part if they're assigned */
			if(p->variables[i].name && p->variables[i].assigned_part==parts &&
				(is_local_variable(&p->variables[i])||is_global_variable(&p->variables[i])))
			{
				struct list* rtls;

				space_used+=p->variables[i].size;

				/*
				if(!p->variables[i].compiling && (p->variables[i].rtls!=NULL))
					fatal("");
				*/
				assert(p->variables[i].compiling || (p->variables[i].rtls==NULL));

				assert( (p->variables[i].compiling && p->variables[i].rtls!=NULL) || parts->default_item);
				
				for( rtls=p->variables[i].rtls; rtls; rtls=rtls->next)
				{
					printf("%s\n",rtls->val);	
				}
			}

		}


		if(!parts->default_item && !parts->instructions)
		{
			assert(space_used<=parts->size);
			printf(".align 4\n");
			printf("heapstart_%s:\n", parts->name);
			printf("\t.space %d\n", (parts->size-space_used));
		}


		printf(".global end_%s\n", parts->name);
		printf("end_%s:\n", parts->name);

		if(!parts->default_item &&  !parts->instructions)
			emit_heap_freelist(parts, space_used);
	}
	emit_heap_assignments(p,&pd);
}

/* returns the globals name */
static char* get_global_name(struct list* rtl)
{
	static char name[1000];
	strcpy(name,rtl->val);

	/* remove the colon at the end of the label */
	name[strlen(name)-1]=NULL;	

	return name;

}


/* read the locality profile from the filename given 
 */
void read_locality(char* lp_fn)
{
	FILE *lp_in;

	lp_in=fopen(lp_fn,"r");
	if(!lp_in) 
		fatal("cannot open locality profile %s", lp_fn);

	if(read_profile(lp_in, &p))
		fatal("locality profile parse unsuccessful\n");
}

/* called when all functions are done being optimized and emitted 
 * used to assign partitions to variables and emit those variables into the
 * proper partition 
 */
void drive_partitioning(char* parts_fn)
{
	FILE *parts_in;
	extern FILE* pdtoolin;

	if(!swg)
		fatal("global variables not gathered, ensure +g is enabled");

	parts_in=fopen(parts_fn,"r");
	if(!parts_in) 
		fatal("cannot open partition description %s", parts_fn);


	/* parse the partition description, not the locality profile, dummy */
	pdtoolin=parts_in;
	if(pdtoolparse())
		fatal("partition description parse unsuccessful\n");


	calc_locality(p,&pd);
	fprintf(stderr,"total count %lld", tcf_counter);

}

static struct list* make_rtl_perm(struct list* rtl)
{
	struct list* new_rtl=(struct list*)CALLOC_PERM(sizeof(struct list));

	/* copy all fields shallow */
	*new_rtl=*rtl;
	
	/* deep copy the rtl_val */
	new_rtl->val=(char*)strcpy((char*)CALLOC_PERM(strlen(rtl->val)+1),rtl->val);

	new_rtl->next=NULL;
	new_rtl->prev=NULL;
	new_rtl->daddy=NULL;

	return new_rtl;
}

static struct list* make_rtls_permenent(struct list* rtl)
{
	struct list *new_start=NULL, *new_end=NULL, *old_end=NULL;
	
	assert(rtl);
	while(rtl)	
	{
		new_end=make_rtl_perm(rtl);

		/* memorize the new start */
		if(!new_start)
			new_start=new_end;
		/* link the lists */
		if(old_end)
			old_end->next=new_end;
		new_end->prev=old_end;

		old_end=new_end;
		rtl=rtl->next;
	}
	return new_start;
}

static void move_global(struct bblock* blk, struct list *start, struct list* end)
{
	int i;
	char* name=get_global_name(start);

	/* grab the alignment directive for this global, if there is one */
	if(start->prev && start->prev->val && strstr(start->prev->val,"align")!=NULL)
		start=start->prev;
	for(i=0;i<p->array_size;i++)
	{
		if(p->variables[i].name!=NULL &&
			strcmp(p->variables[i].name, name)==0)
		{
			if(start->prev)
				start->prev->next=end->next;
			else
				blk->rtls=end->next;
			if(end->next)
				end->next->prev=start->prev;
			else
				blk->rtlend=start->prev;

			start->prev=end->next=NULL;

			p->variables[i].rtls=make_rtls_permenent(start);

			return;
		}
	}
	fprintf(stderr,"%s not found\n", name);


}


static locality_profile_variable_t* lookup_variable(char* name)
{
	int i;
	for(i=0;i<p->array_size;i++)
	{
		if(p->variables[i].name!=NULL &&
			strcmp(p->variables[i].name, name)==0)
		{
			return &p->variables[i];
		}
	}
	return NULL;

}

static void mark_var_as_being_compiled(char* name)
{
	locality_profile_variable_t* var=lookup_variable(name);
	if(var)
		var->compiling=TRUE;
	else
		fprintf(stderr, "Warning:  variable %s not found in profile\n",name);

}

/* 
 * this function is called during pass 1 to memorize which
 * variables are being compiled and gather any info about them.
 */

void memorize_variables()
{
	struct locuse* loc=NULL;
	struct bblock* blk=NULL;
	struct list* rtl=NULL, *start_global=NULL;
	extern struct locuse* locs;
	char name[1000]={0,};
	globals_gathered=TRUE;

	for(blk=top;blk;blk=blk->next)
	{
		for(rtl=blk->rtls;rtl;rtl=rtl->next)
		{
                        if(rtl->type&CASELINE)
                        {
                                start_global=NULL;
                        }
                        if(starts_global(rtl))
                        {
                                assert(!start_global);
                                start_global=rtl;
                        }
                        if(ends_global(rtl))
                        {
                                if(start_global)
                                {
					mark_var_as_being_compiled(get_global_name(start_global));
                                }
                                start_global=NULL;
                        }
		}
	}
	for(loc=locs;loc;loc=loc->next)
	{
		/* note that the arg build area and reg save space is not "compiled".. yet */
		if(!loc->arg && loc->reps)
		{
			sprintf(name,"%s()->%s", fn, loc->item);
			mark_var_as_being_compiled(name);
		}
	}
}

static void promote_local(struct locuse* loc, locality_profile_variable_t* var, int new_num)
{
	char glo_name[1000];
	char encoding[REGSTRING];
	struct list* nrtl;
        struct bblock* blk=NULL;
        struct list *rtl=NULL;


	sprintf(glo_name,"%s_%s\t%s\t0\t0",fn,loc->item,getGlobal(new_num));

	/* add this global to the list of globals */
	declaration(glo_name);


	


	/* .data directive */
	nrtl=new_rtl(NULL,ASMLINE,NULL);
	nrtl->val=strdup(".data");

	/* memorize first rtl */
	var->rtls=nrtl;

	/* align directive */
	nrtl=new_rtl(NULL,ASMLINE,nrtl);
	nrtl->val=strdup(".align 4");


	/* the label*/
	nrtl=new_rtl(NULL,ASMLINE,nrtl);
	/* get the label version of this global name */ 
	sprintf(glo_name,"%s_%s:",fn,loc->item);
	nrtl->val=strdup(glo_name);

	/* the size for this variable */
	nrtl=new_rtl(NULL,ASMLINE,nrtl);
	/* get the label version of this global name */ 
	sprintf(glo_name,".space %d", loc->size);
	nrtl->val=strdup(glo_name);




	/* make stack pointer */
	makereg(encoding,RTREG,29);
	/* make string to look for */
	sprintf(glo_name,"%s+%s", encoding,loc->code);


        for(blk=top;blk;blk=blk->next)
        {
                for(rtl=blk->rtls;rtl;rtl=rtl->next)
                {
			if(cnts(rtl))
			{
				while(fmatch(rtl->val,glo_name))
				{
					/* replace the ref with the new global */
					sub(bpos,epos,getGlobal(new_num));
				}

				assert(!fmatch(rtl->val,loc->code));
			}
		}
	}
	/* clear reps field to say this isn't used anymore */
	loc->reps=NULL;

}


static void promote_locals()
{
	char name[1000]={0,};
	part_list_t* default_part=find_default_part(&pd);
	locality_profile_variable_t* var;
	static int maxGloNum=-1;
	extern int maxGlobalNum;
	struct locuse* loc;
	extern struct locuse* locs;

	if(maxGloNum==-1)
	{
		maxGloNum=maxGlobalNum+100;
	}

	for(loc=locs;loc;loc=loc->next)
	{
		if(!loc->arg && loc->reps)
		{
			sprintf(name,"%s()->%s", fn, loc->item);
			var=lookup_variable(name);
			if(var && var->compiling && var->assigned_part!=default_part)
			{
				promote_local(loc,var,maxGloNum++);
			}
		}
	}
	/* locals aren't promoted yet */
	return;
}

static void gather_globals()
{
	struct bblock* blk=NULL;
	struct locuse* loc=NULL;
	extern struct locuse* locs;
	static int global_counter=0;
	struct list *rtl=NULL, *start_global=NULL, *next=NULL;
	char name[1000]={0,};
	globals_gathered=TRUE;

	for(blk=top;blk;blk=blk->next)
	{
		for(rtl=blk->rtls;rtl;rtl=next)
		{
			next=rtl->next;


			/* if we find a CASELINE, skip this global */
			if(rtl->type&CASELINE)
			{
				start_global=NULL;
			}	
			if(starts_global(rtl))
			{
				assert(!start_global);
				start_global=rtl;
			}
			if(ends_global(rtl))
			{
				if(start_global) 
				{
					move_global(blk, start_global,rtl);
					global_counter++;
				}
				start_global=NULL;
			}
		}
		assert(!start_global);
	}
}

/*
 * This function is called during pass 2 in order to promote assigned
 * local variables to global vars and remove any promoted globals 
 * from the function so they can be emitted to the proper partition later.
 */
void promote_variables_from_this_function()
{
	promote_locals();
	gather_globals();
}

#endif
