/**************************************************************************/
/* remjump.c                                                              */
/* File contains functions to remove jumps using loop replication         */
/**************************************************************************/

#ifdef INTERACT

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/time.h>

#include "vpo.h"
#include "arena.h"
#include "remjump.h"
#include "view.h"
#include "misc.h"


/*find_shortest_path - find the shortest path from some node to all other
 *                     nodes in the graph. This function will return the
 *                     shortest "distance[]" from the start node "start" to
 *                     every other node. It will also return the shortest
 *                     "paths". This function is an implementation of
 *                     dijkstra's shortest path algorithm.
 */
void find_shortest_path(int weights[][MAX_BLOCKS], int start, int nodes,
			int distance[], int paths[])
{
  int dist[MAX_BLOCKS]; /*shortest distances so far*/
  int perm[MAX_BLOCKS]; /*nodes whose shortest dist. is permanant*/
  int current;          /*node added to perm most recently*/
  int smalldist;        /*smallest dist, during this iteration*/
  int newdist, i, j, num; /*temporary variables*/

  /*initialization*/
  for(i=0 ; i<nodes ; i++){
    perm[i] = FALSE;
    distance[i] = INT_INFI;
    dist[i] = INT_INFI;
  }
  perm[start] = TRUE;
  dist[start] = 0;
  current = start;
  distance[start] = dist[current];

  num = 0;
  /*for all nodes*/
  while(num < nodes){

    /*find the distance from start node to each other node, passing through
      the current node to determine if distance through current node is
      smaller than the earlier distance*/
    smalldist = INT_INFI;
    for(i=0 ; i<nodes ; i++){

      /*only check for nodes not in the perm list, as we know for sure that
	nodes in the perm list are the shortest distance*/
      if(perm[i] == FALSE){

	newdist = distance[current] + weights[current][i];
	if(newdist < dist[i]){
	  dist[i] = newdist;
	  paths[i] = current;
	}

	/*determine the smallest distance*/
	if(dist[i] < smalldist){
	  smalldist = dist[i];
	  j = i;
	}
      }
    }

    current = j;
    perm[current] = TRUE;
    distance[current] = dist[current];
    num++;
  }
}


/* remove_jumps - remove jumps by replicating code
 *                This optimization should be performed before fill delay
 *                slots and fix entry exit
 */
int remove_jumps()
{
  unsigned int analysis;
  int i, j, count, start, bnum, max_blk_num, length;
  int change, weights[MAX_BLOCKS][MAX_BLOCKS];
  int dist[MAX_BLOCKS], path[MAX_BLOCKS];
  int rep_path[MAX_REP_BLK], num_uncond_jmp;
  
  /*holds the basic blocks containing uncond. jumps (can be sorted)*/
  struct bblock *uncond_jmp[MAX_UJUMPS];
  struct block_info binfo[MAX_BLOCKS];
  struct bblock *cblk, *blk, *curr, *next, *prev, *new, *target;
  struct blist *bl;
  struct list *rtl, *last_rtl;
  extern struct bblock *top;
  extern char *fn;
  extern int measure_flag;

#ifdef INTERACT
  begin_phase(AVOID_UJUMP_TRANS);
#endif  /*INTERACT*/

  change = FALSE;
  
  /*find all the unconditional jumps in this function and sort them by loop
    nesting depth. This will ensure that loops when duplicated will not
    result in more unconditional jumps*/
  num_uncond_jmp = find_uncond_jumps(uncond_jmp);
  sort_uncond_jumps(num_uncond_jmp, uncond_jmp);

  /*for each block which ends in an unconditional jump*/
  fprintf(stderr, "Function %s\n", fn);
  for(bnum=0 ; bnum<num_uncond_jmp ; bnum++){

    cblk = uncond_jmp[bnum];
    last_rtl = (struct list *)get_last_rtl(cblk);

    /*initialize and set the data structures for this optimization. Return
      the highest block number present*/
    max_blk_num = init_rem_jump(binfo, weights);

    /*find the shortest path from the target block of the jump to all the
      remaining blocks in the function. This way we will get the shortest
      distance between the target block to the block following the JUMP
      block in the control flow, as well the distances between the target
      block and all the return and exit blocks, in one go*/
    /*start = cblk->left->block_id;*/
    start = cblk->block_id;
    find_shortest_path(weights, start, max_blk_num, dist, path);
    
    /*determine the path which will be least costly when replicated and
      still get rid of the unconditional jump*/
    length = det_cheap_path(binfo, start, dist, path, rep_path);
    
    /*check if this control flow would be valid*/
    
    /*remove the current block from the predecessor block of the target*/
    remv_predecessor(cblk, cblk->left);
      
    /*duplicate the blocks and insert them in the control flow*/
    target = cblk->down;
    rep_path[length] = cblk->down->block_id;
    prev = cblk;
    for(i=0 ; i<length ; i++){

      curr = (struct bblock *)binfo[rep_path[i]].ptr;
      next = (struct bblock *)binfo[rep_path[i+1]].ptr;
      new = dup_block(curr, prev, next);
      prev = new;
      binfo[new->block_id].ptr = (struct bblock *)new;
    }

#ifdef INTERACT
      begin_trans(VET_OPT_TRANS);
#endif
      
    /*if the duplication is along the return statement, then the last block
      duplicated should be pointed to the return block, else should be
      pointed to the original "down" block of the JUMP block*/
    rtl = get_last_rtl(new);
    if(new->rtlend && rtl && is_return(rtl->val)){
      new->left = ret_block;
      add_predecessor(new, ret_block);
    }
    else{
      new->left = target;
      add_predecessor(new, target);
    }
    new->down = target;
    target->up = new;

#ifdef INTERACT
    mod_bb_ptrs(MOD_LEFT, new, new->left);
    mod_bb_ptrs(MOD_DOWN, new, new->down);
#endif    
    
    /*remove the unconditional jump statement from the source block*/
    remv_rtl(last_rtl);

#ifdef INTERACT
    end_trans();
#endif    

    change = TRUE;
  }

  /*adjust the predecessor and status field information of the duplicated
    basic blocks*/
  calc_status();

  /*adjust the labels of all branch instructions in the function to reflect
    the correct control flow. This could be done during the replication
    process itself. However, I believe it would be easier to do this here,
    after the main loop*/
  adjust_fn_labels();

  /*This may generate unreachable code which causes problems later during
    delay slots, since some blocks contain pseudo registers. Temporary fix*/
  /*remv_unreachable_code(1);*/

#ifdef INTERACT
  end_phase(AVOID_UJUMP_TRANS);
#endif

  return change;
}


/* find_uncond_jumps - find all the unconditional jumps in this function
 */
int find_uncond_jumps(struct bblock* ujmps[])
{
  int num;
  struct bblock *cblk;
  struct list *last_rtl;
  extern struct bblock *top;

  num = 0;
  for(cblk=top ; cblk ; cblk=cblk->down){
    
    last_rtl = (struct list *)get_last_rtl(cblk);
    if(cblk->status & JUMP && last_rtl && is_ucjmp(last_rtl->val))
      ujmps[num++] = (struct bblock*)cblk;
  }

  /*null terminate the list*/
  ujmps[num] = (struct bblock *)NULL;

  return num;
}

/* sort_uncond_jumps - sort the list of unconditional jumps by nesting depth
 */
void sort_uncond_jumps(int num, struct bblock* ujmps[])
{
  int i, j, max_depth, max_blk;
  struct bblock *temp;

  for(i=0 ; i<num ; i++){

    max_depth = -1;
    for(j=i ; j<num ; j++){

      /*find the block with the innermost depth*/
      if(ujmps[j]->freq > max_depth){
	max_blk = j;
	max_depth = ujmps[j]->freq;
      }
    }

    /*swap block with max_depth with the top block in the current list*/
    temp = ujmps[i];
    ujmps[i] = ujmps[max_blk];
    ujmps[max_blk] = temp;
  }
}


/* dup_block - duplicate the block with all the RTLs and link it in the
 *             control flow. Return the LEFT or RIGHT link of the new basic
 *             block which would point to the next new block duplicate
 *             block and pointer to the new basic block
 */
struct bblock *dup_block(struct bblock *cblk, struct bblock *pblk,
			 struct bblock *nblk)
{
  char *label;
  int i;
  struct bblock *new;
  struct list *n_rtlend, *c_rtlend;

#ifdef INTERACT
  begin_trans(VET_OPT_TRANS);
#endif  
  
  /*allocate space for a new basic block and label it*/
  new = (struct bblock *)copy_block(cblk);

  /*adjust the controlflow information*/

  /*the down and left pointers of all the blocks in the duplicated path point
    to the next block in the new control flow. Since that block will only be
    created in the next iteration we have to link these pointers for the
    previous block in this generation*/
  pblk->down = pblk->left = new;
  new->up = pblk;
#ifdef INTERACT
  mod_bb_ptrs(MOD_LEFT, pblk, pblk->left);
  mod_bb_ptrs(MOD_DOWN, pblk, pblk->down);
#endif  

  /*update predecessor list of the new block and old block*/
  add_predecessor(pblk, new);

  /*if the last instruction in the new block is an unconditional jump, then
    in the duplicated path, this block would be jumping to the block
    following it in the control flow. So, the jump instruction can safely be
    removed*/
  n_rtlend = (struct list *)get_last_rtl(new);
  c_rtlend = (struct list *)get_last_rtl(cblk);
  if(cblk->status & JUMP && n_rtlend && is_ucjmp(n_rtlend->val)){
    remv_rtl(n_rtlend);
  }
  /*if the last instruction is a conditional jump, and the next block in the
    duplicated path is not the block following it in the down list, then we
    have to reverse the condition in the duplicated block and change the jump
    to the block following the original block in the down list in the control
    flow*/
  else if(cblk->left && cblk->right &&
	  c_rtlend && !is_ucjmp(c_rtlend->val)){
    if(nblk != cblk->down){
      label_basic_block(cblk->down);
      reverse(n_rtlend, cblk->down->label);
      new->right = cblk->down;

#ifdef INTERACT
      mod_rtl(n_rtlend, NULL, n_rtlend->val);
      mod_bb_ptrs(MOD_RIGHT, new, new->right);
#endif      
      add_predecessor(new, cblk->down);
    }
    else{
      new->right = cblk->right;

#ifdef INTERACT
      mod_bb_ptrs(MOD_RIGHT, new, new->right);
#endif      
      add_predecessor(new, cblk->right);
    }
  }

#ifdef INTERACT
    end_trans();
#endif    
  
  return (struct bblock*)new;
}


/* in_blist - check if a block is present in a blist
 */
int in_blist(struct blist *list, struct bblock *blk)
{
  struct blist *bl;

  for(bl=list ; bl ; bl=bl->next){
    if(bl->ptr == blk)
      return TRUE;
  }

  return FALSE;
}
    
    
/* copy_block - make a copy of the given basic block
 */
struct bblock *copy_block(struct bblock *cblk)
{
  struct bblock *new;
  struct list *rtl, *prtl, *nrtl;

  new = (struct bblock *)new_basic_block(NULL);
  label_basic_block(new);

  /*add all the RTLs of the current block in this new block*/
  prtl = NULL;
  for(rtl=cblk->rtls ; rtl ; rtl=rtl->next){
    nrtl = (struct list *)new_rtl(new, rtl->type, prtl);
    if(rtl->val){
      nrtl->val = (char *)CALLOC(strlen(rtl->val)+1);
      strcpy(nrtl->val, rtl->val);
    }
    nrtl->kind = rtl->kind;
    rcpy(&nrtl->deads, rtl->deads);
#ifdef INTERACT
    insert_rtl(nrtl);
#endif    
    prtl = nrtl;
  }

  return (struct bblock*)new;
}

/* adjust_fn_labels - adjust the labels of all branch instructions to point
 *                    to the correct blocks
 */
void adjust_fn_labels()
{
  char *label;
  struct bblock *cblk;
  struct list *c_rtlend;
  extern struct bblock *top;

  /*Check all the blocks in the function. Since the control flow (i.e. the
    left and right pointers of the blocks are correct, determining the
    correct labels should be easy*/
  for(cblk=top ; cblk ; cblk=cblk->down){

#ifdef INTERACT
    begin_trans(VET_OPT_TRANS);
#endif    
    c_rtlend = (struct list *)get_last_rtl(cblk);
    if(c_rtlend && c_rtlend->val && is_ucjmp(c_rtlend->val)){
      label = cblk->left->label;
      change_branch_label(c_rtlend, label);
    }
    else if(c_rtlend && cblk->left && cblk->right){
      label = cblk->right->label;
      change_branch_label(c_rtlend, label);
    }
#ifdef INTERACT
    end_trans();
#endif        
  }
}

/* change_branch_label - just replace the target label in the branch
 *                       instruction with the new label. Does not affect any
 *                       basic block pointers
 */
void change_branch_label(struct list *rtl, char *label)
{
  char *pos, *val, new[MAXSTRING];
  int len;

  for(len=0,pos=rtl->val ; *pos!='L' ; pos++)
    new[len++] = *pos;

  for(pos=label ; *pos ; pos++)
    new[len++] = *pos;
  new[len++] = ';';
  new[len] = '\0';
  
  /*copy over the new RTL*/
  FREE(rtl->val);
  rtl->val = (char *)CALLOC(strlen(new)+2);
  strcpy(rtl->val, new);

#ifdef INTERACT
  mod_rtl(rtl, NULL, rtl->val);
#endif    
}



/* init_rem_jump - intialize the data structures for jump removal
 *                 optimization. Return the highest block number present in
 *                 this function.
 */
int init_rem_jump(struct block_info binfo[MAX_BLOCKS], int weights[][MAX_BLOCKS])
{
  int i, j, count, num;
  int max_blk_num;
  struct bblock *cblk;
  struct list *rtl;
  extern struct bblock *top;

  /*initialize*/
  max_blk_num = 0;
  for(cblk=top ; cblk ; cblk=cblk->down){
    if(cblk->block_id > max_blk_num)
      max_blk_num = cblk->block_id;
  }
  max_blk_num++;

  for(i=0 ; i<max_blk_num ; i++){
    binfo[i].ptr = NULL;
    for(j=0 ; j<max_blk_num ; j++)
      weights[i][j] = INT_INFI;
  }

  /*get relevant information about each basic block. Currently, we weigh each
    basic block by the number of RTLs in that block. Also storing the pointer
    to the block in the bblock structure, for quick access later*/
  for(cblk=top ; cblk ; cblk=cblk->down){

    binfo[cblk->block_id].ptr = (struct bblock*)cblk;
    
    /*count the number of RTLs*/
    count = 0;
    for(rtl=cblk->rtls ; rtl ; rtl=rtl->next){
      if(cnts(rtl))
	count++;
    }
    binfo[cblk->block_id].num_rtl = count;
  }
  
  /*find the weights. The weights can be either number of RTLs or basic
    blocks.*/
  for(cblk=top ; cblk ; cblk=cblk->down){
    num = cblk->block_id;
    if(cblk->left)
      weights[num][cblk->left->block_id] = binfo[cblk->left->block_id].num_rtl;
    if(cblk->right)
      weights[num][cblk->right->block_id] = binfo[cblk->right->block_id].num_rtl;
  }

  return max_blk_num;
}


/* det_cheap_path - determine the cheapest path for replication to remove the
 *                  unconditional jump. Return the number of blocks in the
 *                  cheapest path, or -1 in case of error
 */
int det_cheap_path(struct block_info binfo[], int start, int dist[],
		   int path[], int rep_path[])
{
  int i, j, count, min_dist, blk_id, is_ret_blk;
  int target;
  struct bblock *cblk;
  struct blist *bl;
  extern struct bblock *top, *ret_block;

  /*initialize min_dist to the distance of the path from the target of the
    jump to the block following the jump in the control flow*/
  cblk = (struct bblock *)binfo[start].ptr;
  target = cblk->down->block_id;

  /*all distances in the dist vector include distances from the source to the
    target, but also include the weight of the target block which is not
    replicated, so we subtract that*/
  min_dist = dist[target] - binfo[target].num_rtl;

  /*check if any distance from the target block to any of the return blocks
    is smaller. Here we do infact need to replicate the target block as well*/
  is_ret_blk = FALSE;
  for(bl=ret_block->preds ; bl ; bl=bl->next){
    cblk = bl->ptr;
    blk_id = cblk->block_id;
    if(dist[blk_id]+binfo[blk_id].num_rtl < min_dist){
      min_dist = dist[blk_id] + binfo[blk_id].num_rtl;
      target = blk_id;
      is_ret_blk = TRUE;
    }
  }

  /*get the path of the basic blocks from the JUMP to the target which need
    to be replicated*/
  i = 0;
  if(min_dist < INT_INFI){

    /*if this is the return block then also include the pred of the return
      block along which this path is taken*/
    if(is_ret_blk)
      rep_path[i++] = target;

    blk_id = target;
    while(path[blk_id] != start){
      if(dist[blk_id] == INT_INFI){
	fprintf(stderr, "Path along the shortest path does not exist. Do not know if this will happen\n");
	return -1;
      }
      rep_path[i++] = path[blk_id];
      blk_id = path[blk_id];
    }
    /*rep_path[i++] = start;*/

    /*reverse the path*/
    count = i;
    for(i=0, j=count-1 ; i<count/2 ; i++, j--){
      blk_id = rep_path[i];
      rep_path[i] = rep_path[j];
      rep_path[j] = blk_id;
    }
    rep_path[count] = -1;

    /*debug info*/
    fprintf(stderr, "The cheapest path is: ");
    for(i=0 ; rep_path[i]!=-1 ; i++)
      fprintf(stderr, "%d ", rep_path[i]);
    fprintf(stderr, "\n");
    
    return count;
  }

  return -1;
}


/* get_last_rtl - get the last valid RTL line in this basic block
 */
struct list *get_last_rtl(struct bblock *cblk)
{
  struct list *rtl;

  for(rtl=cblk->rtlend ; rtl && !(rtl->type & RTLINE) ; rtl=rtl->prev);

  if(rtl && rtl->type & RTLINE)
    return rtl;

  return (struct list *)NULL;
}


/* draw_graph - debug function
 * display the control flow of the current function. This function creates a
 * file which can either be run through "dot" to create a postscript file, or
 * "display" should be able to interpret it directly, assuming that "dot"
 * executable is present in the current directory
 */
void draw_graph()
{
  FILE *draw_fp;
  char drawbuf[BUFSIZ], str[1001];
  struct bblock *cblk;
  extern struct bblock *top;
  extern char *fn;

  sprintf(str, "draw_%s", fn);
  if((draw_fp = fopen(str, "w")) == NULL){
    fprintf(stderr, "draw_graph: File %s cannot be opened for wrinting\n", str);
    return;
  }
  setbuf(draw_fp, drawbuf);

  fprintf(draw_fp, "digraph G{\n");

  /*just print the blocks to get the control flow right. So, print the edges
    in white so they are not visible*/
  for(cblk=top ; cblk->down ; cblk=cblk->down)
      fprintf(draw_fp, "%d -> %d [color=white]\n",
	      cblk->block_id, cblk->down->block_id);    
  
  for(cblk=top ; cblk ; cblk=cblk->down){

    if(cblk->status&JUMP && is_ucjmp(cblk->rtlend->val)){

      if(cblk->left)
	fprintf(draw_fp, "%d -> %d [style=filled,color=blue]\n",
		cblk->block_id, cblk->left->block_id);
      else
	fprintf(stderr, "Jump block does not have a left block\n");
    }
    else if(cblk->left)
      fprintf(draw_fp, "%d -> %d [color=green]\n",
	      cblk->block_id, cblk->left->block_id);
    
    if(cblk->right)
      fprintf(draw_fp, "%d -> %d\n", cblk->block_id, cblk->right->block_id);
  }
  fprintf(draw_fp, "}\n");
  fclose(draw_fp);
}

#endif /*INTERACT*/
