/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/


#include <stdlib.h>
#include <assert.h>

#include "vpo.h"

/*
 * Functions to read/create register in a size independent way
 */
char* getReg(int type, int number)
{
	/* note the static array
	 * we use it so we don't have to allocate memory
	 * each time this func is called
	 * however, the caller needs to be careful to copy/discard
	 * the results before calling this function again.
	 */
	static char temp[REGLOCLENGTH+1]; 

	/* not all 8 bits are usuable, the upper bit must be 1 */
	int useableBitsPerByte=USEABLECHARBITS;

	/* and here we make masks for the lower 7 bits */
	int maxByteVal=(1<<useableBitsPerByte)-1;

	/* and the top-bit */
	int topBitMask=(1<<useableBitsPerByte);

	/* loop index */
	int i;	
	int typeMask;

	assert(number>=MINREGNUM && number<=MAXREGNUM);
	assert(type>=MINREGTYPENUM && type<=MAXREGTYPENUM);


	/* we set the last character and the first character to 0. */
	/* the rest get set in the for-loop */
	temp[0]=0;
	temp[REGLOCLENGTH]=0;
	for(i=REGLOCLENGTH-1; i>=0; i--) {
		temp[i]=(number&maxByteVal);
		temp[i]|=topBitMask;
		number=number>>useableBitsPerByte;
	}


	/* we'll assume this for now */
	/* and enforce it was an assetion so it's easy */
	/* to track.  We may want more than 2^6 register types */
	/* someday and this'll have to be fixed */
	assert(REGTYPEBITS<useableBitsPerByte);


	/* make a mask for the type bits in the upper byte */
	/* and make sure they are clear in both type, and temp[0]	 */
	typeMask=((1<<REGTYPEBITS)-1);
	assert((type&typeMask)==type);

	type    =type    <<(useableBitsPerByte-REGISTERIDENTSIZE-REGTYPEBITS);
	typeMask=typeMask<<(useableBitsPerByte-REGISTERIDENTSIZE-REGTYPEBITS);

	/* assert that those bits are already 0 */
	assert((temp[0]&(typeMask))==0);

	temp[0]|=type;	

	temp[0]|=(REGISTERIDENTMASK << (USEABLECHARBITS - REGISTERIDENTSIZE));

	return temp;
}


#ifdef SAFE_ISREGISTER
int isRegister(char* location)
{
	int upperBits;
	int i;

	for(i=0;i<REGLOCLENGTH;i++) {	/* note that this bails out if we find
						                  a null terminator */
		unsigned char c;
		c=location[i];
		if((c>>USEABLECHARBITS)!=1)
			return FALSE;
	}
	upperBits=(location[0]>>(USEABLECHARBITS-REGISTERIDENTSIZE));
	upperBits&=((1<<REGISTERIDENTSIZE)-1);
	return (upperBits==REGISTERIDENTMASK);
}
#endif


int getRegType(char* t)
{
	int s;
        /* we'll assume this for now */
        /* and enforce it was an assetion so it's easy */
        /* to track.  We may want more than 2^6 register types */
        /* someday and this'll have to be fixed */
        assert(REGTYPEBITS<USEABLECHARBITS);

	/* some sanity checks, not efficient, but since i'm re-writing */
	/* i need the safety */
	assert(t);
	if(!isRegister(t))
	{
		fprintf(stderr,"Found call to getRegType with param that isn't "
                     " a register.  PID %d going into infinite loop so\n"
                     " that a debugger can be attached\n", getpid());
		fflush(stderr);
		while(1);
	}
	s=(unsigned char)*t;
	

	s=s>>(USEABLECHARBITS-REGISTERIDENTSIZE-REGTYPEBITS);

	s&=((1<<REGTYPEBITS)-1);

	assert(s<=MAXREGNUM && s>=MINREGNUM);

	return s;
}

static int getLocationNumber(unsigned char* t, int len)
{
	int answer=0;
	int i;
	int shiftDist=0;
	for(i=len-1; i>=0;i--) {
		int theBits;
		theBits=t[i]& ((1<<USEABLECHARBITS)-1);
		answer|=(theBits<<shiftDist);
		shiftDist+=USEABLECHARBITS;
	}
	return answer;
}

int getRegNum(char *s)
{
	unsigned char t[REGLOCLENGTH+1];
   int mask;

	strncpy(t, s, REGLOCLENGTH);

        /* we'll assume this for now */
        /* and enforce it with an assertion so it's easy */
        /* to track.  We may want more than 2^6 register types */
        /* someday and this'll have to be fixed */
        assert(REGTYPEBITS<USEABLECHARBITS);

        /* some sanity checks, not efficient, but since i'm re-writing */
        /* i need the safety */
        assert(t);
        /*assert(isRegister((char*)t));*/
	if(!isRegister(s))
	{
		fprintf(stderr,"Found call to getRegNum with param that isn't "
                     " a register.  PID %d going into infinite loop so\n"
                     " that a debugger can be attached\n", getpid());
		fflush(stderr);
		while(1);
	}
        

        mask=(USEABLECHARBITS-REGISTERIDENTSIZE-REGTYPEBITS);

        (*t)&=((1<<mask)-1);

        return getLocationNumber(t, REGLOCLENGTH);
}




/*
 * Functions to read/create globals in a size independent way 
 */
char* getGlobal(int number)
{
	/* note the static array */
	/* we use it so we don't have to allocate memory */
	/* each time this func is called */
	/* however, the caller needs to be careful to copy/discard */
	/* the results before calling this function again. */
	static char temp[GLOBALLOCLENGTH+1]; 

	/* not all 8 bits are usuable, the upper bit must be 1 */
	int useableBitsPerByte=USEABLECHARBITS;

	/* and here we make masks for the lower 7 bits */
	int maxByteVal=(1<<useableBitsPerByte)-1;

	/* and the top-bit */
	int topBitMask=(1<<useableBitsPerByte);

	/* loop index */
	int i;	

	assert(number>=MINGLOBNUM && number<=MAXGLOBNUM);


	/* we set the last character and the first character to 0. */
	/* the rest get set in the for-loop */
	temp[0]=0;
	temp[GLOBALLOCLENGTH]=0;
	for(i=GLOBALLOCLENGTH-1; i>=0; i--) {
		temp[i]=(number&maxByteVal);
		temp[i]|=topBitMask;
		number=number>>useableBitsPerByte;
	}

	temp[0]|=topBitMask;
	temp[0]|=(GLOBALIDENTMASK << (USEABLECHARBITS - GLOBALIDENTMASKSIZE));

	return temp;
}

int isGlobal(char* location)
{
	int upperBits;
	int i;
	for (i = 0; i < GLOBALLOCLENGTH; ++i) {       
		unsigned char c;
		c = location[i];
		if ((c >> USEABLECHARBITS) != 1)
			return FALSE;
	}
	upperBits = (location[0] >> (USEABLECHARBITS - GLOBALIDENTMASKSIZE));
	upperBits &= ((1 << GLOBALIDENTMASKSIZE) - 1);
	return (upperBits == GLOBALIDENTMASK);
}

int getGlobalNum(char* s)
{
	unsigned char t[GLOBALLOCLENGTH+1];
	int mask;

	strncpy(t,s,GLOBALLOCLENGTH);


	/* some sanity checks, not efficient, but since i'm re-writing */
	/* i need the safety */
	assert(t);
	if(!isGlobal(s)) {
		fprintf(stderr,"Found call to getGlobalNum with param that isn't "
                     " a register.  PID %d going into infinite loop so\n"
                     " that a debugger can be attached\n", getpid());
		fflush(stderr);
		while(1);
	}


	mask=(USEABLECHARBITS-GLOBALIDENTMASKSIZE);

	(*t)&=((1<<mask)-1);

	return getLocationNumber(t, GLOBALLOCLENGTH);
}


/* 
 * Functions to read/create locals in a size independent way 
 */
char* getLocal(int number)
{
	/* note the static array */
	/* we use it so we don't have to allocate memory */
	/* each time this func is called */
	/* however, the caller needs to be careful to copy/discard */
	/* the results before calling this function again. */
	static char temp[LOCALLOCLENGTH+1]; 

	/* not all 8 bits are usuable, the upper bit must be 1 */
	int useableBitsPerByte=USEABLECHARBITS;

	/* and here we make masks for the lower 7 bits */
	int maxByteVal=(1<<useableBitsPerByte)-1;

	/* and the top-bit */
	int topBitMask=(1<<useableBitsPerByte);

	/* loop index */
	int i;	

	assert(number>=MINLOCALNUM && number<=MAXLOCALNUM);


	/* we set the last character and the first character to 0. */
	/* the rest get set in the for-loop */
	temp[0]=0;
	temp[LOCALLOCLENGTH]=0;
	for(i=LOCALLOCLENGTH-1; i>=0; i--) {
		temp[i]=(number&maxByteVal);
		temp[i]|=topBitMask;
		number=number>>useableBitsPerByte;
	}

	temp[0]|=topBitMask;
	temp[0]|=(LOCALIDENTMASK << (USEABLECHARBITS - LOCALIDENTMASKSIZE));

	return temp;
}

int isLocal(char* location)
{
	int upperBits;
	int i;
	for(i=0;i<LOCALLOCLENGTH;i++) {       

		unsigned char c;
		c=location[i];
		if((c>>USEABLECHARBITS)!=1)
			return FALSE;
	}
	upperBits=(location[0]>>(USEABLECHARBITS-LOCALIDENTMASKSIZE));
	upperBits&=((1<<LOCALIDENTMASKSIZE)-1);
	return (upperBits==LOCALIDENTMASK);
}

int getLocalNum(char* s)
{
	unsigned char t[LOCALLOCLENGTH+1];
	int mask;

	strncpy(t,s,LOCALLOCLENGTH);


	/* some sanity checks, not efficient, but since i'm re-writing */
	/* i need the safety */
	assert(t);
	if(!isLocal((char*)s))
	{
		fprintf(stderr,"Found call to getLocalNum with param that isn't "
                     " a register.  PID %d going into infinite loop so\n"
                     " that a debugger can be attached\n", getpid());
		fflush(stderr);
		while(1);
	}


	mask=(USEABLECHARBITS-LOCALIDENTMASKSIZE);

	(*t)&=((1<<mask)-1);

	return getLocationNumber(t, LOCALLOCLENGTH);
}

void copyLocation(char* a, char* b)
{
	int i, len;

	if(isRegister(b))
		len=REGLOCLENGTH;
	else if(isGlobal(b))
		len=GLOBALLOCLENGTH;
	else if(isLocal(b))
		len=LOCALLOCLENGTH;
	else
		len=1;

	for(i=0;i<len;i++)
		a[i]=b[i];
}

void copyreg(char *a, char *b)
{
  copyLocation(a, b);
  a[REGLOCLENGTH] = '\0';
}


/* works like strcmp on locations */
int compareLocation(char* a, char* b)
{
	int i, len;

	if(isRegister(a)) {
		if(isRegister(b))
			len=REGLOCLENGTH;
		else
			return TRUE;
	}
	else if(isGlobal(a)) {
		if(isGlobal(b))
			len=GLOBALLOCLENGTH;
		else
			return TRUE;
	}
	else if(isLocal(a)) {
		if(isLocal(b))
			len=LOCALLOCLENGTH;
		else
			return TRUE;
	}
	else
		len=1;

	for(i=0;i<len;i++)
		if(a[i]!=b[i])
			return a[i]-b[i];
	return FALSE;
}


void mangleReg(char* dst, char *src, char* mangler)
{
	int i=0;

	assert(REGLOCLENGTH>1);
	assert(isreg(src));

	*dst++=*src++;
	while(*dst++=*mangler++)
		;
	dst--;
	for(i=1;i<REGLOCLENGTH;i++)
	  *dst++=*src++;
}

void demangleReg(char* dst, char* src, char* mangler)
{
	char* a=dst;
	int i;
	assert(REGLOCLENGTH>1);


	*a++=*src++;

	while(*mangler) {
		assert(*src==*mangler);
		src++; 
		mangler++;
	}

	for(i=1;i<REGLOCLENGTH;i++)
		*a++=*src++;

	assert(isreg(dst));
}


/* return the "mangler" if it is a mangled register */
char* isMangledReg(char* src, char* mangler)
{
	static char dst[REGLOCLENGTH];
	char* a=dst;
	char* b=src+1;
	int i;

	assert(REGLOCLENGTH>1);


	*a++=*src++;

	while(*mangler) {
		/* use a '?' as a one-character wildcard */
		if(*mangler!='?' && *src!=*mangler)
			return 0;
		src++; 
		mangler++;
	}
	for(i=1;i<REGLOCLENGTH;i++)
		*a++=*src++;

	if(isreg(dst)) {
		strncpy(dst, b, strlen(mangler));
		*(dst+strlen(mangler))=0;
		return dst;
	}
	return 0;

}


void locString(char* dst, char* src) 
{
	copyloc(dst,src);
	*(skiploc(dst))=0;
}

