/******************************************************************************
 *
 * Copyright (c) 1993 by Mark W. Bailey.
 *
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made as to
 * whether it serves any particular purpose or even works at all.
 *
 *****************************************************************************/

/******************************************************************************
 *
 * File: expr.c
 *
 * Author: Mark W. Bailey
 *
 * Created: 12/18/93
 *
 * Description:
 *
 *      The file contains all of the constructors, or functions to 
 *      make a new expression of a particular type. Most of these 
 *      functions simply allocate a new node and assign the parameters 
 *      to the appropriate fields of the structures.
 *
 * RCS Header:
 * 	$Header: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/rtlprep/expr.c,v 1.1.1.1 2000/08/22 21:10:33 zephyr Exp $
 *
 * Change Log:
 *	$Log: expr.c,v $
 *	Revision 1.1.1.1  2000/08/22 21:10:33  zephyr
 *	Created directory structure
 *	
 *	Revision 1.3  1998/08/07 16:39:00  jwd
 *	Added support for ARGBUILDAREA interface. This is to
 *	support the new RTL creation interface.
 *	
 *	Revision 1.2  1998/04/08 18:27:15  jwd
 *	Cosmetic changes only.
 *	Revision 1.1  1998/01/29 20:53:54  jch8f
 *	Initial revision
 *
 * Revision 1.1  1994/01/13  16:38:52  mwb5y
 * Initial revision
 *
 *
 * Revsion: 1998/1/29 jch8f
 *   Modifying the code to make the function prototyping ansi-C compatable
 *****************************************************************************/


#include <stdio.h>
#include "gencode.h"
#include "expr.h"
#include <malloc.h>

#define NEW(TYPE) ((TYPE *) malloc(sizeof(TYPE)))

/****
 *
 *  AllocateString - Allocates a string for storing a constant. This 
 *                   will later cause the string to be declared unless 
 *                   DeAllocate string is called beforehand. 
 *                   
 *                   This function only has effect in the ntro 
 *                   generator.
 * 
 ****/

static void AllocateString(CExpr *expr)
{
    expr->stackid = ConstSp;
    ConstStack[ConstSp++] = expr->expr;
}


/****
 *
 *  DeAllocateString - This function deallocates a string that has 
 *                     been allocated by AllocateString.
 * 
 ****/

static void DeAllocateString(CExpr *expr)
{
    if (expr->stackid + 1 == ConstSp)
	ConstSp--;
    ConstStack[expr->stackid] = NULL;
}



/****
 *
 *  C2InternalC - Changes the type of a CType node to an InternalC
 *                node and deallocates the associated string.
 * 
 ****/

Expr *C2InternalC(Expr *expr)
{
    DeAllocateString(&expr->value.cexpr);
    expr->type = InternalCType;
    return(expr);
}


/****
 *
 *  I2InternalConst - Changes the type of a IConst node to an 
 *  		      InternalConst node.
 * 
 ****/

Expr *I2InternalConst(Expr *expr)
{
    expr->type = InternalConstType;
    return(expr);
}

/****
 *
 *  MakeTransfer - Builds a transfer expression node.
 * 
 ****/

Expr *MakeTransfer(Expr *transfers, Expr *deads)
{
    Expr *new = NEW(Expr);

    new->type = TransferLineType;
    new->value.transfer.transfers = transfers;
    new->value.transfer.deads = deads;
    return(new);
}


/****
 *
 *  MakeExprList - Builds an expression list node.
 * 
 ****/

Expr *MakeExprList(Expr *expr, Expr *list)
{
    Expr *new = NEW(Expr);

    new->type = ExprListType;
    new->value.exprlist.expr = expr;
    new->value.exprlist.list = list;
    return(new);
}


/****
 *
 *  MakeEffect - Builds an RTL effect node.
 * 
 ****/

Expr *MakeEffect(char *type, Expr *destination, Expr *source, int optional)
{
    Expr *new = NEW(Expr);

    new->type = EffectType;
    new->value.effect.type = type;
    new->value.effect.destination = destination;
    new->value.effect.source = source;
    new->value.effect.optional = optional;
    return(new);
}


/****
 *
 *  MakeDeadList - Builds a dead variable list node.
 * 
 ****/

Expr *MakeDeadList(Expr *reg, Expr *list)
{
    Expr *new = NEW(Expr);

    new->type = DeadListType;
    new->value.deadlist.reg = reg;
    new->value.deadlist.list = list;
    return(new);
}


/****
 *
 *  MakeUseLine - Builds a USE line node.
 * 
 ****/

Expr *MakeUseLine(Expr *reg)
{
    Expr *new = NEW(Expr);

    new->type = UseLineType;
    new->value.useline.reg = reg;
    return(new);
}

/****
 *
 *  MakeArgBuildAreaLine - Build an Argument Build Area line node.
 *
 ****/

Expr *MakeArgBuildAreaLine(Expr *name)
{
    Expr *new = NEW(Expr);

    new->type = ArgBuildAreaLineType;
    new->value.argbuildarealine.asize = name;
    return(new);
}


/****
 *
 *  MakeAsmLine - Builds an ASSEMBLY line node.
 *
 ****/

Expr *MakeAsmLine(Expr *expr)
{
    Expr *new = NEW(Expr);

    new->type = AsmLineType;
    new->value.asmline.expr = expr;
    return(new);
}


/****
 *
 *  MakeLabelLine - Builds a LABEL line node.
 * 
 ****/

Expr *MakeLabelLine(Expr *label)
{
    Expr *new = NEW(Expr);

    new->type = LabelLineType;
    new->value.labelline.label = label;
    return(new);
}


/****
 *
 *  MakeMemory - Builds a memory expression node.
 *  
 ****/

Expr *MakeMemory(Expr *type, Expr *address)
{
    Expr *new = NEW(Expr);
    
    new->type = MemoryType;
    new->value.memory.type = type;
    new->value.memory.address = address;
    return(new);
}


/****
 *
 *  MakeMemoryType - Builds a memory type node.
 * 
 ****/

Expr *MakeMemoryType(char type)
{
    Expr *new = NEW(Expr);

    new->type = MemoryTypeType;
    new->value.memorytype.type = type;
    return(new);
}


/****
 *
 *  MakeRegisterType - Builds a register node.
 * 
 ****/

Expr *MakeRegister(Expr *type, Expr *number)
{
    Expr *new = NEW(Expr);
    
    new->type = RegisterType;
    new->value.reg.type = type;
    new->value.reg.number = number;
    return(new);
}


/****
 *
 *  MakeRegisterType - Builds a register type node.
 * 
 ****/

Expr *MakeRegisterType(char type)
{
    Expr *new = NEW(Expr);

    new->type = RegisterTypeType;
    new->value.registertype.type = type;
    return(new);
}


/****
 *
 *  MakeGroup - Builds a grouped expression node.
 * 
 ****/

Expr *MakeGroup(Expr *expr)
{
    Expr *new = NEW(Expr);

    new->type = GroupType;
    new->value.group.expr = expr;
    return(new);
}


/****
 *
 *  MakeBinary - Builds a binary operator expression node.
 * 
 ****/

Expr *MakeBinary(char *op, Expr *left, Expr *right)
{
    Expr *new = NEW(Expr);
    
    new->type = BinaryType;
    new->value.binary.op = op;
    new->value.binary.left = left;
    new->value.binary.right = right;
    return(new);
}


/****
 *
 *  MakeUnary - Builds a unary operator expression node.
 * 
 ****/

Expr *MakeUnary(char *op, Expr *operand)
{
    Expr *new = NEW(Expr);
    
    new->type = UnaryType;
    new->value.unary.op = op;
    new->value.unary.operand = operand;
    return(new);
}


/****
 *
 *  MakeMacro - Builds a macro node.
 * 
 ****/

Expr *MakeMacro(char *name, Expr *operand)
{
    Expr *new = NEW(Expr);
    
    new->type = MacroType;
    new->value.macro.name = name;
    new->value.macro.operand = operand;
    return(new);
}


/****
 *
 *  MakeVariable - Builds a variable node.
 * 
 ****/

Expr *MakeVariable(char *type, Expr *name)
{
    Expr *new = NEW(Expr);
    
    new->type = VariableType;
    new->value.variable.type = type;
    new->value.variable.name = name;
    return(new);
}


/****
 *
 *  MakeLabel - Builds a label node.
 * 
 ****/

Expr *MakeLabel(char *type, char *prefix, Expr *name)
{
    Expr *new = NEW(Expr);
    
    new->type = LabelType;
    new->value.label.type = type;
    new->value.label.prefix = prefix;
    new->value.label.name = name;
    return(new);
}


/****
 *
 *  MakeC - Builds a C expression node. Automatically allocates a 
 *  	    string for the C type for use in conversion.
 * 
 ****/

Expr *MakeC(char *expr)
{
    Expr *new = NEW(Expr);
    
    new->type = CType;
    new->value.cexpr.expr = expr;
    AllocateString(&new->value.cexpr);
    return(new);
}


/****
 *
 *  MakeEscExpr - Builds an escaped expression node.
 * 
 ****/

Expr *MakeEscExpr(char *expr)
{
    Expr *new = NEW(Expr);
    
    new->type = EscExprType;
    new->value.escexpr.expr = expr;
    return(new);
}


/****
 *
 *  MakeTypedConst - Builds a typed constant node.
 * 
 ****/

Expr *MakeTypedConst(Expr *type, Expr *value)
{
    Expr *new = NEW(Expr);

    new->type = TypedConstType;
    new->value.typedconstant.type = type;
    new->value.typedconstant.value = value;
    return(new);
}


/****
 *
 *  MakeConstType - Builds a constant type node.
 * 
 ****/

Expr *MakeConstType(char *type)
{
    Expr *new = NEW(Expr);

    new->type = ConstTypeType;
    new->value.consttype.type = type;
    return(new);
}


/****
 *
 *  MakeIConst - Builds an integer constant node.
 * 
 ****/

Expr *MakeIConst(char *value)
{
    Expr *new = NEW(Expr);

    new->type = IConstType;
    new->value.iconst.value = value;
    return(new);
}


/****
 *
 *  MakeFConst - Builds a floating-point constant node.
 * 
 ****/

Expr *MakeFConst(char *value)
{
    Expr *new = NEW(Expr);

    new->type = FConstType;
    new->value.fconst.value = value;
    return(new);
}











