/******************************************************************************
 *
 * Copyright (c) 1993 by Mark W. Bailey.
 *
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made as to
 * whether it serves any particular purpose or even works at all.
 *
 *****************************************************************************/

/******************************************************************************
 *
 * File: expr.h
 *
 * Author: Mark W. Bailey
 *
 * Created: 12/18/93
 *
 * Description:
 *
 *      All the expression types that are used in ART are defined 
 *      here. 
 *
 * RCS Header:
 * 	$Header: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/rtlprep/expr.h,v 1.1.1.1 2000/08/22 21:10:33 zephyr Exp $
 *
 * Change Log:
 *	$Log: expr.h,v $
 *	Revision 1.1.1.1  2000/08/22 21:10:33  zephyr
 *	Created directory structure
 *	
 *	Revision 1.3  1998/08/07 16:39:25  jwd
 *	Added support for ARGBUILDAREA interface. This is to
 *	support the new RTL creation interface.
 *	
 *	Revision 1.2  1998/04/08 18:27:43  jwd
 *	Cosmetic changes only.
 *	Revision 1.1  1998/01/29 20:54:04  jch8f
 *	Making the file compatible with ANSI-C function prototyping.
 *
 *	Revision 1.2  1997/07/09 14:29:29  mwb5y
 *	Removed trailing comma in ExprType to shut up -pedantic for gcc.
 *
 *	Revision 1.1  1994/01/13 16:39:38  mwb5y
 *	Initial revision
 *
 *
 *****************************************************************************/


#ifndef __EXPR_H__
#define __EXPR_H__

typedef enum {
    TransferLineType,		/* Whole transfer line (many effects) */
    ExprListType,		/* Semicolon separated list of expressions */
    EffectType,			/* An effect */
    UseLineType,		/* A USE line */
    AsmLineType,		/* An ASSEMBLY line */
    ArgBuildAreaLineType,       /* An argument build area line */
    LabelLineType,		/* A LABEL line */
    DeadListType,		/* Dead variable list */
    MemoryType,			/* Memory expression */
    MemoryTypeType,		/* Type part of a memory expression*/
    RegisterType,		/* Register expression */
    RegisterTypeType,		/* Type part of a register expression */
    GroupType,			/* Placeholder for parenthesis (for vpo) */
    BinaryType,			/* Binary operator expression */
    UnaryType,			/* Unary operator expression */
    MacroType,			/* Macro expression */
    VariableType,		/* Global or local variable */
    LabelType,			/* Label expression */
    CType,			/* C code that evaluates to an int */
    InternalCType,		/* A converted CType used in a larger expr */
    EscExprType,		/* An escaped expression */
    TypedConstType,		/* A typed constant */
    InternalConstType,		/* A converted TypeConst used in larger */
    ConstTypeType,		/* Type part of constant expression */
    IConstType,			/* Integer constant */
    FConstType			/* Floating-point constant */
} ExprType;

/* 
 * Forward declare these types so that we can use them in defining the 
 * expression types.
 */

struct Expr;
typedef struct Expr Expr;

typedef struct {
    Expr	*transfers;	/* The effects */
    Expr	*deads;		/* The dead register list */
} TransferExpr;

typedef struct {
    Expr	*expr;		/* An effect */
    Expr	*list;		/* The rest of the effects */
} ExprListExpr;

typedef struct {
    char	*type;		/* Effect type */
    Expr	*destination;	/* Destination of transfer */
    Expr	*source;	/* Source of transfer */
    int		optional;	/* Flag indicating if effect is optional */
} EffectExpr;

typedef struct {
    Expr	*reg;		/* A dead register */
    Expr	*list;		/* The rest of the dead registers */
} DeadListExpr;

typedef struct {
    Expr	*reg;		/* The used register */
} UseLineExpr;

typedef struct {
    Expr	*asize;		/* The argument build area size */
} ArgBuildAreaExpr;

typedef struct {
    Expr	*expr;		/* The assembly expression */
} AsmLineExpr;

typedef struct {
    Expr	*label;		/* The label */
} LabelLineExpr;

typedef struct {
    Expr	*type;		/* Type of memory */
    Expr	*address;	/* Address of memory reference */
} MemoryExpr;

typedef struct {
    char	type;		/* Memory type */
} MemoryTypeExpr;

typedef struct {
    Expr	*type;		/* Type of register */
    Expr	*number;	/* Number of register */
} RegisterExpr;
 
typedef struct {
    char	type;		/* Register type */
} RegisterTypeExpr;

typedef struct {
    Expr	*expr;		/* Expression surrounded by parenthesis */
} GroupExpr;
 
typedef struct {
    char	*op;		/* Operator */
    Expr	*left;		/* Left hand side operand */
    Expr	*right;		/* Right hand side operand  */
} BinaryExpr;

typedef struct {
    char	*op;		/* Operator */
    Expr	*operand;	/* Operand */
} UnaryExpr;

typedef struct {
    char	*name;		/* Macro name */
    Expr	*operand;	/* Optional macro operand */
} MacroExpr;

typedef struct {
    char	*type;		/* Type of variable (global or local) */
    Expr	*name;		/* Name of variable */
} VariableExpr;

typedef struct {
    char	*type;		/* Type of label (BE, CE, or FE) */
    char 	*prefix;	/* Prefix of label string (e.g. L0) */
    Expr	*name;		/* Name, or actually, number of label */
} LabelExpr;

typedef struct {
    char	*expr;		/* The C expression */
    int		stackid;	/* String number used to convert */
} CExpr;

typedef struct {
    char	*expr;		/* The actual escaped string */
} EscExprExpr;

typedef struct {
    Expr	*type;		/* Type of constant */
    Expr	*value;		/* Value of constant */
} TypedConstExpr;

typedef struct {
    char	*type;		/* Type (e.g. WORD, DOBL, etc) */
} ConstTypeExpr;

typedef struct {
    char	*value;		/* Integer constant */
} IConstExpr;

typedef struct {
    char	*value;		/* Floating-point constant */
} FConstExpr;

union Exprs {
    TransferExpr	transfer;
    ExprListExpr	exprlist;
    EffectExpr		effect;
    DeadListExpr	deadlist;
    UseLineExpr		useline;
    ArgBuildAreaExpr    argbuildarealine;
    AsmLineExpr		asmline;
    LabelLineExpr	labelline;
    MemoryExpr		memory;
    MemoryTypeExpr	memorytype;
    RegisterExpr	reg;
    RegisterTypeExpr	registertype;
    GroupExpr		group;
    BinaryExpr		binary;
    UnaryExpr		unary;
    MacroExpr		macro;
    VariableExpr	variable;
    LabelExpr		label;
    TypedConstExpr	typedconstant;
    ConstTypeExpr	consttype;
    IConstExpr		iconst;
    FConstExpr		fconst;
    CExpr		cexpr;
    EscExprExpr		escexpr;
};

struct Expr {
    ExprType	type;		/* Flags which element in union is valid */
    union Exprs	value;		/* Actual value of expression */
};

Expr *MakeTransfer(Expr *transfers, Expr *deads);
Expr *MakeExprList(Expr *expr, Expr *list);
Expr *MakeEffect(char *type, Expr *destination, Expr *source, int optional);
Expr *MakeDeadList(Expr *reg, Expr *list);
Expr *MakeUseLine(Expr *reg);
Expr *MakeArgBuildAreaLine(Expr *name);
Expr *MakeAsmLine(Expr *expr);
Expr *MakeLabelLine(Expr *label);
Expr *MakeMemory(Expr *type, Expr *address);
Expr *MakeMemoryType(char type);
Expr *MakeRegister(Expr *type, Expr *number);
Expr *MakeRegisterType(char type);
Expr *MakeGroup(Expr *expr);
Expr *MakeBinary(char *op, Expr *left, Expr *right);
Expr *MakeUnary(char *op, Expr *operand);
Expr *MakeMacro(char *name, Expr *operand);
Expr *MakeVariable(char *type, Expr *name);
Expr *MakeLabel(char *type, char *prefix, Expr *name);
Expr *MakeC(char *expr);
Expr *C2InternalC(Expr *expr);
Expr *MakeEscExpr(char *expr);
Expr *MakeTypedConst(Expr *type, Expr *value);
Expr *I2InternalConst(Expr *expr);
Expr *MakeConstType(char *type);
Expr *MakeIConst(char *value);
Expr *MakeFConst(char *value);

#endif /* __EXPR_H__ */
